/*
 * @(#)ClassLoaderObjectInputStream.java
 * 
 * Utility classes
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util;

import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.StreamCorruptedException;
import java.io.InputStream;
import java.io.ObjectStreamClass;

/**
A stream defining the classes which are deserialized not through the standard classloader but
a classloader specified by the user.

@author  Michael Sonntag
@version 1.0, 1.7.2000
*/
public class ClassLoaderObjectInputStream extends ObjectInputStream
{
    /**
    The classloader associated with this stream. All classes restored will be loaded through this stream.
    */
    private ClassLoader classLoader=null;

    /**
    Create an ClassLoaderObjectInputStream that reads from the specified InputStream.
    @param in the underlying <code>InputStream</code>
    @throws StreamCorruptedException if the version or magic number are incorrect.
    @throws IOException if an exception occurred in the underlying stream.
    */
	public ClassLoaderObjectInputStream(InputStream in) throws IOException, StreamCorruptedException
	{
		super(in);
	}

	/**
	Sets the classloader through which to load the classes.
	@param classLoader the classloader to use when creating objects
	*/
	public final void setClassLoader(ClassLoader classLoader)
	{
	    // classLoader might be 'null', if called by a class loaded from the system!
		if(this.classLoader!=null && classLoader==null)
		    throw new IllegalArgumentException("Classloader must be specified; removing not possible");
		if(this.classLoader!=null)
		    throw new RuntimeException("Classloader already set");
        this.classLoader=classLoader;
	}

	/**
	Returns the classloader used for laoding the classes in this stream.
	Necessary for SignedMessage; it must load the contained class through the classloader it loads itself
	@throws SecurityException if the caller has no permission to access a classloader (<code>RuntimePermission("getClassLoader")</code>)
	@see PkgAgentSystem.Messaging.SignedMessage
	*/
	public final ClassLoader getClassLoader()
	{
	    SecurityManager security=System.getSecurityManager();
	    if (security!=null)
    	    security.checkPermission(new RuntimePermission("getClassLoader"));
	    return classLoader;
	}

    /**
    Load the local class equivalent of the specified stream class description.
    This method will be invoked only once for each unique class in the stream.
    Once returned, the serialVersionUID of the class is compared to the
    serialVersionUID of the serialized class. If there is a mismatch, the
    deserialization fails and an exception is raised. <p>
    @param v The definition of the class to resolve (serialized information)
    @exception ClassNotFoundException If the class of a serialized object cannot be found.
    */
    public Class resolveClass(ObjectStreamClass v) throws IOException, ClassNotFoundException
    {
        Class res=null;
        if(classLoader==null)
            // If not classLoader specified, use standard implementation
            res=super.resolveClass(v);
        else
            // else load the class through the specified classLoader
            res=classLoader.loadClass(v.getName());
        return res;
    }

}
