/*
 * @(#)Base64Coder.java
 *
 * BASE64 Encoding/Decoding
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.Coder.Base64;

/**
A class for encoding and decoding using the algorithm BASE64.
BASE64 Encoding is defined in RFC 1341 chapter 5.2

@author  Michael Sonntag
@version 1.0, 1.7.2000
*/
public final class Base64Coder
{
    /** Base64 encoding characters */
    private final static char[] cEnc="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/".toCharArray();

    /** Padding character */
    private final static char cPAD='=';

    /** Conversion table for decoding */
    private final static byte[] dDec=new byte[256];
    static 
    {
        for(int i=0;i<256;i++) 
            dDec[i]=-1;
        for(int i='A';i<='Z';i++) 
            dDec[i]=(byte)(i-'A');
        for(int i='a';i<='z';i++) 
            dDec[i]=(byte)(26+i-'a');
        for(int i='0';i<='9';i++) 
            dDec[i]=(byte)(52+i-'0');
        dDec['+']=62;
        dDec['/']=63;
    }
    
    /**
    Encodes an array of bytes into an array of bytes by using the base64 algorithm. 
    Uses padding so output is always multiple of 4.
    @param raw the array of byte to encode
    @return the encoded array of bytes
    */
    public final static byte[] base64Encode(byte[] raw)
    {
        int pos=0;
        byte buf[]=new byte[3];
        int buffered;
        byte[] res=new byte[(raw.length+2)/3*4];
        int resPos=0;
        while(pos<raw.length)
        {
            buffered=0;
            // Fill buffer (3 bytes will be encoded to 4 characters)
            while(buffered<3 && pos<raw.length)
                buf[buffered++]=raw[pos++];
            // Fill rest of buffer with 0 bytes
            for(int i=buffered;i<buf.length;i++)
                buf[i]=(byte)'\0';
            // Convert first byte
            if(buffered>0)
            {
                res[resPos++]=(byte)cEnc[(buf[0]>>2)&0x3f];
                res[resPos++]=(byte)cEnc[((buf[0]<<4)&0x30)+((buf[1]>>4)&0x0f)];
            }
            // Convert second byte
            if(buffered>1)
                res[resPos++]=(byte)cEnc[((buf[1]<<2)&0x3c)+((buf[2]>>6)&0x03)];
            else // Only one byte, so pad result
                res[resPos++]=(byte)cPAD;
            // Convert third byte
            if(buffered>2)
                res[resPos++]=(byte)cEnc[buf[2]&0x3f];
            else // Only one or two bytes, so pad result
                res[resPos++]=(byte)cPAD;
        }
        return res;        
    }

    /**
    Encodes an array of bytes into a string by using the base64 algorithm.
    Uses padding so output is always multiple of 4.
    @param raw the array of byte to encode
    @return the encoded string
    */
    public final static String base64EncodeAsString(byte[] raw)
    {
        int pos=0;
        byte buf[]=new byte[3];
        int buffered;
        String res="";
        while(pos<raw.length)
        {
            buffered=0;
            // Fill buffer (3 bytes will be encoded to 4 characters)
            while(buffered<3 && pos<raw.length)
                buf[buffered++]=raw[pos++];
            // Fill rest of buffer with 0 bytes
            for(int i=buffered;i<buf.length;i++)
                buf[i]=(byte)'\0';
            // Convert first byte
            if(buffered>0)
            {
                res+=cEnc[(buf[0]>>2)&0x3f];
                res+=cEnc[((buf[0]<<4)&0x30)+((buf[1]>>4)&0x0f)];
            }
            // Convert second byte
            if(buffered>1)
                res+=cEnc[((buf[1]<<2)&0x3c)+((buf[2]>>6)&0x03)];
            else // Only one byte, so pad result
                res+=cPAD;
            // Convert third byte
            if(buffered>2)
                res+=cEnc[buf[2]&0x3f];
            else // Only one or two bytes, so pad result
                res+=cPAD;
        }
        return res;        
    }

    /**
    Decodes an array of bytes from an array of bytes by using the base64 algorithm
    Input length must be a multiple of 4 (padding required at end if necessary).
    @param encoded the encoded array of bytes
    @return the decoded array of bytes
    */
    public final static byte[] base64Decode(byte[] encoded)
    {
        int len=(encoded.length/4)*3;
        byte[] res=new byte[len];
        int pos=0;
        int buffer=0;
        int buffered=0;
        for(int i=0;i<encoded.length;i++)
        {
            int val=dDec[encoded[i]];
            if(val>=0)  // Skip padding bytes
            {
                buffer<<=6;
                buffered+=6;
                buffer|=val;
                if(buffered>=8)
                {
                    res[pos++]=(byte)((buffer>>(buffered-8))&0xff);
                    buffered-=8;
                }
            }
        }
        return res;
    }

    /**
    Decodes an array of bytes from a string by using the base64 algorithm.
    Input length must be a multiple of 4 (padding required at end if necessary).
    @param encoded the encoded string
    @return the decoded array of bytes
    */
    public final static byte[] base64Decode(String encoded)
    {
        return base64Decode(encoded.trim().getBytes());
    }
    
    public final static void main(String[] args)
    {
        byte[] raw=new byte[256];
        int i;
        for(i=0;i<256;i++)
            raw[i]=(byte)i;
        byte[] raw2=base64Decode(base64Encode(raw));
        boolean error=false;
        for(i=0;i<256 && !error;i++)
            error=raw2[i]!=(byte)i;
        if(error)
            System.out.println("Error at location "+i+": Expected "+i+" but found "+raw2[i]+"!");
        else
            System.out.println("Test 1 passed successfully!");
        raw2=base64Decode(base64EncodeAsString(raw));
        error=false;
        for(i=0;i<256 && !error;i++)
            error=raw2[i]!=(byte)i;
        if(error)
            System.out.println("Error at location "+i+": Expected "+i+" but found "+raw2[i]+"!");
        else
            System.out.println("Test 2 passed successfully!");
    }
}
