/*
 * @(#)Base64InputStream.java
 *
 * BASE64 Coding
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.Coder.Base64;

import java.io.*;
import FIM.Util.Coder.WSStripInputStream;
import FIM.Util.Coder.WrongFormatException;
 
/**
 * An input stream which decodes a base64-encoded stream.  
 *
 * @version     1.1, 12 Jan 1997
 * @author      Glenn Vanderburg
 * @author      Michael Sonntag (Corrections)
 */

public class Base64InputStream extends FilterInputStream
{
    /** Base64 padding byte */
    private static byte pad=(byte)'=';

    /** Code for a not allowed byte */
    private static int BADBYTE=-1;

    /** Base64 decoding table.  */
    private static int b[]=new int[256];
    static
    {
        for (int i=0;i<256;i++)
            b[i]=BADBYTE;
        b['A'] = 0;  b['B'] = 1;  b['C'] = 2;  b['D'] = 3;  b['E'] = 4;
        b['F'] = 5;  b['G'] = 6;  b['H'] = 7;  b['I'] = 8;  b['J'] = 9;
        b['K'] = 10; b['L'] = 11; b['M'] = 12; b['N'] = 13; b['O'] = 14;
        b['P'] = 15; b['Q'] = 16; b['R'] = 17; b['S'] = 18; b['T'] = 19;
        b['U'] = 20; b['V'] = 21; b['W'] = 22; b['X'] = 23; b['Y'] = 24;
        b['Z'] = 25; b['a'] = 26; b['b'] = 27; b['c'] = 28; b['d'] = 29;
        b['e'] = 30; b['f'] = 31; b['g'] = 32; b['h'] = 33; b['i'] = 34;
        b['j'] = 35; b['k'] = 36; b['l'] = 37; b['m'] = 38; b['n'] = 39;
        b['o'] = 40; b['p'] = 41; b['q'] = 42; b['r'] = 43; b['s'] = 44;
        b['t'] = 45; b['u'] = 46; b['v'] = 47; b['w'] = 48; b['x'] = 49;
        b['y'] = 50; b['z'] = 51; b['0'] = 52; b['1'] = 53; b['2'] = 54;
        b['3'] = 55; b['4'] = 56; b['5'] = 57; b['6'] = 58; b['7'] = 59;
        b['8'] = 60; b['9'] = 61; b['+'] = 62; b['/'] = 63;
        // The pad byte doesn't have an encoding mapping, but
        // it's not an automatic error.
        b[pad]=-2;
    }

    /** Buffer for decoded bytes that haven't been read */
    int buf[]=new int[3];
    /** Number of buffered bytes */
    int buffered=0;

    /** Buffer for encoded bytes */
    byte ebuf[]=new byte[4];
    
    /** if <code>true</code> this is based on a textfile and \r\n will be translated to \n */
    boolean textfile;
    
    /**
    Constructs a new Base64InputStream initialized with the specified input stream.
    @param in the input stream
    */
    public Base64InputStream(InputStream in)
    {
        this(in,false);
    }

    /**
    Constructs a new Base64InputStream initialized with the specified input stream, for a text file.
    If the parameter textfile is <code>true</code>, a \r is immediately followed by a \n, the \r
    will be filtered out.
    @param in the input stream
    @param textfile true if the file is a text file
    */
    public Base64InputStream(InputStream in,boolean textfile)
    {
        // To make life easier, we slip a WSStripInputStream in just ahead
        // of us, so that we don't have to worry about whitespace bytes.
        super(new WSStripInputStream(in));
        this.textfile = textfile;
    }

    /**
    Reads a byte of data.  The method will block if no input is available.
    @return  the byte read, or -1 if the end of the stream is reached.
    @throws IOException if an I/O error has occurred.
    */
    public synchronized int read() throws IOException
    {
        if (buffered == 0)
            fill_buffer();
        int b=buf[--buffered];
        if (textfile && b=='\r' && peek()=='\n')
            return read();
        return b;
    }
           
    /**
    Returns the next byte which will be read.  The method will block if no input is available.
    @return  the next byte to be read, or -1 if the end of the stream is reached.
    @throws IOException If in I/O error has occurred.
    */
    public synchronized int peek() throws IOException
    {
        if (buffered == 0)
            fill_buffer();
        return buf[buffered-1];
    }
           
    /**
    Reads into an array of bytes.
    Blocks until some input is available. This method should be overridden in a subclass for 
    efficiency (the default implementation reads 1 byte at a time).
    @param b the buffer into which the data is read
    @param off the start offset of the data
    @param len the maximum number of bytes to read
    @return the actual number of bytes read or -1 when the end of the stream is reached
    @throws IOException if an I/O error has occurred.
    */
    public int read(byte b[],int off,int len) throws IOException
    {
        for (int i=off;i<(off+len);i++)
        {
            int c=read();
            if (c==-1)
                return i-off;
            b[i]=(byte)c;
        }
        return len;
    }

    /**
    Skips bytes of input.
    @param n number of bytes to be skipped
    @return  the actual number of bytes skipped
    @throws IOException if an I/O error has occurred.
    */
    public long skip(long n) throws IOException
    {
        // Can't just read n bytes from 'in' and throw them away, because
        // n bytes from 'in' will result in roughly (4n/3) bytes from 'this',
        // and we can't even calculate the exact number easily, because of
        // the potential of running into the padding at the end of the
        // encoding.  It's  easier to just read from 'this' and throw those
        // bytes away, even though it's less efficient.
        for(int i=1;i<=n;i++)
        {
            if(read()==-1)
                return i-1;
        }
        return n;
    }

    /**
    Fills buf with a new chunk of decoded data.
    @throws IOException if an I/O error has occurred.
    */
    protected void fill_buffer() throws IOException
    {
        if (buffered!=0)
            return; // Just for safety ...
        int l=in.read(ebuf);
        int numbytes=3;
        if (l==0)
        {  // Must've reached EOF last time ...
            // Fill buffer with EOF indicators for read() to return.
            for(int i=0;i<buf.length;i++)
            {
                buf[i]=-1;
                buffered++;
            }
            return;
        }
        if(l<ebuf.length)
            throw new EOFException();
        // Check for bad bytes
        for(int i=0;i<ebuf.length;i++)
        {
            if(b[ebuf[i]]==BADBYTE)
                throw new WrongFormatException("Base64: invalid byte "+(char)ebuf[i]);
            // While we're at it, take notice of padding
/*
BUGFIX: Changed b[ebuf[i]]==pad to ebuf[i]==pad
Padding occurs in the encoded text, not in the decoded!
*/
            if(ebuf[i]==pad)
            {
                if(i<2)
                    throw new WrongFormatException("Base64: padding starts too soon");
                numbytes=i-1;
            }
        }
        // Now do the decoding
/*
BUFGFIX: Looks now prettier, adheres to bit-handling
*/
        // Will only work for 3 bytes at most!
        int shiftUpper=4;
        int shiftLower=2;
        for(int i=0;i<numbytes;i++)
        {
            buf[(numbytes-1)-i]=((b[ebuf[i]]<<shiftLower)&0xff)|(b[ebuf[i+1]]>>shiftUpper);
            buffered++;
            shiftUpper-=2;
            shiftLower+=2;
        }
    }

    /**
    Test for Base64InputStream.    
    */
    public static void main(String[] args)
    {
        if(args.length!=2)
        {
            System.out.println("Usage: java Base64InputStream infile outfile");
            System.exit(1);
        }
        File infile=new File(args[0]);
        if(!infile.exists() || !infile.canRead() || !infile.isFile())
        {
            System.out.println("Cant' read input file "+infile.getPath()+"\nUsage: java Base64InputStream infile outfile");
            System.exit(1);
        }
        File outfile=new File(args[1]);
        if(outfile.exists() && (!outfile.canWrite() || !outfile.isFile()))
        {
            System.out.println("Can't write output file "+outfile.getPath()+"\nUsage: java Base64InputStream infile outfile");
            System.exit(1);
        }
        try
        {
            InputStream in=new Base64InputStream(new BufferedInputStream(new FileInputStream(infile)),true);
            OutputStream out=new BufferedOutputStream(new FileOutputStream(outfile));
            byte[] buf=new byte[128];
            int len;
            while((len=in.read(buf))>0)
                out.write(buf,0,len);
            in.close();
            out.close();
        }
        catch(FileNotFoundException e)
        {
            System.out.println(e.toString());
            System.exit(1);
        }
        catch(IOException e)
        {
            System.out.println(e.toString());
            System.exit(1);
        }
    }

}
