/*
 * @(#)Base64OutputStream.java
 *
 * BASE64 Coding
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.Coder.Base64;

import java.io.*;
import FIM.Util.Coder.WrongFormatException;

/**
An output stream which encodes a file into base64 format.

@version     1.1, 12 Jan 1997
@author      Glenn Vanderburg
@author      Michael Sonntag (Corrections)
*/
public class Base64OutputStream extends FilterOutputStream
{
    /** Buffer for bytes to encode */
    private byte buf[]=new byte[3];
    /** Number of buffered bytes */
    private int buffered=0;
    /** If it is a textfle */
    private boolean textfile;
    /** Number of chunks (3 bytes unencoded to 4 bytes encoded) done */
    int chunksdone=0;
    /** if the encoding is finished */
    boolean finished=false;

    /** Base64 encoding bytes */
/* BUGFIX: Instead of Y and y there was U and u */
    static char b[]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/".toCharArray();

    /** Padding byte */
    static byte PAD=(byte)'=';

    /**
    Creates a new Base64OutputStream.
    @param out the output stream
    */
    public Base64OutputStream(OutputStream out)
    {
        this(out,false);
    }

    /**
    Creates a new Base64OutputStream, for a text file. If the parameter textfile is <code>true</code>,
    a \n will be written as \r\n.
    @param out the output stream
    @param textfile true if the file is a text file.
    */
    public Base64OutputStream(OutputStream out,boolean textfile)
    {
        super(out);
        this.textfile=textfile;
    }

    /**
    Writes a byte. This method will block until the byte is actually written.
    @param b the byte to write
    @throws IOException if an I/O error has occurred.
    */
    public synchronized void write(int b) throws IOException
    {
        if (textfile && b=='\n')
            write('\r');
        buf[buffered++]=(byte)b;
        if (buffered==buf.length)
            empty_buffer();
    }

    /**
    Writes a subarray of bytes.
    @param b the data to be written
    @param off the start offset in the data
    @param len the number of bytes that are written
    @throws IOException if an I/O error has occurred.
    */
    public void write(byte b[],int off,int len) throws IOException
    {
    	for (int i=0;i<len;i++)
	        write(b[off+i]);
    }

    /**
    Closes the stream.
    @throws IOException if an I/O error has occurred.
    */
    public synchronized void close() throws IOException
    {
        empty_buffer();
        super.close();
    }

    /**
    Encode all buffered bytes and write the encoded bytes to the output stream.
    @throws IOException if an I/O error has occurred.
    */
    protected void empty_buffer() throws IOException
    {
        if(buffered==0) // Just for safety ...
            return;
        if(finished)  // Sorry, the padding's already on ...
            throw new WrongFormatException("Already applied padding. Cannot write more data.");
        for(int i=buffered;i<buf.length;i++)
            buf[i]=0;
        // It's not so easy to do this with nice math as it is with
        // the decoding operation.
/* BUGFIX: Shift left only 4 not 6 and added &0x3f/&0x30/&0x0f. This is needed, as byte
is enlarged to int before shifting WITH sign extension, so shifting yields a large int
on negative numbers. */
        if(buffered>0)
        {
            out.write(b[(buf[0]>>2)&0x3f]);
            out.write(b[((buf[0]<<4)&0x30) | ((buf[1]>>4)&0x0f)]);
        }
/* BUGFIX: Shift left only 2 not 4 and &0x3c/0x03, see above */
        if(buffered>1)
            out.write(b[((buf[1]<<2)&0x3c) | ((buf[2]>>6)&0x03)]);
        else
            out.write(PAD);
        if(buffered>2)
        {
            out.write(b[buf[2]&0x3f]);
            chunksdone++;
        }
        else
        {
            out.write(PAD);
            finished=true;
        }
        buffered=0;
        if ((chunksdone==18) || finished)
        {   // Line completed
            out.write('\n');
            chunksdone=0;
        }
    }

    /**
    Test for Base64OutputStream.
    */
    public static void main(String[] args)
    {
        if(args.length!=2)
        {
            System.out.println("Usage: java Base64OutputStream infile outfile");
            System.exit(1);
        }
        File infile=new File(args[0]);
        if(!infile.exists() || !infile.canRead() || !infile.isFile())
        {
            System.out.println("Cant' read input file "+infile.getPath()+"\nUsage: java Base64OutputStream infile outfile");
            System.exit(1);
        }
        File outfile=new File(args[1]);
        if(outfile.exists() && (!outfile.canWrite() || !outfile.isFile()))
        {
            System.out.println("Can't write output file "+outfile.getPath()+"\nUsage: java Base64OutputStream infile outfile");
            System.exit(1);
        }
        try
        {
            InputStream in=new BufferedInputStream(new FileInputStream(infile));
            OutputStream out=new Base64OutputStream(new BufferedOutputStream(new FileOutputStream(outfile)),true);
            byte[] buf=new byte[128];
            int len;
            while((len=in.read(buf))>0)
                out.write(buf,0,len);
            in.close();
            out.close();
        }
        catch(FileNotFoundException e)
        {
            System.out.println(e.toString());
            System.exit(1);
        }
        catch(IOException e)
        {
            System.out.println(e.toString());
            System.exit(1);
        }
    }

}
