package FIM.Util.Coder;

import java.io.*;

/**
 * An input stream which strips out all whitespace bytes (space, tab, linefeed, carriage return and formfeed).
 *
 * @version     1.1, 11 Jan 1997
 * @author      Glenn Vanderburg
 */

public class WSStripInputStream extends FilterInputStream
{
    /**
     * Constructs a new WSStripInputStream initialized with the
     * specified input stream
     * @param in the input stream
     */
    public WSStripInputStream (InputStream in)
    {
        super(in);
    }

    /**
     * Reads a byte of data.  The method will block if no input is available.
     * @return  the byte read, or -1 if the end of the stream is reached.
     * @exception IOException If an I/O error has occurred.
     */
    public int read () throws IOException
    {
        // This is the routine that really implements the special
        // functionality of this class; the others just call this
        // one to get the data that they need.
        int c;
        do
        {
            c=in.read();
        } while ((c==' ' || c=='\t' || c=='\r' || c=='\n' || c=='\f') && c!=-1);
        return c;
    }

    /**
     * Reads into an array of bytes.
     * Blocks until some input is available.
     * @param b the buffer into which the data is read
     * @param off the start offset of the data
     * @param len the maximum number of bytes read
     * @return  the actual number of bytes read, -1 is
     *          returned when the end of the stream is reached.
     * @exception IOException If an I/O error has occurred.
     */
    public int
    read (byte b[], int off, int len) throws IOException
    {
        for (int i=off; i<(off+len); i++) {
            int c = read();
            if (c == -1) {
                return i - off;
            }
            b[i] = (byte) c;
        }
        return len;
    }

    /**
     * Skips bytes of input.
     * @param n         bytes to be skipped
     * @return  actual number of bytes skipped
     * @exception IOException If an I/O error has occurred.
     */
    public long skip (long n) throws IOException
    {
        // Can't just read n bytes from 'in' and throw them
        // away, because n bytes from 'in' doesn't necessarily
        // correspond to n bytes from 'this'.
        for (int i=1;i<=n;i++)
        {
            if (read()==-1)
                return i-1;
        }
        return n;
    }

    /**
     * Returns the number of bytes that can be read without blocking.
     * @return the number of available bytes
     * @exception IOException If an I/O error has occurred.
     */
    public int available () throws IOException
    {
        // We don't really know.  We can ask 'in', but some of those bytes
        // are probably whitespace, and it's possible that all of them are.
        // So we have to be conservative and return zero.
        return 0;
    }
}
