/*
 * @(#)CopyFilterInputStream.java
 * 
 * Utility classes
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util;

import java.io.*;

/**
A stream which copies all data sent through it to another stream. The content itself is not touched in any way.
If bytes are skipped, they will NOT be copied to the output. If <code>mark</code> and <code>reset</code> is used,
the double read bytes will be copied twice.

@author  Michael Sonntag
@version 1.0, 12.9.2001
*/
public class CopyFilterInputStream extends FilterInputStream
{
    /** The stream to copy to */
    private OutputStream copy=null;
    
    /**
    Create a new stream to copy the content read through to another stream.
    @param out the underlying <code>OutputStream</code>
    */
    public CopyFilterInputStream(InputStream in, OutputStream copy)
    {
        super(in);
        if(copy==null)  
            throw new IllegalArgumentException("No copy stream provided!");
        this.copy=copy;
    }
    
    /**
    Closes the source and the copy stream.
    @throws IOException if an exception occurred in the underlying stream (either the source or the copy)
    */
    public void close() throws IOException
    {
        super.close();
        copy.close();
    }
    
    /**
    Reads the next byte of data from this input stream. The value byte is returned as an int in the range
    0 to 255. If no byte is available because the end of the stream has been reached, the value -1 is returned.
    This method blocks until input data is available, the end of the stream is detected, or an exception is thrown. 
    @return the next byte of data, or -1 if the end of the stream is reached
    @throws IOException if an exception occurred in the underlying stream (either the source or the copy)
    */
    public int read() throws IOException
    {
        int res=super.read();
        if(res!=-1)
            copy.write(res);
        return res;
    }
    
    /**
    Reads up to <code>byte.length</code> bytes of data from this input stream into an array of bytes.
    This method blocks until some input is available. 
    @param b the buffer into which the data is read
    @return the total number of bytes read into the buffer, or -1 if there is no more data because the 
            end of the stream has been reached
    @throws IOException if an exception occurred in the underlying stream (either the source or the copy)
    */
    public int read(byte[] b) throws IOException
    {
        int res=super.read(b);
        if(res!=-1)
            copy.write(b,0,res);
        return res;
    }
    
    /**
    Reads up to len bytes of data from this input stream into an array of bytes. This method blocks 
    until some input is available. 
    @param b the buffer into which the data is read
    @param off the start offset of the data
    @param len the maximum number of bytes read
    @return the total number of bytes read into the buffer, or -1 if there is no more data because the 
            end of the stream has been reached
    @throws IOException if an exception occurred in the underlying stream (either the source or the copy)
    */
    public int read(byte[] b,int off,int len) throws IOException
    {
        int res=super.read(b,off,len);
        if(res!=-1)
            copy.write(b,off,res);
        return res;
    }                

}