/*
 * @(#)CopyFilterOutputStream.java
 * 
 * Utility classes
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util;

import java.io.*;

/**
A stream which copies all data sent through it to another stream. The content itself is not touched in any way.

@author  Michael Sonntag
@version 1.0, 12.9.2001
*/
public class CopyFilterOutputStream extends FilterOutputStream
{
    /** The stream to copy to */
    private OutputStream copy=null;
    
    /**
    Create a new stream to copy the content written through to another stream.
    @param out the underlying <code>OutputStream</code>
    */
    public CopyFilterOutputStream(OutputStream out, OutputStream copy)
    {
        super(out);
        if(copy==null)  
            throw new IllegalArgumentException("No copy stream provided!");
        this.copy=copy;
    }
    
    /**
    Closes the destination and the copy stream.
    @throws IOException if an exception occurred in the underlying stream (either the destination or the copy)
    */
    public void close() throws IOException
    {
        super.close();
        copy.close();
    }
    
    /**
    Flushes the destination and the copy stream.
    @throws IOException if an exception occurred in the underlying stream (either the destination or the copy)
    */
    public void flush() throws IOException
    {
        super.flush();
        copy.flush();
    }
    
    /**
    Writes len bytes from the specified byte array starting at offset off to the output and the copy stream. 
    @param b the data
    @param off the start offset in the data
    @param len the number of bytes to write
    @throws IOException if an exception occurred in the underlying stream (either the destination or the copy)
    */
    public void write(byte[] b,int off,int len) throws IOException
    {
        super.write(b,off,len);
        copy.write(b,off,len);
    }
    
    /**
    Writes the byte to the output and the copy stream. 
    @param b the byte to be written
    @throws IOException if an exception occurred in the underlying stream (either the destination or the copy)
    */
    public void write(int b) throws IOException
    {
        super.write(b);
        copy.write(b);
    }
    
    /**
    Writes <code>b.length</code> bytes from the specified byte array to the output and the copy stream. 
    @param b the data
    @throws IOException if an exception occurred in the underlying stream (either the destination or the copy)
    */
    public void write(byte[] b) throws IOException
    {
        super.write(b);
        copy.write(b);
    }
}