/*
 * @(#)JarUtils.java
 *
 * Part of the AgentSystem "POND"
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.Crypto;

import java.util.Enumeration;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.File;
import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.BufferedInputStream;
import java.io.ObjectInputStream;
import java.io.FileOutputStream;
import java.io.FileInputStream;
import java.io.InputStream;
import java.util.zip.ZipEntry;
import java.util.jar.JarFile;
import java.util.jar.Attributes;
import java.util.jar.Manifest;
import java.util.jar.JarOutputStream;
import java.util.jar.JarEntry;
import java.security.PrivateKey;
import java.security.SignatureException;
import java.security.KeyFactory;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidKeyException;
import java.security.DigestException;
import java.security.spec.PKCS8EncodedKeySpec;
import FIM.Util.Coder.Base64.Base64Coder;

/**
A class with utilities for JAR-files. Contains the following:
<ul>
<li>Digesting of jar-files</li>
<li>Signing of jar-files</li>
<li>Creating a digested (MD5, SHA-1) and signed jar-file.
</ul>
All procedures require that an jar-file already exists.
<p>This class can also be used directly to sign jar-files: <code>java FIM.Util.JarUtils JAR-file CertificateAndKeyFile NameOfSigner</code>
</p>
<p>ATTENTION: This uses undocumented classes from Sun. Might not work in future releases of JDK 1.2.<br>
I couldn't get IAIK to read the SignedData structure of the actual signature files (*.DSA); 
always dropped with "ParsingException: Couldn't parse object" (Wonderful message!!!!).</p>

@author  Michael Sonntag
@version 1.0, 1.7.2000
*/
public class JarUtils
{
    
    /**
    Signs a jar-file.
    @param filename the jar-file to sign
    @param signerName the name of the signer (=entry name for signature file in directory META-INF)
    @param chain the certificate chain for the signer
    @param privateKey the private key of the signer
    @throws SignatureException if an error occured during signing
    @throws FileNotFoundException if the jar-file could not be found
    */
    public static void signJarFile(String filename,String signerName,java.security.cert.X509Certificate[] chain,PrivateKey privateKey) throws FileNotFoundException,SignatureException
    {
        // Check parameters
        if(filename==null || filename.length()<1)
            throw new IllegalArgumentException("Filename must be provided");
        if(signerName==null || signerName.length()<1)
            throw new IllegalArgumentException("Name of the signer must be provided");
        if(chain==null || chain.length<1)
            throw new IllegalArgumentException("Chain of certificates signer must be provided (minimum length: 1)");           
        if(privateKey==null)
            throw new IllegalArgumentException("Private key of the signer must be provided");           
        JarFile file=null;
        File tmp=null;
        Manifest man=null;
        JarOutputStream os=null;
        ByteArrayOutputStream bos=null;
        sun.security.util.SignatureFile signature=null;
        sun.security.util.SignatureFile.Block signedBlock=null;
        // Read manifest
        try
        {
            file=new JarFile(filename);
            man=(Manifest)file.getManifest();
            bos=new ByteArrayOutputStream();
            man.write(bos);
            bos.close();
            // Create temporary file
            tmp=File.createTempFile("tmp_",".jar");
            os=new JarOutputStream(new FileOutputStream(tmp));        
        }
        catch(FileNotFoundException e)
        {
            throw e;
        }
        catch(IOException e)
        {
	        throw new SignatureException("Error reading jar file: "+e.toString());
        }
        try
        {
            MessageDigest[] digs={MessageDigest.getInstance("SHA1")};
            sun.security.util.ManifestDigester mandig=new sun.security.util.ManifestDigester(bos.toByteArray());
            signature=new sun.security.util.SignatureFile(digs,man,mandig,signerName,true);
            signedBlock=signature.generateBlock(privateKey,chain,true);
        }
        catch(IOException e)
        {
	        throw new SignatureException("Error signing jar file (during generating actual signature): "+e.toString());
        }
        catch(NoSuchAlgorithmException e)
        {
	        throw new SignatureException("Error signing jar file (algorithm for hashing/signing not found): "+e.toString());
        }
        catch(InvalidKeyException e)
        {
	        throw new SignatureException("Error signing jar file (Invalid key): "+e.toString());
        }
        try
        {           
            // Now create a new jar file with manifest and signature files at the beginning
            // Write manifest
	        JarEntry entry=new JarEntry(JarFile.MANIFEST_NAME);
    	    os.putNextEntry(entry);
	        man.write(os);
	        os.closeEntry();
	        // Write Signature file
	        entry=new JarEntry(signature.getMetaName());
    	    os.putNextEntry(entry);
    	    signature.write(os);
	        os.closeEntry();
	        // Write the signed signature file (including the certificate)
	        entry=new JarEntry(signedBlock.getMetaName());
    	    os.putNextEntry(entry);
    	    signedBlock.write(os);
	        os.closeEntry();
	    }
	    catch(IOException e)
	    {
	        throw new SignatureException("Error adding meta-entries to jar file: "+e.toString());
	    }
	    try
	    {
	        // Add the original files
            Enumeration entries=file.entries();
            while(entries.hasMoreElements())
            {
                JarEntry entry=(JarEntry)entries.nextElement();
                if(!entry.getName().equalsIgnoreCase(JarFile.MANIFEST_NAME) && !entry.isDirectory() &&
                !entry.getName().equalsIgnoreCase(signature.getMetaName()) && !entry.getName().equalsIgnoreCase(signedBlock.getMetaName()))
                {   // Ignore manifest and signature files
      	            os.putNextEntry(entry);
                    byte[] buf=new byte[1024];
                    int len=1024;
                    InputStream in=file.getInputStream(entry);
                    while((len=in.read(buf))>0)
                        os.write(buf,0,len);
         	        in.close();
           	        os.closeEntry();
	            }
    	    }
            os.close();
            file.close();
	    }
	    catch(IOException e)
	    {
	        throw new SignatureException("Error adding entries to jar file: "+e.toString());
	    }
	    // Copy from temporary over original
	    File jarFile=new File(file.getName());
	    try
	    {
    	    BufferedOutputStream out=new BufferedOutputStream(new FileOutputStream(jarFile));
	        BufferedInputStream in=new BufferedInputStream(new FileInputStream(tmp));
	        byte[] buf=new byte[1024];
    	    int len=1024;
	        while((len=in.read(buf))>0)
	            out.write(buf,0,len);
    	    in.close();
            out.close();
        }
        catch(IOException e)
        {
	        throw new SignatureException("Error copying temporary file over original: "+e.toString());
        }
        // Delete temporary file
	    tmp.delete();
    }

    
    /**
    Digests a jar-file.
    @param filename the jar-file to digest
    @param algorithm the algorithm used for digesting
    @throws DigestException if an error occured during digesting
    @throws FileNotFoundException if the jar-file could not be found
    */
    public static void digestJarFile(String filename,String algorithm) throws FileNotFoundException, DigestException
    {
        JarFile file=null;
        File tmp=null;
        Manifest man=null;
        JarOutputStream os=null;
        // Get manifest
        try
        {
            file=new JarFile(filename);
            man=file.getManifest();
            if(man!=null)
                man=(Manifest)man.clone();
            else
                man=new Manifest();
            // Create temporary file
            tmp=File.createTempFile("tmp_",".jar");
            os=new JarOutputStream(new FileOutputStream(tmp));        
        }
        catch(FileNotFoundException e)
        {
            throw e;
        }
        catch(IOException e)
        {
            throw new DigestException("Error reading jar file: "+e.toString());
        }
        Enumeration entries=file.entries();
        while(entries.hasMoreElements())
        {
            JarEntry entry=(JarEntry)entries.nextElement();
            if(!entry.getName().toLowerCase().startsWith("meta-inf/"))
            {   // Ignore manifest, signatures, ...
                InputStream in=null;
                try
                {
                    file.getInputStream(entry);                    
                    Attributes attr=man.getAttributes(entry.getName());
                    if(attr==null)
                        attr=new Attributes();
                    // Create digest
                    MessageDigest md=null;
                    try
                    {
                        md=MessageDigest.getInstance(algorithm);
                    }
                    catch(NoSuchAlgorithmException ex)
                    {
                        throw new DigestException("Couldn't get message digest: "+ex.toString());
                    }
                    in=file.getInputStream(entry);
                    byte[] buf=new byte[1024];
                    int len=1024;
                    while((len=in.read(buf))>0)
                        md.update(buf,0,len);
                    byte[] myDigest=md.digest();
                    String digest=Base64Coder.base64EncodeAsString(myDigest);
                    String digestAlgos=(String)attr.get(new Attributes.Name("Digest-Algorithms"));
                    if(digestAlgos==null)
                    {   // Not yet any digests
                        attr.put(new Attributes.Name("Digest-Algorithms"),algorithm);
                    }
                    else
                    {   // Add to list if necessary
                        if(digestAlgos.toLowerCase().indexOf(algorithm.toLowerCase())<0)
                        {
                            digestAlgos+=", "+algorithm;
                            attr.put(new Attributes.Name("Digest-Algorithms"),digestAlgos);
                        }
                    }
                    attr.put(new Attributes.Name(algorithm+"-Digest"),digest);
                    man.getEntries().put(entry.getName(),attr);
                }
                catch(IOException e)
                {   // Should never happen
                    System.out.println(e.toString());
                }
            }
        }        
        // Create temporary file for manifest
	    ZipEntry e=new ZipEntry(JarFile.MANIFEST_NAME);
	    try
	    {
    	    os.putNextEntry(e);
	        man.write(os);
	        os.closeEntry();
            entries=file.entries();
            while(entries.hasMoreElements())
            {
                JarEntry entry=(JarEntry)entries.nextElement();
                if(!entry.getName().equalsIgnoreCase(JarFile.MANIFEST_NAME) && !entry.isDirectory())
                {   // Ignore manifest
      	            os.putNextEntry(entry);
                    byte[] buf=new byte[1024];
                    int len=1024;
                    InputStream in=file.getInputStream(entry);
                    while((len=in.read(buf))>0)
                        os.write(buf,0,len);
         	        in.close();
           	        os.closeEntry();
	            }
    	    }
            os.close();
            file.close();
	    }
	    catch(IOException ex)
	    {
            throw new DigestException("Error creating new jar file: "+ex.toString());
	    }
        // We must read the manifest and store it as we read it, so it won't be changed on
        // subsequent reads (as it will happen in signature). Directly after creating the stored
        // manifest differs from reading and again writing it (lines swapped).
        try
        {
            file=new JarFile(filename);
            File tmp2=File.createTempFile("tmp_",".jar");
            JarOutputStream jos=new JarOutputStream(new FileOutputStream(tmp2));
            Manifest newMan=(new JarFile(tmp.getPath())).getManifest();
    	    e=new ZipEntry(JarFile.MANIFEST_NAME);
    	    jos.putNextEntry(e);
	        newMan.write(jos);
	        jos.closeEntry();
            entries=file.entries();
            while(entries.hasMoreElements())
            {
                JarEntry entry=(JarEntry)entries.nextElement();
                if(!entry.getName().equalsIgnoreCase(JarFile.MANIFEST_NAME) && !entry.isDirectory())
                {   // Ignore manifest
      	            jos.putNextEntry(entry);
                    byte[] buf=new byte[1024];
                    int len=1024;
                    InputStream in=file.getInputStream(entry);
                    while((len=in.read(buf))>0)
                        jos.write(buf,0,len);
         	        in.close();
           	        jos.closeEntry();
	            }
    	    }
            jos.close();
            file.close();
            File del=tmp;
            tmp=tmp2;
            del.delete();
        }
        catch(IOException ex)
        {
            throw new DigestException("Error creating file with new manifest: "+e.toString());
        }
        // Copy file to original
	    File jarFile=new File(filename);
	    try
	    {
    	    BufferedOutputStream out=new BufferedOutputStream(new FileOutputStream(jarFile));
	        BufferedInputStream in=new BufferedInputStream(new FileInputStream(tmp));
	        byte[] buf=new byte[1024];
    	    int len=1024;
	        while((len=in.read(buf))>0)
	            out.write(buf,0,len);
    	    in.close();
            out.close();
        }
        catch(IOException ex)
        {
            throw new DigestException("Error copying temporary jar file: "+e.toString());
        }
        // Delete temporary file
	    tmp.delete();
    }
    
    /**
    Digests a jar-file (MD5 and SHA-1) and signs it afterwards.
    @param filename the jar-file to sign
    @param signerName the name of the signer (=entry name for signature file in directory META-INF)
    @param chain the certificate chain for the signer
    @param privateKey the private key of the signer
    @throws DigestException if an error occured during digesting
    @throws SignatureException if an error occured during signing
    @throws FileNotFoundException if the jar-file could not be found
    */
    public static void createSignedJar(String filename,String signerName,java.security.cert.X509Certificate[] chain,PrivateKey privateKey) throws SignatureException,DigestException,FileNotFoundException
    {
        digestJarFile(filename,"MD5");
        digestJarFile(filename,"SHA1");
        signJarFile(filename,signerName,chain,privateKey);
    }
    
    /**
    This class can also be used directly to sign jar-files:<br>
    <code>java FIM.Util.JarUtils JAR-file CertificateAndKeyFile NameOfSigner</code>
    */
    public static void main(String[] args)
    {
        System.out.println("JarUtils: Digest and sign jar-files\nby Michael Sonntag (using proprietary classes from Sun)\n");
        if(args.length!=3)
        {
            System.out.println("Wrong number of parameters!\nUsage: java FIM.Util.JarUtils JAR-file CertificateAndKeyFile NameOfSigner\n");
            System.exit(1);
        }
        PrivateKey priv=null;
        java.security.cert.X509Certificate[] chain=new java.security.cert.X509Certificate[1];
        // Read key and certificate
        try
		{                            
            ObjectInputStream ois=new ObjectInputStream(new FileInputStream(args[1]));
            String keyType=(String)ois.readObject();
            byte[] buf=(byte[])ois.readObject();
 	        PKCS8EncodedKeySpec keySpec=new PKCS8EncodedKeySpec(buf);
    		KeyFactory keyFactory=KeyFactory.getInstance(keyType);
	    	priv=keyFactory.generatePrivate(keySpec);
            java.security.cert.CertificateFactory certFact=java.security.cert.CertificateFactory.getInstance("X.509");
            java.security.cert.Certificate cert=certFact.generateCertificate(ois);
            chain[0]=(java.security.cert.X509Certificate)cert;
            System.out.println("Certificate and private key successfully loaded.");
        }
        catch(Exception e)
        {
            System.err.println("Could not restore the private key or certificate: "+e.toString());
            return;
		}
		// Digest and sign
		try
		{
		    System.out.println("Digesting MD5...");
            digestJarFile(args[0],"MD5");
		    System.out.println("Digesting SHA-1...");
            digestJarFile(args[0],"SHA1");
		    System.out.println("Signing...");
            signJarFile(args[0],args[2],chain,priv);
            System.out.println("File successfully signed!");
        }
        catch(Exception e)
        {
            System.err.println("Error digesting/signing the jar file: "+e.toString());
        }
    }
 }
 