/*
 * @(#)NamedKeyAndCertificate.java
 *
 * Cryptographic utilities
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.Crypto;

import java.security.KeyPair;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.io.IOException;
import java.io.Serializable;
import java.io.ObjectOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.ObjectInputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayInputStream;
import java.security.PrivateKey;
import java.security.cert.Certificate;
import java.security.cert.CertificateFactory;
import java.security.cert.CertificateException;
import java.security.cert.CertificateEncodingException;
import java.security.KeyFactory;
import java.security.spec.X509EncodedKeySpec;
import java.security.NoSuchAlgorithmException;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.PKCS8EncodedKeySpec;
import FIM.Util.ClassLoaderObjectInputStream;

/**
A certificate and the corresponding private key is associated to a name. Can be normally serialized 
(certificate and the key will be stored and retrieved through a factory).

@author  Michael Sonntag
@version 1.0, 1.7.2000
*/
public final class NamedKeyAndCertificate implements Serializable
{
	/**
	The name of the certificate and the private key.
	@serial
    */
    private String name;
	/**
	The private key.
	@serial
    */
    private  PrivateKey priv=null;
	/**
	The certificate.
	@serial
    */
    private Certificate cert;
    
    /**
    Creates a new private key and certificate associated to a name. The name
    and a certificate must be provided, the private key is optional.
    @param name the name of the certificate and the private key
    @param priv the private key
    @param cert the certificate
    */
    public NamedKeyAndCertificate(String name,PrivateKey priv,Certificate cert)
    {
        if(name==null || name.length()<1)
            throw new IllegalArgumentException("Name must be provided!");
        if(cert==null)
            throw new IllegalArgumentException("Certificate must be provided!");
        this.name=name;
        this.priv=priv;
        this.cert=cert;
    }
    
    /**
    Retrieve the name of this keypair.
    @return the name of the keypair
    */
    public String getName()
    {
        return name;
    }
    
    /**
    Retrieve the private key.
    @return the private key (might be null if none contained)
    */
    public PrivateKey getPrivateKey()
    {
        return priv;
    }
    
    /**
    Retrieve the certificate.
    @return the certificate
    */
    public Certificate getCertificate()
    {
        return cert;
    }
    
    /**
    Writes the name, the private key and the certificate to the stream using a factory.
    @param outs the stream to write to
    @throws IOException if an IO error occurs during serialization
    */
    private void writeObject(ObjectOutputStream outs) throws IOException
    {
        /**
        Must store the objects in a byte array, as we must store it in a separate
        ObjectOutputStream. This is necessary to allow us loading the information
        through a ClassLoaderObjectInputStream.
        */
        ByteArrayOutputStream bos=new ByteArrayOutputStream();
        ObjectOutputStream out=new ObjectOutputStream(bos);
		try
		{
            out.writeObject(name);
			out.writeObject(cert.getType());
			out.writeObject(cert.getEncoded());
            out.writeBoolean(priv!=null);
            if(priv!=null)            
            {
        		out.writeObject(priv.getAlgorithm());
	        	KeyFactory keyFactory=KeyFactory.getInstance(priv.getAlgorithm());
		        PKCS8EncodedKeySpec spec=(PKCS8EncodedKeySpec)keyFactory.getKeySpec(priv,PKCS8EncodedKeySpec.class);
		        out.writeObject(spec.getEncoded());
		    }
		}
		catch(InvalidKeySpecException e)
		{
		    throw new IOException("Error storing private key: "+e.toString());
		}
		catch(NoSuchAlgorithmException e)
		{
		    throw new IOException("Error storing private key: "+e.toString());
		}
		catch(CertificateEncodingException e)
		{
		    throw new IOException("Error storing certificate: "+e.toString());
		}
		finally
		{
		    out.close();
		}
		outs.writeObject(bos.toByteArray());
    }
    
    /**
    Reads the name, the private key and the certificate from the stream using a factory. Loads 
    the classes through the same classloader as the ObjectInputStream was loaded from.
    @param ins the stream to read from
    @throws IOException if an IO error occurs during serialization
    @throws ClassNotFoundException if a class could not be restored
    */
    private void readObject(ObjectInputStream ins) throws IOException, ClassNotFoundException
    {
        // Recreate stream
        byte[] buffer=(byte[])ins.readObject();
        ByteArrayInputStream bis=new ByteArrayInputStream(buffer);
        final ObjectInputStream instream=ins;
        final ClassLoaderObjectInputStream in=new ClassLoaderObjectInputStream(bis);
        // Called from within AgentBase.readObject, which is loaded through the AgentClassLoader, so privileged
        AccessController.doPrivileged(new PrivilegedAction()
        {
            public Object run()
            {
                in.setClassLoader(instream.getClass().getClassLoader());
                return null;
            }
        }
        );
        // Read name
        name=(String)in.readObject();
		try
		{
    	    // Read certificate
			String type=(String)in.readObject();
			byte[] buf=(byte[])in.readObject();
			CertificateFactory cf=CertificateFactory.getInstance(type);
			ByteArrayInputStream bais=new ByteArrayInputStream(buf);
		    cert=cf.generateCertificate(bais);
		    if(in.readBoolean())
		    {
    		    // Read private key
        		type=(String)in.readObject();
	        	buf=(byte[])in.readObject();
 	            PKCS8EncodedKeySpec keySpec=new PKCS8EncodedKeySpec(buf);
    		    KeyFactory keyFactory=KeyFactory.getInstance(type);
    	    	priv=keyFactory.generatePrivate(keySpec);
    	    }
 		}
		catch(InvalidKeySpecException e)
		{
		    throw new IOException("Error restoring private key: "+e.toString());
		}
		catch(NoSuchAlgorithmException e)
		{
		    throw new ClassNotFoundException("Error restoring private key: "+e.toString());
		}
 		catch(CertificateException e)
		{
		    throw new IOException("Error restoring certificate: "+e.toString());
		}
  }
    
}