/*
 * @(#)NamedKeyPair.java
 *
 * Cryptographic utilities
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.Crypto;

import java.security.KeyPair;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.io.IOException;
import java.io.Serializable;
import java.io.ObjectOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.ObjectInputStream;
import java.io.ByteArrayInputStream;
import java.security.PublicKey;
import java.security.PrivateKey;
import java.security.KeyFactory;
import java.security.spec.X509EncodedKeySpec;
import java.security.NoSuchAlgorithmException;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.PKCS8EncodedKeySpec;
import FIM.Util.ClassLoaderObjectInputStream;

/**
A <code>KeyPair</code> which is associated to a name. Can be normally serialized (the keys 
will be stored and retrieved through a factory).

@author  Michael Sonntag
@version 1.0, 1.7.2000
*/
public final class NamedKeyPair implements Serializable
{
	/**
	The name of this keypair.
	@serial
    */
    private String name;
	/**
	The keypair itself.
	@serial
    */
    private  KeyPair kp;
    
    /**
    Creates a new keypair associated to a name. The private key may be empty but a public key
    must be present at least.
    @param name the name of the keypair
    @param kp the keypair
    */
    public NamedKeyPair(String name,KeyPair kp)
    {   // Private key may be null
        if(name==null || name.length()<1)
            throw new IllegalArgumentException("Name must be supplied!");
        if(kp==null || kp.getPublic()==null)
            throw new IllegalArgumentException("Public key must be supplied!");
        this.name=name;
        this.kp=kp;
    }
    
    /**
    Retrieve the name of this keypair.
    @return the name of the keypair
    */
    public String getName()
    {
        return name;
    }
    
    /**
    Retrieve the keypair itself.
    @return the keypair
    */
    public KeyPair getKeyPair()
    {
        return kp;
    }
    
    /**
    Writes the name and the keys to the stream using a factory.
    @param outs the stream to write to
    @throws IOException if an IO error occurs during serialization
    */
    private void writeObject(ObjectOutputStream outs) throws IOException
    {
        /**
        Must store the objects in a byte array, as we must store it in a separate
        ObjectOutputStream. This is necessary to allow us loading the information
        through a ClassLoaderObjectInputStream.
        */
        ByteArrayOutputStream bos=new ByteArrayOutputStream();
        ObjectOutputStream out=new ObjectOutputStream(bos);
        try
        {
            // Write name
            out.writeObject(name);
	    	try
		    {
		        // Write public key
                PublicKey pub=kp.getPublic();
        		out.writeObject(pub.getAlgorithm());
        		KeyFactory keyFactory=KeyFactory.getInstance(pub.getAlgorithm());
	        	X509EncodedKeySpec spec=(X509EncodedKeySpec)keyFactory.getKeySpec(pub,X509EncodedKeySpec.class);
		        byte[] buf=spec.getEncoded();
    		    out.writeObject(buf);
	    	}
		    catch(InvalidKeySpecException e)
    		{
	    	    throw new IOException("Error storing public key: "+e.toString());
		    }
    		catch(NoSuchAlgorithmException e)
	    	{
		        throw new IOException("Error storing public key: "+e.toString());
    		}
            PrivateKey priv=kp.getPrivate();
            out.writeBoolean(priv!=null);
    		if(priv!=null)
    		{
        		// Write private key
	        	try
		        {
        	    	out.writeObject(priv.getAlgorithm());
	        	    KeyFactory keyFactory=KeyFactory.getInstance(priv.getAlgorithm());
    		        PKCS8EncodedKeySpec spec=(PKCS8EncodedKeySpec)keyFactory.getKeySpec(priv,PKCS8EncodedKeySpec.class);
    	    	    out.writeObject(spec.getEncoded());
	    	    }
    		    catch(InvalidKeySpecException e)
        		{   
	        	    throw new IOException("Error storing private key: "+e.toString());
    		    }
    		    catch(NoSuchAlgorithmException e)
	        	{
		            throw new IOException("Error storing private key: "+e.toString());
		        
		        }
		    }
		}
		finally
		{
		    out.close();
		}
		outs.writeObject(bos.toByteArray());
    }
    
    /**
    Reads the name and the keys from the stream using a factory. Loads the classes through the same
    classloader as the ObjectInputStream was loaded from.
    @param ins the stream to read from
    @throws IOException if an IO error occurs during serialization
    @throws ClassNotFoundException if a class could not be restored
    */
    private void readObject(ObjectInputStream ins) throws IOException, ClassNotFoundException
    {
        // Recreate stream
        byte[] buffer=(byte[])ins.readObject();
        ByteArrayInputStream bis=new ByteArrayInputStream(buffer);
        final ObjectInputStream instream=ins;
        final ClassLoaderObjectInputStream in=new ClassLoaderObjectInputStream(bis);
        // Called from within AgentBase.readObject, which is loaded through the AgentClassLoader, so privileged
        AccessController.doPrivileged(new PrivilegedAction()
        {
            public Object run()
            {
                in.setClassLoader(instream.getClass().getClassLoader());
                return null;
            }
        }
        );
        // Read name
        name=(String)in.readObject();
        // Read public key
        PublicKey pub=null;
        try
        {
    		String keyType=(String)in.readObject();
	    	byte[] buf=(byte[])in.readObject();
 	        X509EncodedKeySpec keySpec=new X509EncodedKeySpec(buf);
    		KeyFactory keyFactory=KeyFactory.getInstance(keyType);
	    	pub=keyFactory.generatePublic(keySpec);
		}
		catch(InvalidKeySpecException e)
		{
		    throw new IOException("Error restoring public key: "+e.toString());
		}
		catch(NoSuchAlgorithmException e)
		{
		    throw new ClassNotFoundException("Error restoring public key: "+e.toString());
		}
        PrivateKey priv=null;
		if(in.readBoolean())
		{   // Read private key
    		try
	    	{
    	    	String keyType=(String)in.readObject();
	    	    byte[] buf=(byte[])in.readObject();
     	        PKCS8EncodedKeySpec keySpec=new PKCS8EncodedKeySpec(buf);
        		KeyFactory keyFactory=KeyFactory.getInstance(keyType);
	        	priv=keyFactory.generatePrivate(keySpec);
     		}
	    	catch(InvalidKeySpecException e)
		    {
		        throw new IOException("Error restoring private key: "+e.toString());
    		}
	    	catch(NoSuchAlgorithmException e)
		    {
		        throw new ClassNotFoundException("Error restoring private key: "+e.toString());
    		}
    	}
    	// Create keypair
        kp=new KeyPair(pub,priv);
   }
    
}