/*
 * @(#)SealedAndSignedObject.java
 *
 * Part of the AgentSystem "POND"
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.Crypto;

import java.io.Serializable;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.ByteArrayInputStream;
import java.security.KeyStore;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.security.cert.CertificateParsingException;
import java.security.cert.CertificateExpiredException;
import java.security.cert.CertificateNotYetValidException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.InvalidKeyException;
import java.security.SignatureException;
import java.security.SignedObject;
import java.security.Signature;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.KeyStoreException;
import java.security.Principal;
import javax.crypto.Cipher;
import javax.crypto.SealedObject;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.BadPaddingException;
import java.util.Enumeration;

/**
An object to contain another object which is first signed and afterwards encrypted. The data-object
must be serializable.
<br>Based on SignedObject and SealedObject.

@author Michael Sonntag
@version 1.0, 1.7.2000
@see java.security.SignedObject
@see javax.crypto.SealedObject
*/
public class SealedAndSignedObject implements Serializable
{
	/**
	The object containing the encrypted (and previously signed) data-object.
	@serial
    */
	SealedObject encrypted=null;

	/**
	The algorithm used for encryption.
	@serial
    */
	String cryptAlgorithm="";

	/**
	Additional data to tell the recipient e. g. which public key was used for encryption.
	@serial
    */
	Serializable cryptData=null;

    /**
    Private class to encapsulate a SignedObject and the certificate chain.
    */
	private class SignedContent implements Serializable
	{
	    /** The contained <code>SignedObject</code> containing the actual data-object. */	    
		public SignedObject signedObject=null;
		/** The chain of certificates for the signature of the signed object */
		public transient Certificate[] certificateChain=null;

        /**
        Writes the object to a stream.
        @param out the stream to write to
        @throws IOException if an IO error occured during writing to the stream
        */
		private void writeObject(ObjectOutputStream out) throws IOException
		{
			out.defaultWriteObject();
			// Write certificate chain
			try
			{
				out.writeInt(certificateChain.length);
				for(int i=0;i<certificateChain.length;i++)
				{
    			    out.writeUTF(certificateChain[i].getType());
					out.writeObject(certificateChain[i].getEncoded());
				}
			}
			catch(CertificateEncodingException e)
			{
				throw new IOException("Can't encode certificate chain: "+e.toString());
			}
		}

        /**
        Reads the object from a stream.
        @param in the stream to read from
        @throws IOException if an IO error occured during reading from the stream
        @throws ClassNotFoundException if a class could not be found (especially from the contained data-object)
        @throws CertificateException if an error occured during restoring the certificate chain
        */
		private void readObject(ObjectInputStream in) throws IOException,ClassNotFoundException,CertificateException
		{
			in.defaultReadObject();
			int len=in.readInt();
			certificateChain=new Certificate[len];
			for(int i=0;i<len;i++)
			{
				String type=in.readUTF();
        	    CertificateFactory cf=CertificateFactory.getInstance(type);
				byte buf[]=(byte[])in.readObject();
    			ByteArrayInputStream bais=new ByteArrayInputStream(buf);
	    	    certificateChain[i]=cf.generateCertificate(bais);
			}
		}
	}

    /**
    Creates a new encrypted and signed object with a certain data-object.
    @param c the ciher to use for encrypting the object
    @param cryptAlgo the algorithm to be used for encryption
    @param cryptData the additional data as information for the recipient
    @param s the signature to use for signing the object
    @param pk the private key to be used for signing
    @param certs the certificate chain for the public key to be used for verification of the signature
    @param content the data-object to be signed and encrypted
    @throws SignatureException if an error occured during singing the content
    @throws InvalidKeyException if the key could not be used for signing
    @throws IOException If an error occured during serialization of the data object or the signed content
    @throws IllegalBlockSizeException if an error occured during encrypting the signed content
    */
    public SealedAndSignedObject(Cipher c, String cryptAlgo, Serializable cryptData,Signature s, PrivateKey pk, Certificate[] certs, Serializable content)
    throws SignatureException, InvalidKeyException, IOException, IllegalBlockSizeException
    {
    	if(certs==null || certs.length<1 || content==null)
    		throw new IllegalArgumentException();
    	SignedContent sc=new SignedContent();
    	sc.certificateChain=certs;
    	sc.signedObject=new SignedObject(content,pk,s);
    	encrypted=new SealedObject(sc,c);
    	cryptAlgorithm=cryptAlgo;
    	this.cryptData=cryptData;
    }

    /**
    Retrieve the algorithm used for encrypting the data.
    @return the encryption algorithm
    */
    public String getEncryptionAlgorithm()
    {
        return cryptAlgorithm;
    }
    
    /**
    Retrieve the additional data stored with encrypted object. E. g. which public key was used.
    @return the additional data
    */
    public Serializable getEncryptionData()
    {
        return cryptData;
    }

    /**
    Retrieve the algorithm used for signing the data. This is only accessible after decryption,
    so a initialized cipher must be provided.
    @param c the <code>Cipher</code> for decrypting the data
    @return the signature algorithm
    @throws ClassNotFoundException if the object could not be decrypted because a class was not found
    @throws IOException IO error decrypting the object
    @throws BadPaddingException  padding error decrypting the object
    @throws IllegalBlockSizeException block size error decrypting the object
    */
    public String getSignAlgorithm(Cipher c) throws ClassNotFoundException, IOException, BadPaddingException, IllegalBlockSizeException
    {
        SignedContent sig=(SignedContent)encrypted.getObject(c);
        return sig.signedObject.getAlgorithm();
    }

    /**
    Retrieve the encrypted data-object. A initialized cipher must be provided.
    @param c the <code>Cipher</code> for decrypting the data
    @return the stored object
    @throws ClassNotFoundException if the object could not be decrypted because a class was not found
    @throws IOException IO error decrypting the object
    @throws BadPaddingException  padding error decrypting the object
    @throws IllegalBlockSizeException block size error decrypting the object
    */
    public Object getObject(Cipher c) throws ClassNotFoundException, IOException, BadPaddingException, IllegalBlockSizeException
    { 
        SignedContent sig=(SignedContent)encrypted.getObject(c);
        return sig.signedObject.getObject();
    }

    /**
    Retrieve the certificates associated with the signature. They are only accessible after decryption,
    so a initialized cipher must be provided.
    @param c the <code>Cipher</code> for decrypting the data
    @return the certificate chain of the signature
    @throws ClassNotFoundException if the object could not be decrypted because a class was not found
    @throws IOException IO error decrypting the object
    @throws BadPaddingException  padding error decrypting the object
    @throws IllegalBlockSizeException block size error decrypting the object
    */
    public Certificate[] getCertificates(Cipher c) throws ClassNotFoundException, IOException, BadPaddingException, IllegalBlockSizeException
    {
        SignedContent sig=(SignedContent)encrypted.getObject(c);
        return sig.certificateChain;
    }

    /**
    Verifies the signature of the object using a keystore. To get access, it must be decrypted so a initialized
    cipher must be provided.
    @param c the cipher used for decrypting the object
    @param ks the key store used for verification
    @throws ClassNotFoundException if the object could not be decrypted because a class was not found
    @throws IOException IO error decrypting the object
    @throws BadPaddingException  padding error decrypting the object
    @throws IllegalBlockSizeException block size error decrypting the object
    @throws SignatureException if signature verification failed
    @throws InvalidKeyException if the verification key is invalid
    @throws NoSuchAlgorithmException if the algorihtm used for signing is not available
    */
    public boolean verify(Cipher c,KeyStore ks) throws ClassNotFoundException, IOException, BadPaddingException, IllegalBlockSizeException, InvalidKeyException, SignatureException, NoSuchAlgorithmException
    {
        SignedContent sig=(SignedContent)encrypted.getObject(c);
        Signature s=Signature.getInstance(sig.signedObject.getAlgorithm());
        java.security.cert.Certificate[] certs=sig.certificateChain;
        if(!processCertificates(certs,ks))
        	return false;
        // The content itself must be verified with the first certificate form the chain
		return sig.signedObject.verify(certs[0].getPublicKey(),s);
    }
    
    /**
    Process a chain of certificates. Will return <code>true</code> if either the last certificate in the chain is
    a trusted one or it is signed by a trusted certificate. Trusted certificates are retrieved
    from the keystore. Any error will lead to a failure.
    @param cert the chain if certificates to verify
    @param ks the keystore with the trusted certificates
    @return <code>true</code> if the chain is trustworthy
    @throws CertificateParsingException if a certificate could not be parsed
    */
    private boolean processCertificates(java.security.cert.Certificate[] cert,KeyStore ks) //throws CertificateParsingException
    {
    	for(int i=0;i<cert.length-1;i++)
    	{
    		java.security.cert.X509Certificate xcert=(java.security.cert.X509Certificate)cert[i];
	    	try
    		{
    		    // Verify with public key of next certificate
    			xcert.verify(((java.security.cert.X509Certificate)cert[i+1]).getPublicKey());
    			// Check validity period
    			xcert.checkValidity();
	    	}
    		catch(InvalidKeyException e)
    		{	// Invalid key in next certificate
    			return false;
	    	}
    		catch(SignatureException e)
    		{	// Something wrong with signature (possible tampered with)
    			return false;
	    	}
    		catch(NoSuchProviderException e)
    		{	// We cannot check
    			return false;
	    	}
    		catch(NoSuchAlgorithmException e)
    		{	// We cannot check
    			return false;
	    	}
    		catch(CertificateExpiredException e)
    		{	// Certificate already expired
    			return false;
	    	}
    		catch(CertificateNotYetValidException e)
   			{	// certificate not yet valid
	    		return false;
    		}
    		catch(CertificateException e)
    		{	// Unkown to us (encoding error)
    			return false;
	    	}
    	}
		// Last one must be accepted by us
   		java.security.cert.X509Certificate xcert=(java.security.cert.X509Certificate)cert[cert.length-1];
    	try
   		{
   		    // Check the last certificate
   			if(!verifySigner(xcert,xcert.getIssuerDN().getName(),ks))
   			    return false;
   			// Check period of validity
   			xcert.checkValidity();
    	}
   		catch(CertificateExpiredException e)
   		{
   			return false;
    	}
   		catch(CertificateNotYetValidException e)
		{
    		return false;
   		}
   		catch(CertificateException e)
   		{	// Unkown to us
   			return false;
    	}
    	return true;
    }
    
    /**
    Check whether a certificate is a trusted one or signed by a trusted one.
    @param c the certificate to check
    @param name the name of the certificate
    @param ks the key store with the trusted certificates
    @return <code>true</code> if the verification succeeded
    */
    private boolean verifySigner(Certificate c,String name,KeyStore ks)
    {
    	Certificate issuerCert=null;
    	X509Certificate sCert=null;
    	String signer=null;
    	Enumeration al=null;
    	try
    	{
	    	signer=ks.getCertificateAlias(c);
    		if(signer!=null)	// We already have this certificate as trusted
	    		return true;
	        // Go through all aliases
	    	al=ks.aliases();
	    }
	    catch(KeyStoreException e)
	    {
    		return false;
	    }
    	while(al.hasMoreElements() && issuerCert==null)
    	{
    		String s=(String)al.nextElement();
    		try
    		{
    			sCert=(X509Certificate)ks.getCertificate(s);
    		}
    		catch(Exception e)
    		{	// Just skip this on error
    			continue;
    		}
    		// We found an issuer certificate
    		if(sCert!=null && name.equals(sCert.getSubjectDN().getName()))
    			issuerCert=sCert;
    	}
    	if(issuerCert==null)	// No certificate from the authority which signed c found
    		return false;
        // Check the signature
    	try
    	{
    		c.verify(issuerCert.getPublicKey());
    	}
    	catch(Exception e)
    	{
    		return false;
    	}
    	return true;
    }

}
