/*
 * @(#)FilterEnumeration.java
 *
 * Utility classes
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util;

import java.util.Enumeration;
import java.util.NoSuchElementException;

/**
An enumeration of objects matching a certain criteria. It is based on another enumeration from
which only elements will be returned, which match certain criteria. Subclasses must implement 
the actual filtering function <code>protected boolean matchesFilter(Object toTest)</code>.

@author  Michael Sonntag
@version 1.0, 1.7.2000
*/
public abstract class FilterEnumeration implements Enumeration
{
    /** The complete enumeration to filter. */
    private Enumeration baseEnum;
    /** The next object to return or <code>null</code> if no more matching elements are available. */
    private Object nextMatching;
    
    /**
    Creates a new enumeration for filtering another enumeration.
    */
    protected FilterEnumeration()
    {
    }
    
    /**
    Sets the enumeration for filtering objects from another enumeration.
    <br><strong>The filtering parameters of the subclass must be set when this function is 
    called. Calls to <code>matchesFilter(Object toTest)</code> must be possible!</strong>
    @param fullEnum the enumeration of all elements which will be filtered
    @throws IllegalArgumentException if the parameter is <code>null</code>
    @throws IllegalStateException if the base enumeration is already set
    */
    protected void setBaseEnumeration(Enumeration fullEnum)
    {
        if(fullEnum==null)
            throw new IllegalArgumentException("Base enumeration must be provided");
        if(baseEnum!=null)
            throw new IllegalStateException("Base enumeration already set");
        baseEnum=fullEnum;
        goNextMatchingElement();
    }
    
    /**
    Tests if this enumeration contains more matching elements.
    @return <code>true</code> if this enumeration contains more matching elements; <code>false</code> otherwise
    @throws IllegalStateException if the base enumeration is not set
    */
    public boolean hasMoreElements()
    {
        if(baseEnum==null)
            throw new IllegalStateException("Base enumeration not set");
        return nextMatching!=null;
    }

    /**
    Returns the next matching element of this enumeration.
    @return the next matching element of this enumeration
    @throws NoSuchElementException if no more matching elements exist
    @throws IllegalStateException if the base enumeration is not set
    */
    public Object nextElement()
    {
        if(baseEnum==null)
            throw new IllegalStateException("Base enumeration not set");
        Object res=nextMatching;
        if(res==null)
            throw new NoSuchElementException(this.getClass().getName());
        goNextMatchingElement();
        return res;
    }

    /**
    The function implementing the filtering function.
    @param totest the object to test if it matches the filter
    @return <code>true</code> if the element should be returned; <code>false</code> if this
            element does not match the criteria and should be suppressed
    */
    protected abstract boolean matchesFilter(Object toTest);
    
    /**
    Moves on to the next matching object.
    */
    private void goNextMatchingElement()
    {
		Object tmp=null;       
		while(baseEnum.hasMoreElements() && (tmp==null || !matchesFilter(tmp)))
			tmp=baseEnum.nextElement();
	    if(tmp!=null && matchesFilter(tmp))
		    nextMatching=tmp;
		else
		    nextMatching=null;
    }
    
}