/*
 * @(#)FilterIterator.java
 *
 * Utility classes
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util;

import java.util.Iterator;
import java.util.NoSuchElementException;

/**
An iterator of objects matching a certain criteria. It is based on another iterator from
which only elements will be returned, which match certain criteria. Subclasses must implement 
the actual filtering function <code>protected boolean matchesFilter(Object toTest)</code>. The
<code>remove</code> method can only be called between <code>next</code> and <code>hasNext</code>
(After <code>hasNext</code> was called removal is no longer possible as the underlying iterator
is at another position already).

@author  Michael Sonntag
@version 1.0, 1.7.2000
*/
public abstract class FilterIterator implements Iterator
{
    /** The complete enumeration to filter. */
    private Iterator baseIter;
    /** The next object to return or <code>null</code> if no more matching elements are available. */
    private Object nextMatching;
    /**
    Marks whether we the current element was already returned to the user or not. Is <code>true</code>
    if we called <code>next</code> before. If <code>hasNext</code> is called, this will be 
    <code>false</code>: We must move on at least one object to see if it matches the criteria
    before we can tell the user whether at least one more matching element exists. Therefore
    in the underlying iterator we are at a different object and cannot anymore remove the
    last by us returned object.
    */
    private boolean currentReturned=true;
    
    /**
    Creates a new iterator for filtering another iterator.
    */
    protected FilterIterator()
    {
    }
    
    /**
    Sets the iterator for filtering objects from another iterator.
    <br><strong>The filtering parameters of the subclass must be set when this function is 
    called. Calls to <code>matchesFilter(Object toTest)</code> must be possible!</strong>
    @param fullIter the iterator of all elements which will be filtered
    @throws IllegalArgumentException if the parameter is <code>null</code> 
    @throws IllegalStateException if the base iterator is already set
    */
    protected void setBaseIterator(Iterator fullIter)
    {
        if(fullIter==null)
            throw new IllegalArgumentException("Base iterator must be provided");
        if(baseIter!=null)
            throw new IllegalStateException("Base iterator already set");
        baseIter=fullIter;
    }
    
    /**
    Returns <code>true</code> if the iteration has more elements. (In other words, 
    returns <code>true</code> if <code>next</code> would return an element rather than throwing an exception.)
    @return <code>true</code> if this iterator has more matching elements; <code>false</code> otherwise
    @throws IllegalStateException if the base iterator is not set
    */
    public boolean hasNext()
    {
        if(baseIter==null)
            throw new IllegalStateException("Base iterator not set");
        if(!currentReturned)
            return nextMatching!=null; // Already moved on in previous hasNext
        goNextMatchingElement();        
        currentReturned=false;
        return nextMatching!=null;
    }

    /**
    Returns the next matching element of this iterator.
    @return the next matching element of this iterator
    @throws NoSuchElementException if no more matching elements exist
    @throws IllegalStateException if the base iterator is not set
    */
    public Object next()
    {
        if(baseIter==null)
            throw new IllegalStateException("Base iterator not set");
        if(currentReturned)// Already moved on in hasNext?
            goNextMatchingElement();
        currentReturned=true;
        if(nextMatching==null)
            throw new NoSuchElementException(this.getClass().getName());
        return nextMatching;
    }

    /**
    The function implementing the filtering function.
    @param totest the object to test if it matches the filter
    @return <code>true</code> if the element should be returned; <code>false</code> if this
            element does not match the criteria and should be suppressed
    */
    protected abstract boolean matchesFilter(Object toTest);
    
    /**
    Removes from the underlying collection the last element returned by the iterator. This 
    method can be called only once per call to next. The behavior is unspecified if the 
    underlying collection is modified while the iteration is in progress in any way 
    other than by calling this method.
    If <code>hasNext</code> is called after <code>next</code>, this will throw a 
    <code>IllegalStateException</code>: We must move on at least one object to see if it 
    matches the criteria before we can tell the user whether at least one more matching 
    element exists. Therefore in the underlying iterator we are at a different object and 
    cannot anymore remove the last by us returned object.
    @throws IllegalStateException if the base iterator is not set or if the <code>next</code> 
            method has not yet been called, or the <code>remove</code> method has already 
            been called after the last call to the <code>next</code> method.
    */
    public void remove()
    {
        if(baseIter==null)
            throw new IllegalStateException("Base iterator not set");
        if(!currentReturned)
            throw new IllegalStateException("Cannot delete element now (do not call 'hasNext()' between 'next' and 'remove'!)");
        baseIter.remove();
    }
    
    /**
    Moves on to the next matching object.
    */
    private void goNextMatchingElement()
    {
    	Object tmp=null;       
		while(baseIter.hasNext() && (tmp==null || !matchesFilter(tmp)))
	        tmp=baseIter.next();
    	if(tmp!=null && matchesFilter(tmp))
		    nextMatching=tmp;
	    else
            nextMatching=null;
    }

}