/*
 * @(#)FlowingLayout.java
 *
 * GUI utility classes.
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.GUI;

import java.awt.FlowLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Insets;

/**
A layout where the components are arranged from left to right in rows. Afer the first row
is full, the next row is started. Row height is the height of the tallest object in this row.

@author Michael Sonntag
@version 1.0, 1.7.2000
*/
public class FlowingLayout extends FlowLayout
{
    /**
    Constructs a new FlowingLayout with a centered alignment and a
    default 5-unit horizontal and vertical gap.
    */
    public FlowingLayout() 
    {
	    this(CENTER,5,5);
    }

    /**
    Constructs a new FlowingLayout with the specified alignment and a
    default 5-unit horizontal and vertical gap.
    The value of the alignment argument must be one of
    <code>FlowLayout.LEFT</code>, <code>FlowLayout.RIGHT</code>,
    or <code>FlowLayout.CENTER</code>. Alignment means, that the complete row is aligned in this way,
    not the individual objects.
    @param align the alignment value
    */
    public FlowingLayout(int align)
    {
	    this(align,5,5);
    }

    /**
    Creates a new FlowingLayout manager with the indicated alignment
    and the indicated horizontal and vertical gaps.
    <p>
    The value of the alignment argument must be one of
    <code>FlowLayout.LEFT</code>, <code>FlowLayout.RIGHT</code>,
    or <code>FlowLayout.CENTER</code>.  Alignment means, that the complete row is aligned in this way,
    @param align the alignment value
    @param hgap the horizontal gap between components
    @param vgap the vertical gap between components
    */
    public FlowingLayout(int align,int hgap,int vgap)
    {
        super(align,hgap,vgap);
    }
    
    /**
    Returns the minimum or preferred dimensions needed to layout the components contained in the specified target container.
    @param target the component which needs to be laid out 
    @param minimum if <code>true</code>, calculates the minimum size else the preferred size
    @return the minimum dimensions to lay out the subcomponents of the specified container
    */
    private Dimension calculateSize(Container target,boolean minimum)
    {
        synchronized (target.getTreeLock())
        {
	        Dimension dim=new Dimension(0, 0);
	        Insets insets = target.getInsets();
	        int parWidth = target.getWidth() - (insets.left + insets.right + getHgap()*2);
	        int nMembers=target.getComponentCount();
	        int curWidth=0;
	        int maxWidth=0;
	        int curHeight=0;
	        int maxHeight=0;
	        int column=0;
	        int row=0;
	        for(int i=0;i<nMembers;i++)
	        {
	            Component m=target.getComponent(i);
         	    if(m.isVisible())
         	    {
		            Dimension d=null;
		            if(minimum)
		                d=m.getMinimumSize();
		            else
		                d=m.getPreferredSize();
		            int tempWidth=curWidth+d.width;
		            if(column>0)
		                tempWidth+=getHgap();
		            if(tempWidth>parWidth)
		            {   // Start new row
		                maxWidth=Math.max(maxWidth,curWidth);
		                if(row++>0)
		                    maxHeight+=getVgap();
		                maxHeight+=curHeight;
		                column=1;   // We already have the first object; it is the current one
		                curWidth=d.width;
		                curHeight=d.height;
		            }
                    else
                    {
    		            curHeight=Math.max(curHeight,d.height);               
		                if(column++>0)
		                    curWidth+=getHgap();
		                curWidth+=d.width;
		            }
	            }
	        }
	        // Final adjustment
            maxWidth=Math.max(maxWidth,curWidth);
            if(row++>0)
                 maxHeight+=getVgap();
            maxHeight+=curHeight;
        	dim.width=maxWidth+insets.left+insets.right+getHgap()*2;
	        dim.height=maxHeight+insets.top+insets.bottom+getVgap()*2;
        	return dim;
        }
    }

    /**
    Returns the preferred dimensions for this layout given the components in the specified target container.
    @param target component which needs to be laid out 
    @return the preferred dimensions to lay out the subcomponents of the specified container
    */
    public Dimension preferredLayoutSize(Container target)
    {
        return calculateSize(target,false);
    }
    

    /**
    Returns the minimum dimensions needed to layout the components contained in the specified target container.
    Exactly the same as <code>preferredLayoutSize</code>, only the contained components are asked for
    their minimum size instead of their preferred size.
    @param target the component which needs to be laid out 
    @return the minimum dimensions to lay out the subcomponents of the specified container
    */
    public Dimension minimumLayoutSize(Container target)
    {
        return calculateSize(target,true);
    } 

    /**
    Centers the elements in the specified row, if there is any slack.
    @param target the component which needs to be moved
    @param x the x coordinate
    @param y the y coordinate
    @param width the width dimensions
    @param height the height dimensions
    @param rowStart the beginning of the row
    @param rowEnd the the ending of the row
    @param ltr left-to-right layout
    */
    private void moveComponents(Container target,int x,int y,int width,int height,int rowStart,int rowEnd,boolean ltr)
    {
        synchronized(target.getTreeLock())
        {
	        switch(getAlignment())
	        {
	            case LEFT:
	                x+=ltr?0:width;
	            break;
	            case CENTER:
	                x+=width/2;
	            break;
	            case RIGHT:
	                x+=ltr?width:0;
	            break;
	            case LEADING:
	            break;
            	case TRAILING:
	                x+=width;
	            break;
	        }
	        for(int i=rowStart;i<rowEnd;i++)
	        {
	            Component m=target.getComponent(i);
        	    if(m.isVisible())
        	    {
	                if(ltr)
        	            m.setLocation(x,y+(height-m.getHeight())/2);
        	        else
	                    m.setLocation(target.getWidth()-x-m.getWidth(),y+(height-m.getHeight())/2);
                    x+=m.getWidth()+getHgap();
	            }
	        }
        }
    }

    /**
    Lays out the container. This method lets each component take its preferred size by reshaping 
    the components in the target container in order to satisfy the constraints of this FlowingLayout object.
    @param target the specified component being laid out
    */
    public void layoutContainer(Container target)
    {
        synchronized(target.getTreeLock()) 
        {
            target.setSize(target.getWidth(),Math.max(preferredLayoutSize(target).height,target.getHeight()));
	        Insets insets=target.getInsets();
	        int maxwidth=target.getWidth()-(insets.left+insets.right+getHgap()*2);
	        int nmembers=target.getComponentCount();
	        int x=0;
	        int y=insets.top+getVgap();
	        int rowh=0;
	        int start=0;
            boolean ltr=target.getComponentOrientation().isLeftToRight();
	        for(int i=0;i<nmembers;i++)
	        {
	            Component m=target.getComponent(i);
	            if(m.isVisible()) 
	            {
		            Dimension d=m.getPreferredSize();
		            m.setSize(d.width,d.height);
		            if ((x==0) || ((x+d.width)<=maxwidth))  // At least one object per line, up to what fits
		            {
		                if (x>0) 
			                x+=getHgap();
            		    x+=d.width;
		                rowh=Math.max(rowh,d.height);		            
		            }
		            else 
		            {   // Start a new line
		                // Adjust previous row
		                moveComponents(target,insets.left+getHgap(),y,maxwidth-x,rowh,start,i,ltr);
		                // First object in new line
		                x=d.width;
		                y+=getVgap()+rowh;
                	    rowh=d.height;
		                start=i;
		            }
	            }
	        }
	        // Adjust last line
    	    moveComponents(target,insets.left+getHgap(),y,maxwidth-x,rowh,start,nmembers,ltr);
        }
    }

}