/*
 * @(#)NamedListModel.java
 *
 * GUI utility classes.
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */
package FIM.Util.GUI;

import javax.swing.AbstractListModel;
import java.util.Vector;

/**
A model for a list consisting of objects which are represented through names instead of themselves.
Abstract class; subclasses must return the name of the object in the list, so will be mostly
lists of one type of element.

@author Michael Sonntag
@version 1.0, 1.7.2000
*/
public abstract class NamedListModel extends AbstractListModel
{
	/**
	Array for storing the additional data for the entries.
	@serial
    */
    private Vector data=new Vector();
    
    /**
    Inserts an element before the first entry.
    @param id the object to insert
    */
    public void insertData(Object id)
    {
        if(data!=null)
        {
            data.insertElementAt(id,0);
            fireIntervalAdded(this,0,0);
        }
    }
    
    /**
    Adds an element after the last entry.
    @param id the object to add
    */
    public void addData(Object id)
    {
        if(id!=null)
        {
            data.addElement(id);
            fireIntervalAdded(this,data.size()-1,data.size()-1);
        }
    }
    
    /**
    Retrieve the element at a specified index.
    @param index the index of the element to return
    @return the element at the provided index or <code>null</code> if the index is out of range
    */
    public Object getDataAt(int index)
    {
        if(index<0 || index>=data.size())
            return null;
        return data.elementAt(index);
    }
    
    /**
    Returns the name of the element at a certain index. Subclasses must provide an implementation.
    @param index the index of the element whose name should be retrieved
    @return the name of the element at the provided index
    */
    public abstract String getName(int index);
    
    // ListModel iterface from now on
    
    /**
    Retrieve the number of elements in the list.
    Equals the number of entries in this model.
    @return number of elements in the list
    */
    public int getSize()
    {
        return data.size();
    }
    
    /**
    Retrieve the element at a certain index.
    Returns here the name of the element at this index only, not the element itself.
    @param index the index of the element to return
    @return the elements at the index or <code>null</code> if the index is out of range
    */
    public Object getElementAt(int index)
    {
        if(index<0 || index>=data.size())
            return null;
        return getName(index);
    }
    
    /**
    Removes an element at a certain index from the list.
    Deletes also the data in this model.
    @param index the index of the element to remove
    */
    public void removeElement(int index)
    {
        if(index<0 || index>=data.size())
            return;
        data.removeElementAt(index);
        fireIntervalRemoved(this,index,index);
    }
    
    
}