/*
 * @(#)ReadOnlySelectionModel.java
 *
 * GUI utility classes.
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.GUI;

import javax.swing.*;
import javax.swing.event.*;

/**
This class extends the class <code>DefaultListSelectionModel</code>, but does only 
allow to set the selection if it is not marked as read-only: All requests are ignored then.
Used for read-only list-boxes. They can be scrolled, but the selection cannot be changed.

@author  Michael Sonntag
@version 1.0, 1.7.2000
@see javax.swing.JList
*/
public final class ReadOnlySelectionModel extends DefaultListSelectionModel
{
    /**
    If <code>true</code>, the selection cannot be changed 
    @serial
    */
    private boolean isReadOnly=false;
    
    /** 
    Retrieve the current mode of the selection model.
    @return <code>true</code> if the selection cannot be changed
    */
    public boolean isReadOnly()
    {
        return isReadOnly;
    }

    /**
    Sets the current mode of the selection model.
    The selection cannot be changed if the mode is set to <code>true</code> until
    it is set again to <code>false</code>.
    @param readOnly if <code>true</code> the selection cannot be changed anymore
    */
    public void setReadOnly(boolean readOnly)
    {
        isReadOnly=readOnly;
    }

    /**
    Change the selection to the empty set.  If this represents
    a change to the current selection then notify each ListSelectionListener.
    Does nothing if the model is set to read-only.
    */
    public void clearSelection()
    {
        if(!isReadOnly)
            super.clearSelection();
    }

    /** 
    Change the selection to be between index0 and index1 inclusive.
    If this represents a change to the current selection, then
    notify each ListSelectionListener. Note that index0 doesn't have
    to be less than or equal to index1.  
    Does nothing if the model is set to read-only.
    @param index0 one end of the interval.
    @param index1 other end of the interval
    */
    public void setSelectionInterval(int index0,int index1)
    {
        if(!isReadOnly)
            super.setSelectionInterval(index0,index1);
    }
        
    /** 
    Change the selection to be the set union of the current selection
    and the indices between index0 and index1 inclusive.  If this represents 
    a change to the current selection, then notify each 
    ListSelectionListener. Note that index0 doesn't have to be less
    than or equal to index1.  
    Does nothing if the model is set to read-only.
    @param index0 one end of the interval.
    @param index1 other end of the interval
    */
    public void addSelectionInterval(int index0,int index1)
    {
        if(!isReadOnly)
            super.addSelectionInterval(index0,index1);
    }
        
    /** 
    Change the selection to be the set difference of the current selection
    and the indices between index0 and index1 inclusive.  If this represents 
    a change to the current selection, then notify each 
    ListSelectionListener.  Note that index0 doesn't have to be less
    than or equal to index1.  
    Does nothing if the model is set to read-only.
    @param index0 one end of the interval.
    @param index1 other end of the interval
    */
    public void removeSelectionInterval(int index0,int index1)
    {
        if(!isReadOnly)
            super.removeSelectionInterval(index0,index1);
    }
        
    /** 
    Insert length indices beginning before/after index. This is typically 
    called to sync the selection model with a corresponding change
    in the data model.
    Does nothing if the model is set to read-only.
    @param index the index before/after which to insert
    @param length the number of indices to insert
    @param before if <code>true</code>, the indices will be inserted before the provided index
    */
    public void insertIndexInterval(int index,int length,boolean before)
    {
        if(!isReadOnly)
            super.insertIndexInterval(index,length,before);
    }
    
    /** 
    Remove the indices in the interval index0,index1 (inclusive) from
    the selection model.  This is typically called to sync the selection
    model width a corresponding change in the data model.
    Does nothing if the model is set to read-only.
    @param index0 the first index to be removed
    @param index1 the last index to be removed
    */
    public void removeIndexInterval(int index0,int index1)
    {
        if(!isReadOnly)
            super.removeIndexInterval(index0,index1);
    }
    
    /**
    Set the anchor selection index. 
    Does nothing if the model is set to read-only.
    @param anchorIndex the anchor selection index 
    @see #getAnchorSelectionIndex
    */
    public void setAnchorSelectionIndex(int anchorIndex)
    { 
        if(!isReadOnly)
            super.setAnchorSelectionIndex(anchorIndex);
    }
    
    /**
    Set the lead selection index. 
    Does nothing if the model is set to read-only.
    @param leadIndex the lead selection index
    @see #getLeadSelectionIndex
    */
    public void setLeadSelectionIndex(int leadIndex)
    {
        if(!isReadOnly)
            super.setLeadSelectionIndex(leadIndex);
    }
    
}