/*
 * @(#)VerticalExtendingPanel.java
 *
 * GUI utility classes.
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.GUI;

import java.awt.Dimension;
import java.awt.LayoutManager;
import java.awt.Rectangle;
import javax.swing.SwingConstants;

/**
A panel which will extend itself to fill the complete width of the parent and the height necessary
to contain the complete subelement. Will display a vertical scrollbar if necessary.

@author Michael Sonntag
@version 1.0, 1.7.2000
*/
public class VerticalExtendingPanel extends javax.swing.JPanel implements javax.swing.Scrollable
{
    
    /**
    Creates a new VerticalExtendingPanel with the specified layout manager and buffering strategy.
    @param layout the layout manager to use
    @param isDoubleBuffered a boolean, true for double-buffering, which uses additional memory space 
            to achieve fast, flicker-free updates
    */
    public VerticalExtendingPanel(LayoutManager layout,boolean isDoubleBuffered)
    {
        super(layout,isDoubleBuffered);
    }

    /**
    Create a new buffered VerticalExtendingPanel with the specified layout manager.
    @param layout the layout manager to use
    */
    public VerticalExtendingPanel(LayoutManager layout)
    {
        this(layout,true);
    }

    /**
    Create a new VerticalExtendingPanel with <code>FlowingLayout</code> and the specified buffering
    strategy. If <code>isDoubleBuffered</code> is true, the JPanel will use a double buffer.
    @param layout the layout manager to use
    @param isDoubleBuffered a boolean, <code>true</code> for double-buffering, which uses additional 
            memory space to achieve fast, flicker-free updates
    @see FIM.Util.GUI.FlowingLayout 
    */
    public VerticalExtendingPanel(boolean isDoubleBuffered)
    {
        this(new FlowingLayout(),isDoubleBuffered);
    }

    /**
    Create a new VerticalExtendingPanel with a double buffer and a <code>FlowingLayout</code>
    @see FIM.Util.GUI.FlowingLayout 
    */
    public VerticalExtendingPanel()
    {
        this(new FlowingLayout(),true);
    }

    /**
    Retrieve the preferred size of the viewport for a view component. Returns the preferred size
    of the component itself.
    @return the preferred size of the component
    */
    public Dimension getPreferredScrollableViewportSize()
    {
        return getPreferredSize();
    }

    /**
    Return <code>true</code> if a viewport should always force the width of this Scrollable to match the width 
    of the viewport. This is exactly the desired here, so always <code>true</code> is returned, disabling
    horizontal scrolling.
    @return always <code>true</code>
    */
    public boolean getScrollableTracksViewportWidth()
    {
        return true;
    }

    /**
    The height of the scrollable will never be made to match the viewport: On the contrary,
    we want the scrollable to extend till it fits all.
    @return always <code>false</code>
    */
    public boolean getScrollableTracksViewportHeight()
    {
        return false;
    }

    /**
    Retrieve the increment to expose a new row. As we cannot tell exactly the height of each row
    (they may differ) without recalculation, we return 1/4 of the visible height.
    @param visibleRect the view area visible within the viewport 
    @param orientation must be SwingConstants.VERTICAL here (HORIZONTAL will always return 0) 
    @param direction less than zero to scroll up, greater than zero for down
    @return the "unit" increment for scrolling in the specified direction
    */
    public int getScrollableUnitIncrement(Rectangle visibleRect,int orientation,int direction)
    {
        if(orientation==SwingConstants.VERTICAL)
        {
            int blHeight=0;
            blHeight=visibleRect.height/4;
            return blHeight;
        }
        return 0;
    }

    /**
    Retrieve the increment to expose a new block/page. As we cannot tell exactly the height of each row
    (they may differ) without recalculation, we return the visible height.
    @param visibleRect the view area visible within the viewport 
    @param orientation must be SwingConstants.VERTICAL here (HORIZONTAL will always return 0) 
    @param direction less than zero to scroll up, greater than zero for down
    @return the "unit" increment for scrolling in the specified direction
    */
    public int getScrollableBlockIncrement(Rectangle visibleRect,int orientation,int direction)
    {
        if(orientation==SwingConstants.VERTICAL)
        {
            int blHeight=0;
            blHeight=visibleRect.height;
            return blHeight;
        }
        return 0;
    }

}