/*
 * @(#)ReadOnlyStore.java
 *
 * Utility classes
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util;

import java.util.*;
import java.io.*;

/**
Similar to a <code>Hashtable</code>, but only allows to set a value once. Removal of entries 
or overwriting values is not allowed.<br>
The data is stored in serialized form. Therefore value must implement <code>java.io.Serializable</code>.
Changes after adding the object to the store are <strong>not</strong> reflected.

@author  Michael Sonntag
@version 1.0, 1.7.2000
@see java.util.Hashtable
@see java.io.Serializable
*/
public class ReadOnlyStore extends Dictionary implements Serializable
{
    /**
    The hashtable for storing the data
    */
    private Hashtable table=null;
    
    /**
    Constructs a new, empty read-only store.
    */
    public ReadOnlyStore()    
    {
        table=new Hashtable();
    }

    /**
    Constructs a new, empty read-only store with the specified initial capacity.
    @param initialCapacity the initial capacity 
    @exception IllegalArgumentException if the initial capacity is less than one
    */
    public ReadOnlyStore(int initialCapacity)
    {
        if(initialCapacity<1)
            throw new IllegalArgumentException("Size of store must be at least 1");
        table=new Hashtable(initialCapacity);
    }
    
    /**
    Tests if some key maps into the specified value in this store.
    @param value a value to search for
    @return <code>true</code> if and only if some key maps to the <code>value</code> argument 
    in this store as determined by the <tt>equals</tt> method, <code>false</code> otherwise
    @exception NullPointerException if the value is <code>null</code>
    @see #containsKey(Object)
    */
    public synchronized boolean contains(Object value)
    {
        return table.contains(value);
    }
    
    /**
    Tests if the specified object is a key in this store.
    @param key possible key
    @return <code>true</code> if and only if the specified object is a key in this hashtable,
    as determined by the <tt>equals</tt> method, <code>false</code> otherwise
    @see #contains(Object)
    */
    public synchronized boolean containsKey(Object key)
    {
        return table.containsKey(key);
    }
    
    /**
    Returns an enumeration of the values in this store.
    Use the Enumeration methods on the returned object to fetch the elements sequentially.
    @return  an enumeration of the values in this store
    @see java.util.Enumeration
    @see #keys()
    */
    public synchronized Enumeration elements()
    {
        return new ElementEnumeration(table);
    }

    /**
    Returns an enumeration of the keys in this store.
    @return  an enumeration of the keys in this store
    @see java.util.Enumeration
    @see #elements()
    */
    public synchronized Enumeration keys()
    {
        return table.keys();
    }
    
    /**
    Tests if this store maps no keys to values.
    @return  <code>true</code> if this store maps no keys to values, <code>false</code> otherwise.
    */
    public boolean isEmpty()
    {
        return table.isEmpty();
    }
    
    /**
    Returns the number of keys in this store.
    @return  the number of keys in this store
    */
    public int size()
    {
        return table.size();
    }

    /**
    Returns the value to which the specified key is mapped in this table.
    @param   key   a key in the htable.
    @return  the value to which the key is mapped in this table;
              <code>null</code> if the key is not mapped to any value in
              this table.
    @see     #put(Object, Object)
    */
    public synchronized Object get(Object key) 
    {
        Object res=table.get(key);
        if(res!=null)
            return reCreate((byte[])res);
        return res;
    }
    
    /**
    Removes an object from the store. Not allowed in a read-only store.
    @param key ignored
    @throws IllegalStateException always
    */
    public synchronized Object remove(Object key) 
    {
        throw new IllegalStateException("ReadOnlyStore does not allow removal of elements");
    }
    
    /**
    Maps the specified <code>key</code> to the specified 
    <code>value</code> in this table. Neither the key nor the 
    value can be <code>null</code>. The value must be <code>Serializable</code>. The key
    may not currently exist in the table (Overwriting value is prohibited).<p>
    The value can be retrieved by calling the <code>get</code> method 
    with a key that is equal to the original key. 
    @param key the table key
    @param value the value
    @return always <code>null</code>, as overwriting is not allowed
    @exception NullPointerException if the key or value is <code>null</code>.
    @exception IllegalArgumentException if the value is not serializable.
    @see Object#equals(Object)
    @see #get(Object)
    */
    public synchronized Object put(Object key, Object value) 
    {
	    // Make sure the key is not null
    	if(key==null)
	        throw new NullPointerException("Key must be provided");
	    // Make sure the value is not null
    	if(value==null)
	        throw new NullPointerException("Value must be provided");
        if(!(value instanceof Serializable))
            throw new IllegalArgumentException("Only instances of java.io.Serializable can be stored in a ReadOnlyStore");
        if(table.containsKey(key))
            throw new IllegalArgumentException("Table already contains the key "+key.toString()+"!");
        try
        {
            ByteArrayOutputStream bos=new ByteArrayOutputStream();
            ObjectOutputStream out=new ObjectOutputStream(bos);
            out.writeObject(value);
            out.close();
            ByteArrayInputStream bin=new ByteArrayInputStream(bos.toByteArray());
            ObjectInputStream in=new ObjectInputStream(bin);
            in.readObject();
            table.put(key,bos.toByteArray());
        }
        catch(Exception e)
        { 
            throw new IllegalArgumentException("Cannot store object: "+e.toString());
        }
        return null;
    }
    
    /**
    Recreates an object from the byte array it is stored in.
    @param obj the object encoded in a byte array
    @return the recreated object
    */
    private Object reCreate(byte[] obj)
    {
        try
        {
            ByteArrayInputStream bin=new ByteArrayInputStream(obj);
            ObjectInputStream in=new ObjectInputStream(bin);
            return in.readObject();
        }
        catch(IOException e)
        {   // Should never happen; we checked that it can be deserialized!
            System.err.println();
        }
        catch(ClassNotFoundException e)
        {   // Should never happen; we checked that it can be deserialized!
            System.err.println();
        }
        return null;
    }
    
    /**
    Helper class: Simple wrapping enumeration which recreates the object from 
    the byte-array upon calling <code>nextElement</code>.
    */
    private class ElementEnumeration implements Enumeration
    {
        private Enumeration base=null;
        
        public ElementEnumeration(Hashtable table)
        {
            base=table.elements();
        }
        
        public boolean hasMoreElements()
        {
            return base.hasMoreElements();
        }

        public Object nextElement()
        {
            Object res=base.nextElement();
            return reCreate((byte[])res);
       }

    }
    
}