/*
 * @(#)SMSPost.java
 *
 * Classes for sending SMS messages using websites
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.SMS;

import java.io.IOException;
import java.util.Vector;
import java.net.URL;
import java.net.URLEncoder;
import java.net.MalformedURLException;
import FIM.Util.WWW.Cookies.CookieStore;
import FIM.Util.WWW.WWWPageUtils;
import FIM.Util.WWW.Form.HTMLForm;
import FIM.Util.WWW.Form.FormElement;
import FIM.Util.WWW.Form.TextElement;
import FIM.Util.WWW.Form.SelectableElement;

/*

@author  Michael Sonntag
@version 1.0, 1.7.2000
*/
public final class SMSPost
{

    /** Print a trace of the pages received if <code>true</code>. */
    private final static boolean TRACE=false;

    public final static String PROVIDER_MAX="676";
    public final static String PROVIDER_A1="664";
    public final static String PROVIDER_ONE="699";
    public final static String PROVIDER_TELERING="650";


    private URL providerUrl=null;
    private String country_prefix="";
    private String provider_prefix="";
    private String number_prefix="";
    private String message_prefix="";
    private int maxMessageLen=140;
    private CookieStore cookies=new CookieStore();

    /* No longer operational!
    public final static SMSPost Lion_SMS()
    {
        return new SMSPost("http://sms.lion.cc/","ls_country","do","detail","li_message",144);
    }
    */

    public final static SMSPost ORF_SMS()
    {
        return new SMSPost("http://sms.orf.at/",null,"gsm_type","tel_number","text",147);
    }

    /*
    Not fully functional: Must start on homepage, there unique ID is given under which to request the send-page
    public final static SMSPost SMS_AT()
    {
        return new SMSPost("http://www.sms.at/",null,"gsmnetz","cellnr","nachr",151);
    }
    */

public SMSPost(String provider_url,String country_prefix,String provider_prefix,String number_prefix,String message_prefix,int maxMessageLen)
{
    if(provider_url==null || provider_url.length()<1)
        throw new IllegalArgumentException("URL to SMS-provider must be provided");
    try
    {
        providerUrl=new URL(provider_url);
    }
    catch(MalformedURLException e)
    {
        throw new IllegalArgumentException("Could not parse provider URL");
    }
    this.country_prefix=country_prefix;
    if(provider_prefix==null || provider_prefix.length()<1)
        throw new IllegalArgumentException("SMS-provider prefix must be provided");
    this.provider_prefix=provider_prefix;
    if(number_prefix==null || number_prefix.length()<1)
        throw new IllegalArgumentException("Number prefix must be provided");
    this.number_prefix=number_prefix;
    if(message_prefix==null || message_prefix.length()<1)
        throw new IllegalArgumentException("Message prefix must be provided");
    this.message_prefix=message_prefix;
    this.maxMessageLen=maxMessageLen;
}

/**
Retrieve the <code>URL</code> of the service provider (Base URL).
@return provider base <code>URL</code> for pages
*/
public URL getProvider()
{
    return providerUrl;
}

/**
@return null on success, error message from server otherwise (status code, not page!)
*/
public String sendSMS(String country,String provider,String number,String message) throws CannotSendException
{
    String[] parts=canonicalizeMessage(message);
    String res=null;
    for(int i=0;i<parts.length && res==null;i++)
        res=internalSendSMS(country,provider,number,parts[i]);
    return res;
}

protected String internalSendSMS(String country,String provider,String number,String message) throws CannotSendException
{
    try
    {
        String page=WWWPageUtils.fetchPage(providerUrl,cookies);
        HTMLForm form=getSMSForm(page);
        // Short delay to allow website to do some stuff (Pretend to be human and read the page!)
        try
        {
            Thread.sleep(8000);
        }
        catch(InterruptedException e)
        {   // If interrupted: just continue and hope we waited long enough
            System.err.print("");
        }
        if(country_prefix!=null)
        {
            if(!selectCountry(form,country))
                throw new CannotSendException("No country element found");
        }
        if(!selectProvider(form,provider))
        {
            try
            {
                page=form.doSubmit(null,cookies,null);
                // Short delay to allow website to do some stuff
                try
                {
                    Thread.sleep(8000);
                }
                catch(InterruptedException e)
                {   // If interrupted: just continue and hope we waited long enough
                    System.err.print("");
                }
                form=getSMSForm(page);
                if(!selectProvider(form,provider))
                    throw new CannotSendException("No provider element found (second page)");
            }
            catch(IOException e)
            {
                throw new CannotSendException("No provider element found");
            }
        }
        if(!selectNumber(form,number))
        {
            try
            {
                page=form.doSubmit(null,cookies,null);
                try
                {
                    Thread.sleep(8000);
                }
                catch(InterruptedException e)
                {   // If interrupted, just continue and hope we waited enough
                    System.err.print("");
                }
                form=getSMSForm(page);
                if(!selectNumber(form,number))
                    throw new CannotSendException("No number element found (second page)");
            }
            catch(IOException e)
            {
                throw new CannotSendException("No number element found");
            }
        }
        if(!insertMessage(form,message))
            throw new CannotSendException("No message element found");
        page=form.doSubmit(null,cookies,null);
        // Try checking the result: "unsuccessful" or "Fehler" will let us conclude that it did NOT work
        // "successful" or "erfolgreich" was a success
        // If neither is found, we guess it did NOT work (which is probably right in this case)
        String test=page.toLowerCase();
        if(test.indexOf(" unsuccessful ")>0 || test.indexOf(" Fehler ")>0)
        {
            throw new CannotSendException("\"Unsuccessful\" or \"Fehler\"in reply!");
        }
        if(test.indexOf(" successful ")<0 && test.indexOf(" erfolgreich ")<0)
        {
            throw new CannotSendException("No \"successful\" or \"erfolgreich\"in reply!");
        }
        if(TRACE)
            System.out.println(page);
    }
    catch(IOException e)
    {
        throw new CannotSendException(e.getMessage());
    }
    return null;
}

public String[] canonicalizeMessage(String message)
{
    String[] res;
    String tmp=message;
    String part="";
    int len=0;
    while(len<maxMessageLen && tmp.length()>0)
    {
        if(tmp.startsWith("\n"))
            len++;  // Linebreak counts as two characters!
        len++;
        part=part+tmp.substring(0,1);
        tmp=tmp.substring(1);
    }
    if(tmp.length()>0)
    {
        // have to make parts
        Vector parts=new Vector();
        while(message.length()>0)
        {
            part="";
            len=6;
            while(len<maxMessageLen && message.length()>0)
            {
                if(message.startsWith("\n"))
                    len++;  // Linebreak counts as two characters!
                len++;
                part=part+message.substring(0,1);
                message=message.substring(1);
            }
            parts.add(part);
        }
        res=new String[parts.size()];
        parts.copyInto(res);
        for(int i=0;i<parts.size();i++)
            res[i]=""+(i+1)+"/"+res.length+"\n"+res[i];
    }
    else
    {
        res=new String[1];
        res[0]=part;
    }
    return res;
}

protected HTMLForm getSMSForm(String page) throws CannotSendException
{
    HTMLForm[] forms=null;
    try
    {
        forms=HTMLForm.getForms(page);
    }
    catch(java.text.ParseException e)
    {
        throw new CannotSendException("Parsing error in form: "+e.getMessage()+" at "+e.getErrorOffset());
    }
    HTMLForm form=null;
    for(int i=0;form==null && i<forms.length;i++)
    {
            if(forms[i].getActionURL()==null || forms[i].getActionURL().getHost().equalsIgnoreCase(providerUrl.getHost()))
                form=forms[i];
    }
    if(form==null && forms.length<1)
        throw new CannotSendException("Could not find the inital form on the page");
    if(form==null)
        form=forms[0];
    if(form.getBaseURL()==null)
        form.setBaseURL(providerUrl);
    return form;
}

protected boolean insertMessage(HTMLForm form,String message) throws CannotSendException
{
    if(message==null || message.length()<1)
        return false;   // No value passed in; we can never find it (at insert a value)
    FormElement[] prov=form.getElements(message_prefix);
    if(prov.length!=1)
        return false;
    if(!(prov[0] instanceof TextElement))
        throw new CannotSendException("Message element of wrong form type");
    ((TextElement)prov[0]).setText(message);
    return true;
}

protected boolean selectNumber(HTMLForm form,String number) throws CannotSendException
{
    if(number==null || number.length()<1)
        return false;   // No value passed in; we can never find it (at insert a value)
    FormElement[] prov=form.getElements(number_prefix);
    if(prov.length!=1)
        return false;
    if(!(prov[0] instanceof TextElement))
        throw new CannotSendException("Number element of wrong form type");
    ((TextElement)prov[0]).setText(number);
    return true;
}

protected boolean selectProvider(HTMLForm form,String service) throws CannotSendException
{
    if(service==null || service.length()<1)
        return false;   // No value passed in; we can never find it (at insert a value)
    FormElement[] prov=form.getElements(provider_prefix);
    if(prov.length<1)
        return false;
    if(prov[0] instanceof TextElement)
        ((TextElement)prov[0]).setText(service);
    else if(prov[0] instanceof SelectableElement)
    {
        for(int i=0;i<prov.length;i++)
        {
            SelectableElement el=(SelectableElement)prov[i];
            if(el.getLabel()!=null && el.getLabel().indexOf(service)>=0 || el.getValue().indexOf(service)>=0)
                el.select();
        }
        if(((SelectableElement)prov[0]).getGroup().selectedCount()!=1)
            throw new CannotSendException("Could not select the provider");
    }
    else
        throw new CannotSendException("Provider element not recognized");
    return true;
}

protected boolean selectCountry(HTMLForm form,String country) throws CannotSendException
{
    if(country==null || country.length()<1)
        return false;   // No value passed in; we can never find it (at insert a value)
    FormElement[] prov=form.getElements(country_prefix);
    if(prov.length<1)
        return false;
    if(prov[0] instanceof TextElement)
        ((TextElement)prov[0]).setText(country);
    else if(prov[0] instanceof SelectableElement)
    {
        for(int i=0;i<prov.length;i++)
        {
            SelectableElement el=(SelectableElement)prov[i];
            if(el.getLabel()!=null && el.getLabel().indexOf(country)>=0 || el.getValue().indexOf(country)>=0)
                el.select();
        }
        if(((SelectableElement)prov[0]).getGroup().selectedCount()!=1)
            throw new CannotSendException("Could not select the country");
    }
    else
        throw new CannotSendException("Country element not recognized");
    return true;
}

/**
This class can also be used as stand-alone.
Usage: SMSPost country provider number message<br>
Country: Only needed for lion.cc (sterreich, Deutschland, Schweiz)<br>
Provider (sterreich): 676 (Max), 664 (A1), 699 (One) or 650 (Telering)<br>
Number: At least 7 digits
*/
static public void main(String args[]) throws Exception
{
    System.out.println("SMSPost: Send SMS via command line\n(c) Michael Sonntag 2000\n");
    String result=null;
    if(args.length==4 && args[1].startsWith("6") && args[2].length()>=7)
    {
        result=SMSPost.ORF_SMS().sendSMS(args[0],args[1],args[2],args[3]);
    }
    else
    {
        System.err.println("\nUsage: SMSPost provider number message\nProvider: 676 (Max), 664 (A1) or 699 (One)\nNumber: At least 7 digits\n");
        System.exit(2);
    }
    if(result!=null)
    {
        System.out.println("Error sending SMS: "+result);
        System.exit(1);
    }
    System.out.println("SMS successfully sent.");
    System.exit(0);
}

}

