/*
 * @(#)CancellableThread.java
 *
 * Part of the AgentSystem "POND"
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.Threads;

/**
A thread which can be cancelled by another thread. If it does not want to die, its priority
is set to a minimum ({@link java.lang.Thread#stop() Thread.stop()} is deprecated).

@author  Michael Sonntag
@version 1.0, 1.7.2000
*/
public abstract class CancellableThread extends Thread
{
    /** true if this thread has been cancelled */
	private transient volatile boolean cancelled=false;
	
	/**
	Cancel this thread. Will try to stop it, but returns immediately.
	*/
	public void cancel()
	{
		synchronized(this)
		{
			cancelled=true;
		}
		// If it is waiting, set it free
		interrupt();
	}

    /**
    Returns <code>true</code> if this thread-object (not the current thread!) is cancelled.
    @return if this thread is cancelled
    */
	public boolean isCancelled()
	{
		if(cancelled)
			return true;
	    // If false, try it in synchronized form to be sure
		synchronized(this)
		{
			return cancelled;
		}
	}

    /**
    Cancels this tread and wait till it has stopped (or it is impossible).
    Will wait for a number of seconds or indefinitely (called with 0).
    If called from within the thread a <code>ThreadDeath</code> error will be thrown
    if it dies not gracefully.
    @param maxWaitToDie number of seconds to wait for the thread to stop gracefully
    @return <code>true</code> if the thread is stopped (false: continues to run)
    */
	public boolean terminate(long maxWaitToDie)
	{
		if(!isAlive())
			return true;    // Already dead
	    // Cancel it
		cancel();
		// Wait to die
		try
		{
			join(maxWaitToDie*1000);
		}
		catch(InterruptedException e)
		{   // Ignore it and chekc below whether still alive
		    System.err.print("");
		}
		if(!isAlive())
			return true;    // Now dead
	    if(Thread.currentThread()==this)
	        throw new ThreadDeath();    // If called from within us, throw ThreadDeath to terminate
	    // Nothing works, so set priority to a minimum
		setPriority(Thread.MIN_PRIORITY);
		// Thread.stop() is deprecated, which would completely terminate it
//		Thread.currentThread().stop();
		return false;
	}

    /**
    Main method of the thread. Subclasses must override and place funtion in here.
    Any subclasses should regularly check if this thread is cancelled, especially
    when a <code>wait</code> was interrupted.
    */
	public abstract void run();
}

