/*
 * @(#)Timer.java
 *
 * Part of the AgentSystem "POND"
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.Threads;

import java.util.Vector;
import java.util.Enumeration;
import java.io.IOException;
import java.io.Serializable;

/**
A timer, which can be set for recurring or once calling a number of listeners.
The timer can be serialized and will be created exactly as it was serialized, including
a new Thread for waiting if the serialized version was running (elapsed time is reset although).

@author Michael Sonntag
@version 1.0, 1.7.2000
@see FIM.Util.Threads.TimerAction
*/
public final class Timer implements Serializable
{

    /** Contained thread for waiting for the event */
	private class TimerThread extends CancellableThread
	{
	    /** Number of seconds to wait till the event */
		protected transient long secondsToWait=-1;
		/** List of all objects to notify */
		protected transient Vector listeners=null;
		/** If the thread should end */
		protected volatile transient boolean end=false;
		/** If true the event occurs only once, otherwise the thread is immediately restarted */
		protected boolean once=true;
		/** ID of the timer */
		protected long id=-1;
	
	    /**
	    Creates a new thread for waiting for an event
	    @param actionListeners object to notify after time has elapsed
	    @param interval interval between/till event in seconds
	    @param onlyOnce if true, event will occur only once
	    @param id the ID of the timer
	    */
		public TimerThread(Vector actionListeners,long interval,boolean onlyOnce,long id)
		{
			secondsToWait=interval;
			listeners=actionListeners;
			once=onlyOnce;
			this.id=id;
			this.setDaemon(true);
			this.setPriority(Thread.MIN_PRIORITY);
		}
		
		/**
		Main method for waiting for the event to occur.
		*/
		public void run()
		{
			do
			{
			    // Do we have to wait?
				if(secondsToWait>0)
				{
					try
					{
						synchronized(this)
						{
							wait(1000*secondsToWait);
						}
					}
					catch(InterruptedException e)
					{	// No notification on interruption
						return;
					}
				}
				// Notify all
    			Enumeration enum=listeners.elements();
				while(enum.hasMoreElements())
					    ((TimerAction)enum.nextElement()).doTimerAction(id);
			}while(!end && !once);  // restart if continuos event
		}
	}


    /**
    Number of seconds to wait till the event occurs.
	@serial
	*/
	private long waitSeconds=-1;
    /**
    The listeners, which have to be notified when the event occurs.
	@serial
	*/
	private Vector actionReceivers=new Vector();
    /**
    Whether the thread is currently running or not.
	@serial
	*/
	private boolean running=false;
    /**
    If <code>true</code> the event will only occur one, otherwise the delay will
    be restarted immediately after notifying the listeners.
	@serial
	*/
	private boolean once=true;
    /**
    The ID of the timer.
	@serial
	*/
	private long id=0;
	
    /**
    The next ID for a timer.
	@serial
	*/
	private static long ID=0;
	
	/**
	Out thread for waiting for the event.
	*/
	private transient TimerThread tt=null;

    /**
    Create a new timer for a certain period, which might be recurring.
    The timer will not be started.
    @param secondsToWait number of seconds till (or between) the event
    @param recurring if <code>true</code> the timer will be restarted 
           automatically after notifying all listeners
    */
	public Timer(long secondsToWait,boolean recurring)
	{
		waitSeconds=secondsToWait;
		once=!recurring;
		setID();
	}
	
	/**
    Create a new timer for a certain period which occurs once.
    The timer will not be started.
    @param secondsToWait number of seconds till (or between) the event
	*/
	public Timer(long secondsToWait)
	{
		waitSeconds=secondsToWait;
		setID();
	}
	
	/**
	Setd the ID of the current timer and advances the global count.
	Extra procedure needed to allow being synchronized.
	*/
	private synchronized void setID()
	{
		id=ID++;
	}
	
	/**
	Returns the ID of this timer.
	@return the ID of this timer
	*/
	public long getID()
	{
	    return id;
	}
	
	/**
	Starts the timer.
	*/
	public void startTimer()
	{
		if(tt!=null && Thread.currentThread()==tt)
		{   // We are called from within a work-procedure
		    // Stop timer
			tt.end=true;
			tt=null;
			running=false;
		}
		if(tt==null && waitSeconds>0 && actionReceivers!=null)
		{
			tt=new TimerThread(actionReceivers,waitSeconds,once,id);
			tt.start();
			running=true;
		}
	}
	
	/**
	Stops the timer.
	*/
	public void stopTimer()
	{
		if(tt!=null)
		{
			if(Thread.currentThread()==tt)
			{
				tt.end=true;
				tt=null;
				running=false;
			}
			else
			{
				tt.terminate(5);
				tt=null;
				running=false;
			}
		}
	}
	
	/**
	Returns the interval of the timer (Complete time, not the remaining one!).
	@return the interval of the timer
	*/
	public long getInterval()
	{
		return waitSeconds;
	}
	
	/**
	Sets the interval of the timer.
	If the timer is currently running, it is stopped and restarted with the new interval.
	*/
	public void setInterval(long newInterval)
	{
		if(tt==null)
			waitSeconds=newInterval;
		else
		{
			if(Thread.currentThread()==tt)
			{
				tt.secondsToWait=newInterval;
				waitSeconds=newInterval;			
			}
			else
			{
				stopTimer();
				waitSeconds=newInterval;			
				startTimer();		
			}	
		}
	}
	
	/**
	Adds a new listener for this timer.
	@param ta the listener to add
	*/
	public void addTickListener(TimerAction ta)
	{
		if(!actionReceivers.contains(ta))
			actionReceivers.addElement(ta);
	}
	
	/**
	emoves a listener for this timer.
	@param ta the listener to remove
	*/
	public void removeTickListener(TimerAction ta)
	{
		actionReceivers.removeElement(ta);
	}
	
	/**
	Writes the timer to a stream.
	@param out the stream to write to
	@throws IOException if an IO error occurs during writing to the stream
	*/
	private void writeObject(java.io.ObjectOutputStream out) throws IOException
	{
/*	    
		if(running)
		{
			StopTimer();
			running=true;
		}
*/		
		out.defaultWriteObject();
		out.writeLong(ID);  // Must write this to avoid getting duplicate IDs!
	}
	 
	/**
	Reads the timer from a stream. If it was running when it was written to the stream the timer ist started.
	@param in the stream to read from
	@throws IOException if an IO error occurs during reading from the stream
	@throws ClassNotFoundException if a class could not be found
	*/
	private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException
	{
		in.defaultReadObject();
		ID=in.readLong();
		if(running)
			startTimer();
	}

}