/*
 * @(#)TimerQueue.java
 *
 * Part of the AgentSystem "POND"
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.Threads;

import java.util.Date;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Vector;
import java.util.Enumeration;
import java.text.DateFormat;

/**
Creates a queue, where an arbitrary number of alarms can be set for a given time or after
a specified interval. Only one thread will be running at every time, waiting for the time to elapse.
The time when the alarms are called is not guaranteed, but they will not be called BEFORE their time and
they will be called in ASCENDING order. Alarms with exactly the same point in time will be called in
the order they are added.<br>
Each event is assigned an ID, which is returned on creation. It can be used for cancelling it and
is sent as a paremeter upon notification.

@author Michael Sonntag
@version 1.0, 1.7.2000
@see FIM.Util.Threads.Timer
*/
public class TimerQueue
{
    /** The list of events */
    private Vector queue=new Vector();
    /** The thtead waiting for the next event to occur */
    private TimingThread waitThread;    
    /** The next ID for an event */
    private static long ID=0;
    
    /**
    Class of the thread waiting for the next event.
    */
    private class TimingThread extends Thread
    {
        /** If this thread should be cancelled */
        private volatile Boolean cancel=new Boolean(false);

        /**
        Main method of the thread
        */
        public synchronized void run()
        {
            boolean cont=true;
            long delay=1;
            // Get delay till next event
            synchronized(queue)
            {
                cont=queue.size()>0;
                if(cont)
                {
                    long now=(new Date()).getTime();
                    long then=((Alarm)queue.elementAt(0)).atTime.getTime();
                    delay=then-now;
                }
            }
            // While events to wait for
            while(cont)
            {
                try
                {
                    if(delay>0) // If now or already passed, don't wait; notify immediately
                        wait(delay);
                }
                catch(InterruptedException e)
                {          
                    return;
                }
                synchronized(queue)
                {   // Cancel necessary, as interrupt will not work on all platforms (????)
                    if(!cancel.booleanValue() && queue.size()>0)
                    {
                        Alarm toCall=(Alarm)queue.elementAt(0);
                        toCall.recipient.doTimerAction(toCall.id);
                        queue.removeElementAt(0);
                    }
                    // Next event
                    cont=queue.size()>0;
                    if(cont)
                    {
                        long now=(new Date()).getTime();
                        long then=((Alarm)queue.elementAt(0)).atTime.getTime();
                        delay=then-now;
                    }
                }
            }
        }
    }
    
    /**
    Storage class for a single event.
    */
    private class Alarm implements java.io.Serializable 
    {
        /** The time at which the event happens */
        private Date atTime;
        /** Where to send the notification */
        private TimerAction recipient;
        /** The ID of the event */
        private long id;
        
        /**
        Creates a new event.
        @param atTime the time of the event
        @param recipient whom to notify
        @param id the ID of the event
        */
        public Alarm(Date atTime,TimerAction recipient,long id)
        {
            this.atTime=atTime;
            this.recipient=recipient;
            this.id=id;
        }
    }
    
    /**
    Creates a new and empty timer queue.
    */
    public TimerQueue()
    {
    }
    
    /**
    Adds a new alarm at an absolute point in time.
    @param atTime the point in time when the alarm will be called
    @param recipient the interface, where the method doTimerAction will be called
    @return ID of the event
    */
    public long addAtTimeAction(Date atTime,TimerAction recipient)
    {   
        Alarm newAlarm=new Alarm(atTime,recipient,ID++);
        int i=0;
        synchronized(queue)
        {   // Add AFTER all events using exactly the same date
            while(i<queue.size() && !newAlarm.atTime.before(((Alarm)queue.elementAt(i)).atTime))
                i++;
            queue.insertElementAt(newAlarm,i);
            if(i==0 && queue.size()>1)
            {   // Stop current thread
                // Only ONE thread will be running always, so stop old one and start new and shorter wait-period
                waitThread.cancel=new Boolean(true);
                waitThread.interrupt();
                waitThread.setPriority(Thread.MIN_PRIORITY);
                waitThread=null;    // Abandon; will eventually terminate
            }
            if(i==0)
            {   // Start/Restart thread
                waitThread=new TimingThread();
                waitThread.start();
            }
        }
        return newAlarm.id;
    }
    
    /**
    Adds a new alarm at a relative point in time. The alarm will be called the 
    specified number of seconds after this method is called.
    @param atTime the duration in seconds before the alarm will be called
    @param recipient the interface, where the method doTimerAction will be called
    @return ID of the event
    */
    public long addAfterIntervalAction(int afterSecondsInterval,TimerAction recipient)
    {
        GregorianCalendar cal=new GregorianCalendar();
        cal.setTime(new Date());
        cal.add(Calendar.SECOND,afterSecondsInterval);
        return addAtTimeAction(cal.getTime(),recipient);
    }
    
    /**
    Removes a alarm from the queue.
    If it is called before the activation of the event: There is NO guarantee, that the action
    will not be called if the time before is too short. After this method returns, the action
    will guaranteed NOT be called any more.
    @param id The id of the alarm
    */
    public void cancelAction(long id)
    {
        int i=0;
        synchronized(queue)
        {
            Alarm toRemove=null;
            // Retrieve event from queue
            do
            {
                Alarm check=(Alarm)queue.elementAt(i++);
                if(check.id==id)
                    toRemove=check;
            }
            while(toRemove==null && i<queue.size());
            if(toRemove!=null)
            {
                // Remove event from queue
                queue.removeElement(toRemove);
                if(i==1)
                {   // Was the first element in the queue; the one the thread was waiting for
                    waitThread.cancel=new Boolean(true);
                    waitThread.interrupt();
                    waitThread.setPriority(Thread.MIN_PRIORITY);
                    waitThread=null;    // Abandon; will eventually terminate
                    if(queue.size()>0)
                    {   // Start/Restart new thread for next alarm in queue
                        waitThread=new TimingThread();
                        waitThread.start();
                    }
                }
            }
        }
    }
    
    /**
    Removes all alarms from the queue.
    */
    public void killAll()
    {
        synchronized(queue)
        {
            queue.removeAllElements();
            if(waitThread!=null)
            {
                waitThread.cancel=new Boolean(true);
                waitThread.interrupt();
                waitThread.setPriority(Thread.MIN_PRIORITY);
                waitThread=null;    // Abandon; will eventually terminate
            }
        }
    }
    
    /**
    Returns a string representation of this object. The number or alarms and details for all of them
    will be returned.
    @return A String containing information about the state of this object
    */
    public String toString()
    {
        String res="";
        synchronized(queue)
        {
            Enumeration enum=queue.elements();
            res+="Timerqueue has "+queue.size()+" alarms waiting:\n";
            DateFormat df=DateFormat.getDateTimeInstance(DateFormat.LONG,DateFormat.LONG);
            while(enum.hasMoreElements())
            {
                Alarm elem=(Alarm)enum.nextElement();
                res+="At "+df.format(elem.atTime)+" call "+elem.recipient.toString()+" with id "+elem.id+"\n";
            }
        }
        return res+"\n";
    }

}