/*
 * @(#)CanonicalFilterReader.java
 *
 * Utility classes for managing webpages
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.WWW;

import java.io.Reader;
import java.io.FilterReader;
import java.io.IOException;
import java.util.Hashtable;

/**
Removes all whitespaces and spaces at the beginning of the stream and converts all 
multiple space characters within the stream to a single space (2nd, 3rd, ... are suppressed).

@author  Michael Sonntag
@version 1.0, 1.7.2000
*/
public class CanonicalFilterReader extends FilterReader
{
    /** Marks that we are at the beginning of the stream (The first "letter"-character was not yet encountered). */
    protected boolean atBeginning=true;
    /** The character we returned on the last read-request. */
    protected int lastChar=0;

    /**
    Creates a new reader.
    @param in the reader to read from 
    */
	public CanonicalFilterReader(Reader in)
	{
		super(in);
	}
	
	/**
    Tell whether this stream supports the mark() operation (this stream does not).
    @return always <code>false</code>
	*/
	public boolean markSupported()
	{
	    return false;
	}

	/**
    Read characters into an array. This method will block until some input is available, 
    an I/O error occurs, or the end of the stream is reached.
    @param cbuf destination buffer 
    @param off the offset in the buffer the first character will be written to
    @param len the number of characters to read
    @return the number of bytes read, or -1 if the end of the stream has been reached
    @throws IOException if an I/O error occurs 
    */
    public int read(char[] cbuf,int off,int len) throws IOException
    {
        if(off<0)
            throw new IllegalArgumentException("Offset must be at least 0");
        if(len>cbuf.length)
            throw new IllegalArgumentException("Length must be smaller or equal to length of buffer");
        if(off+len>cbuf.length)
            throw new IllegalArgumentException("Would write over end of buffer");
        int count=0;
        int cur=read();
        while(count<len && cur>=0)
        {
            cbuf[off+count]=(char)cur;
            count++;
            cur=read();
        }
        if(cur<0)
            return -1;
        return count;
    }

    /**
    Read a single character. This method will block until a character is available, 
    an I/O error occurs, or the end of the stream is reached. 
    @return the character read, as an integer in the range 0 to 16383 (0x00-0xffff), or -1 if the end of the stream has been reached
    @throws IOException if an I/O error occurs
    */
    public int read() throws IOException
    {
        if(lastChar<0)
            return -1;
        int res=in.read();
        while(res>=0 && atBeginning && (Character.isWhitespace((char)res) || Character.isSpaceChar((char)res)))
            res=in.read();
        atBeginning=false;
        while(res>=0 && Character.isSpaceChar((char)lastChar) && Character.isSpaceChar((char)res))
            res=in.read();
        lastChar=res;
        return res;
    }
    
    /**
    Skip characters. This method will block until some characters are available, 
    an I/O error occurs, or the end of the stream is reached.
    @param n the number of characters to skip 
    @return the number of characters actually skipped
    @throws IOException if an I/O error occurs 
    */
    public long skip(long n) throws IOException	
    {
        // Cannot tell in advance how much we must read, so read characters till count 
        // reached or end of stream found
        int count;
        int cur=0;
        for(count=0;count<n && cur>=0;count++)
            cur=read();
        return count;
    }
}