/*
 * @(#)CanonicalHTMLFilterReader.java
 *
 * Utility classes for managing webpages
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.WWW;

import java.io.Reader;
import java.io.FilterReader;
import java.io.IOException;
import java.util.Hashtable;

/**
Removes all line breaks and multiple space characters within a tag.
Removes all spaces after the starting "&lt;" of a tag.
E. g. "&lt;\n  /  p&gt; will be changed to "&lt;/p&gt;".

@author  Michael Sonntag
@version 1.0, 1.7.2000
*/
public class CanonicalHTMLFilterReader extends FilterReader
{   
    /** <code>true</code> if we are currently within a tag. */
    protected boolean inTag=false;
    /** Single or double quote if we are currently within quotation marks within a tag, space otherwise. */
    protected char inQuotes=' ';    
    /** The buffer of characters already read from the input stream but not yet returned. */
    protected String buffer="";
    /** The last character returned */
    protected int lastChar=0;

    /**
    Creates a new reader.
    @param in the reader to read from 
    */
	public CanonicalHTMLFilterReader(Reader in)
	{
		super(in);
	}
	
	/**
    Tell whether this stream supports the mark() operation (this stream does not).
    @return always <code>false</code>
	*/
	public boolean markSupported()
	{
	    return false;
	}
	
	/**
    Read characters into an array. This method will block until some input is available, 
    an I/O error occurs, or the end of the stream is reached.
    @param cbuf destination buffer 
    @param off the offset in the buffer the first character will be written to
    @param len the number of characters to read
    @return the number of bytes read, or -1 if the end of the stream has been reached
    @throws IOException if an I/O error occurs 
    */
    public int read(char[] cbuf,int off,int len) throws IOException
    {
        if(off<0)
            throw new IllegalArgumentException("Offset must be at least 0");
        if(len>cbuf.length)
            throw new IllegalArgumentException("Length must be smaller or equal to length of buffer");
        if(off+len>cbuf.length)
            throw new IllegalArgumentException("Would write over end of buffer");
        int count=0;
        int cur=read();
        while(count<len && cur>=0)
        {
            cbuf[off+count]=(char)cur;
            count++;
            cur=read();
        }
        return count;
    }

    /**
    Read a single character. This method will block until a character is available, 
    an I/O error occurs, or the end of the stream is reached. 
    @return the character read, as an integer in the range 0 to 16383 (0x00-0xffff), or -1 if the end of the stream has been reached
    @throws IOException if an I/O error occurs
    */
    public int read() throws IOException
    {
        // If we were at eof, don't ask stream, just return -1
        if(lastChar==-1)
            return -1;
        int cur=-1;
        do
        {
            if(cur=='>' && inQuotes==' ')
                inTag=false;
            cur=in.read();
            if(cur<0)
                return -1;
            if(cur=='<' && inQuotes==' ')
                inTag=true;
            // Quotation marks only matter within tags. Outside of them they are ordinary characters
            if(cur=='\"' && inTag)
            {
                if(inQuotes=='\"')
                    inQuotes=' ';
                else if(inQuotes==' ')
                    inQuotes='\"';
            }
            if(cur=='\'' && inTag && inQuotes==' ')
            {
                if(inQuotes=='\'')
                    inQuotes=' ';
                else if(inQuotes==' ')
                    inQuotes='\'';
            }
        }while(inTag && lastChar=='<' && (Character.isWhitespace((char)cur) || Character.isSpaceChar((char)cur)));
        lastChar=cur;
        return cur;
    }
       
    /**
    Tell whether this stream is ready to be read.
    @return <code>true</code> if the next <code>read()</code> is guaranteed not to block for input,
            <code>false</code> otherwise. Note that returning <code>false</code> does not guarantee 
            that the next read will block.
    @throws IOException if an I/O error occurs 
    */
    public boolean ready() throws IOException
    {   // This is correct, but there may be additional cases where we could return true
        return buffer.length()>0;
    }

    /**
    Skip characters. This method will block until some characters are available, 
    an I/O error occurs, or the end of the stream is reached.
    @param n the number of characters to skip 
    @return the number of characters actually skipped
    @throws IOException if an I/O error occurs 
    */
    public long skip(long n) throws IOException	
    {
        // Cannot tell in advance how much we must read, so read characters till count 
        // reached or end of stream found
        int count;
        int cur=0;
        for(count=0;count<n && cur>=0;count++)
            cur=read();
        return count;
    }
}