/*
 * @(#)ConvertHTMLCharFilterReader.java
 *
 * Utility classes for managing webpages
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.WWW;

import java.io.Reader;
import java.io.FilterReader;
import java.io.IOException;
import java.util.Hashtable;

/**
Converts the special characters of HTML (&amp;nbsp; = " ", &amp;auml; = "", ...) and types of 
"&amp;#160;" (Only decimal specification allowed, hex will be ignored) to normal Java characters.

@author  Michael Sonntag
@version 1.0, 1.7.2000
*/
public class ConvertHTMLCharFilterReader extends FilterReader
{
    /**
    A table containing all the special characters as keys and their associated cleartext as the value.
    The value may be an empty string.
    <p>Currently initialized with the following characters:<br>
    &quot;&amp;nbsp;&quot;, &quot;&amp;auml;&quot;, &quot;&amp;ouml;&quot;, &quot;&amp;uuml;&quot;,
    &quot;&amp;Auml;&quot;, &quot;&amp;Oml;&quot;, &quot;&amp;Uuml;&quot;, &quot;&amp;szlig;&quot;,
    &quot;&amp;amp;&quot;, &quot;&amp;lt;&quot;, &quot;&amp;gt;&quot;, &quot;&amp;quot;&quot;
    </p>
    */
    protected static Hashtable specialChars=new Hashtable();   
    static
    {
        specialChars.put("&nbsp;"," ");
        specialChars.put("&auml;","");
        specialChars.put("&uuml;","");
        specialChars.put("&ouml;","");
        specialChars.put("&Auml;","");
        specialChars.put("&Uuml;","");
        specialChars.put("&Ouml;","");
        specialChars.put("&szlig;","");
        specialChars.put("&amp;","&");
        specialChars.put("&lt;","<");
        specialChars.put("&gt;",">");
        specialChars.put("&quot;","\"");
        specialChars.put("&euro;","&#8364");
    }
    
    /** The buffer of characters already read from the input stream but not yet returned. */
    protected String buffer="";

    /**
    Creates a new reader.
    @param in the reader to read from 
    */
	public ConvertHTMLCharFilterReader(Reader in)
	{
		super(in);
	}
	
	/**
    Tell whether this stream supports the mark() operation (this stream does not).
    @return always <code>false</code>
	*/
	public boolean markSupported()
	{
	    return false;
	}
	
	/**
    Read characters into an array. This method will block until some input is available, 
    an I/O error occurs, or the end of the stream is reached.
    @param cbuf destination buffer 
    @param off the offset in the buffer the first character will be written to
    @param len the number of characters to read
    @return the number of bytes read, or -1 if the end of the stream has been reached
    @throws IOException if an I/O error occurs 
    */
    public int read(char[] cbuf,int off,int len) throws IOException
    {
        if(off<0)
            throw new IllegalArgumentException("Offset must be at least 0");
        if(len>cbuf.length)
            throw new IllegalArgumentException("Length must be smaller or equal to length of buffer");
        if(off+len>cbuf.length)
            throw new IllegalArgumentException("Would write over end of buffer");
        int count=0;
        int cur=read();
        while(count<len && cur>=0)
        {
            cbuf[off+count]=(char)cur;
            count++;
            cur=read();
        }
        return count;
    }

    /**
    Read a single character. This method will block until a character is available, 
    an I/O error occurs, or the end of the stream is reached. 
    @return the character read, as an integer in the range 0 to 16383 (0x00-0xffff), or -1 if the end of the stream has been reached
    @throws IOException if an I/O error occurs
    */
    public int read() throws IOException
    {
        int res=-1;
        while(buffer.length()==0)
        {
            res=in.read();
            if(res>=0)
            {
                buffer+=(char)res;
                if(res=='&')
                    convertSpecialChar();
            }
            else
                return -1;
        }
        res=buffer.charAt(0);
        buffer=buffer.substring(1);
        return res;
    }
    
    /**
    Converts special characters that might be at the beginning of the buffer.
    Limited to 10 characters, so the special char must be terminated then or it will not 
    be converted (Should be enough).
    @throws IOException if an I/O error occurs
    */
    protected void convertSpecialChar() throws IOException
    {
        int len=1;
        int specialCharLen=0;
        while(len<11 && specialCharLen==0)
        {
            int res=in.read();
            if(res<0)
                return;
            buffer+=(char)res;
            len++;
            specialCharLen=bufferStartsWithSpecialChar();
        }
        if(specialCharLen>0)
        {
            String spec=buffer.substring(0,specialCharLen);
            String replace=(String)specialChars.get(spec);
            if(replace==null)
                replace=(String)specialChars.get(spec.toLowerCase());
            buffer=specialChars.get(spec)+buffer.substring(specialCharLen);
        }
        if(buffer.startsWith("&#"))
        {
            int pos=2;
            String num="";
            while(pos<buffer.length() && Character.isDigit(buffer.charAt(pos)))
                num+=buffer.charAt(pos++);
            if(pos<buffer.length() && buffer.charAt(pos)==';')
            {   // Found a numerical character
                try
                {
                    int number=Integer.parseInt(num);
                    if(Character.isSpaceChar((char)number) ||number==160) // "&nbsp;" is converted by the HTML classes to this. We want it as a simple space character however
                        buffer=" "+buffer.substring(pos+1);
                    else
                        buffer=""+((char)number)+buffer.substring(pos+1);
                }
                catch(NumberFormatException e)
                {   // Is not a number, so leave the buffer as it is
                    System.err.print("");
                }
            }
        }
    }
    
    /**
    Checks wether the buffer starts with a special character.
    @return 0 if the buffer does not start with a special character or the length of the special character
    */
    private int bufferStartsWithSpecialChar()
    {
        int len=1;
        while(len<=buffer.length())
        {
            if(specialChars.get(buffer.substring(0,len))!=null || specialChars.get(buffer.substring(0,len).toLowerCase())!=null)
                return len;
            len++;
        }
        return 0;
    }
        
    /**
    Tell whether this stream is ready to be read.
    @return <code>true</code> if the next <code>read()</code> is guaranteed not to block for input,
            <code>false</code> otherwise. Note that returning <code>false</code> does not guarantee 
            that the next read will block.
    @throws IOException if an I/O error occurs 
    */
    public boolean ready() throws IOException
    {   // This is correct, but there may be additional cases where we could return true
        return buffer.length()>0;
    }

    /**
    Skip characters. This method will block until some characters are available, 
    an I/O error occurs, or the end of the stream is reached.
    @param n the number of characters to skip 
    @return the number of characters actually skipped
    @throws IOException if an I/O error occurs 
    */
    public long skip(long n) throws IOException	
    {
        // Cannot tell in advance how much we must read, so read characters till count 
        // reached or end of stream found
        int count;
        int cur=0;
        for(count=0;count<n && cur>=0;count++)
            cur=read();
        return count;
    }
}