/*
 * @(#)Cookie.java
 *
 * Utility classes for managing cookies
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.WWW.Cookies;

import java.util.Date;
import java.util.Calendar;
import java.util.TimeZone;
import java.text.SimpleDateFormat;
import java.text.ParsePosition;
import java.io.Serializable;

/**
A single cookie. It can be stored to a stream and supports matching the domain and path.
Additionally stores also when the cookie was last sent (to allow deleting the least recently
used cookie).

@author  Michael Sonntag
@version 1.0, 1.7.2000
*/
public class Cookie implements Serializable
{
    /** The maximum size of the name and the value of a cookie together. */
    public final static int COOKIE_SIZE=4096;

    /**
	The name of the cookie.
	@serial
	*/
    protected String name;
    /**
	The value of the cookie.
	@serial
	*/
    protected String value;
    /**
	The path of the cookie (the subset of URLs for which this cookie is valid).
	@serial
	*/
    protected String path;
    /**
	The domain of the cookie (from where it was sent; hostname).
	@serial
	*/
    protected String domain;
    /**
	The date when to cookie expires (will be deleted after this time).
	@serial
	*/
    protected Date expires;
    /**
	Flag to indicate whether the cookie is secure or not (if <code>true</code> it will
    only be sent when using a secure communications channel).
	@serial
	*/
    protected boolean secure;
    /**
	Whether checking should be done strict (e. g. domain).
	@serial
	*/
    protected boolean strict=false;

    /**
    The time this cookie was last handed out. Necessary for purging when the limits are exceeded.
    Not the oldest, but the least recently used cookies should be thrown away.
    If <code>null</code> it was never sent.
    @serial
    */
    protected Date lastSent=null;

    /**
    Creates a new cookie.
    @param name the name of the cookie (may not contain ";", "," or whitespace)
    @param value the value of the cookie (may not contain ";", "," or whitespace)
    @param path the path of the cookie (<code>null</code> or "" is changed to "/")
    @param domain the domain of the cookie (must not be <code>null</code> or "")
    @param expires the date the cookie expires (in string form as defined in the cookie definition)
    @param secure whether it is a secure cookie or not
	@param strict if <code>true</code> checking will be done strict (currently only for checking the domain).
    @throws IllegalArgumentException if the domain is <code>null</code> or not valid according to the specification
    */
    public Cookie(String name,String value,String path,String domain,String expireString,boolean secure,boolean strict) throws IllegalArgumentException
    {
        if(name.length()+value.length()>COOKIE_SIZE)
        {
            if(name.length()<COOKIE_SIZE)  // Preserve name if possible
                value=value.substring(0,COOKIE_SIZE-name.length());
            else
            {
                /*
                Reduce name to 90% and data to first 10%. Just a (possibly) sensible
                distribution (clients should never expect this however; at least
                according to the standard!)
                */
                name=name.substring(0,(COOKIE_SIZE*10)/9);
                value=value.substring(0,COOKIE_SIZE-(COOKIE_SIZE*10)/9);
            }
        }
        this.name=name;
        this.value=value;
        checkString(name);
        checkString(value);
        this.path=path;
        // Empty path is equal to '/' (for consistency)
        if(path==null || path.length()==0)
            this.path="/";
        this.domain=domain.toLowerCase();
        if(domain==null || domain.length()==0)
            throw new IllegalArgumentException("Domain must be provided");
        checkDomainValidity();
        // Parse expiry date
        this.expires=null;
 		SimpleDateFormat fo=new SimpleDateFormat("EEE, dd-MMM-yyyy HH:mm:ss 'GMT'",java.util.Locale.US);
 		fo.setLenient(true);
        fo.setTimeZone(TimeZone.getTimeZone("GMT"));
        ParsePosition pos=new ParsePosition(0);
        // Set last sent to a date very far in the past
        lastSent=fo.parse("Mon, 01-Jan-1980 00:00:00 GMT",pos);
        if(expireString!=null)
        {
            pos=new ParsePosition(0);
            expires=fo.parse(expireString,pos);
        }
        // if expires==null it is a session cookie and this.expires will stay null
        this.secure=secure;
    }

    /**
    Creates a new cookie with loose checking.
    @param name the name of the cookie
    @param value the value of the cookie
    @param path the path of the cookie
    @param domain the domain of the cookie
    @param expires the date the cookie expires (in string form as defined in the cookie definition)
    @param secure whether it is a secure cookie or not
    @throws IllegalArgumentException if the domain is <code>null</code> or not valid according to the specification
    */
    public Cookie(String name,String value,String path,String domain,String expires,boolean secure) throws IllegalArgumentException
    {
        this(name,value,path,domain,expires,secure,false);
    }

    /**
    Returns a string representation of the cookie. Conforms to the definition for the transmission
    of cookies (i. e. sequence, ";" between elements, ...).
    @return string representation as in the definition
    */
    public String toString()
    {
        String res=name+"="+value+"; path="+path+"; domain="+domain;
        if(expires!=null)
        {
 	    	SimpleDateFormat fo=new SimpleDateFormat("EEE, dd-MMM-yyyy HH:mm:ss 'GMT'",java.util.Locale.US);
    		fo.setTimeZone(TimeZone.getTimeZone("GMT"));
            res+="; "+fo.format(expires);
        }
        if(secure)
            res+="; secure";
        return res;
    }

    /**
    Returns a hash code value for the object. It consists of the name, the path and the domain.
    @return a hash code value for this object
    */
    public int hashCode()
    {
        return name.hashCode()+path.hashCode()+domain.hashCode();
    }

    /**
    Compares two cookies for equality. Compares the name and the path in case-sensitive form and
    the domain in case-insensitive form.
    @param obj the reference object with which to compare
    @return <code>true</code> if this object is the same as the obj argument
    */
    public boolean equals(Object obj)
    {
        if(!(obj instanceof Cookie))
            return false;
        Cookie other=(Cookie)obj;
        return name.equals(other.name) && path.equals(other.path) && domain.equals(other.domain);
    }

    /**
    Checks whether this cookie matches a certain domain. The domain of the cookie is matched
    against the tail of the fully qualified domain name of the host provided as parameter.
    In addition to the specification it must be a full part of the domain (so "acme.com" would
    match both "acme.com" and "www.acme.com" but not "wwwacme.com" or "www.myacme.com").
    @param dom the domain to match
    @return <code>true</code> if the domain of the cookie matches the parameter
    */
    public boolean matchDomain(String dom)
    {
        if(dom==null || dom.length()<1)
            return false;   // Empty domain does not match any cookie
        return dom.endsWith(domain) &&
        // Tail must match. If no exact match, there has to be a dot before the matching end
            (dom.length()==domain.length() || domain.startsWith(".") ||
            dom.charAt(dom.length()-domain.length()-1)=='.');
    }

    /**
    Checks whether this cookie matches a certain path. Checks whether the parameter starts
    with the path or not. The path "/foo" would match "/foobar" and "/foo/bar.html".
    @param p the path to match (<code>null</code> or "" is interpreted as "/" )
    @return <code>true</code> if the path of the cookie matches the parameter
    */
    public boolean matchPath(String p)
    {
        // Take care of empty parameter; this will be like a top-level path ("/")
        if(p==null || p.length()<1)
            p="/";
        return p.startsWith(path);
    }

    /**
    Checks wether the string contains ";", "," or a whitespace (" ", \r, \n, \t, \f).
    @param str the string to check
    @throws IllegalArgumentException if an unallowed character is in the string
    */
    protected void checkString(String str) throws IllegalArgumentException
    {
        if(str.indexOf(";")>=0)
            throw new IllegalArgumentException("May not contain ';'");
        if(str.indexOf(",")>=0)
            throw new IllegalArgumentException("May not contain ','");
        if(str.indexOf(" ")>=0)
            throw new IllegalArgumentException("May not contain ' '");
        if(str.indexOf("\r")>=0)
            throw new IllegalArgumentException("May not contain '\r'");
        if(str.indexOf("\n")>=0)
            throw new IllegalArgumentException("May not contain '\n'");
        if(str.indexOf("\t")>=0)
            throw new IllegalArgumentException("May not contain '\t'");
        if(str.indexOf("\f")>=0)
            throw new IllegalArgumentException("May not contain '\f'");
    }

    /**
    Checks whther the domain provided is valid according to the specification or not.
    If the matching is strict, domains with 2 dots must end in one of the following to be valid:
    com, edu, net, org, gov, mil or int. This does not allow e. g. "www.acme.at"; however these
    names are quite common so the default is non-strict matching (see constructors).
    @return <code>true</code> if the domain in valid
    @see #Cookie(String,String,String,String,String,boolean,boolean)
    @see #Cookie(String,String,String,String,String,boolean)
    */
    protected void checkDomainValidity() throws IllegalArgumentException
    {
        int dot=0;
        int pos=0;
        while(pos<domain.length() && (pos=domain.indexOf(".",pos))>0)
        {
            dot++;
            pos++;
        }
        if(dot<2)
            throw new IllegalArgumentException("Domain must contain at least 2 dots (e. g. www.acme.com)");
        if(dot>2)
            return;
        if(strict)
        {
        // Should be in by standard, but does not work then (many have only 2 dots but are not of these domains (e. g. www.orf.at)!)
            String end=domain.substring(domain.length()-3);
            if(end.equalsIgnoreCase("com") || end.equalsIgnoreCase("edu") || end.equalsIgnoreCase("net"))
                return;
            if(end.equalsIgnoreCase("org") || end.equalsIgnoreCase("gov") || end.equalsIgnoreCase("mil"))
                return;
            if(end.equalsIgnoreCase("int"))
                return;
            throw new IllegalArgumentException("Domain must contain at least 3 dots if not from a special top-level domain (com, edu, net, org, gov, mil, int)");
        }
    }

    /**
    Retrieve whether this cookie has expired or not.
    Expired cookies should be deleted.
    @return <code>true</code> if the cookie has expired
    */
    public boolean hasExpired()
    {
        // Session cookies will never expire for a certain time
        if(expires==null)
            return false;
        Calendar cal=Calendar.getInstance();
        return expires.before(cal.getTime());
    }

    /**
    Retrieve whether this is a session-only cookie or not.
    These should not be written to a stream when a set of cookies is stored.
    @return <code>true</code> if the cookie expires with the session
    */
    public boolean isSessionOnly()
    {
        // Session cookies will expire with the session
        // (and therefore not be saved)
        return expires==null;
    }

    /**
    Retrieve the name of this cookie.
    @return the name of the cookie
    */
    public String getName()
    {
        return name;
    }

    /**
    Retrieve the value of this cookie.
    @return the value of the cookie
    */
    public String getValue()
    {
        return value;
    }

    /**
    Retrieve the path for this cookie.
    @return the path for the cookie
    */
    public String getPath()
    {
        return path;
    }

    /**
    Retrieve the domain for this cookie.
    @return the domain for the cookie
    */
    public String getDomain()
    {
        return domain;
    }

    /**
    Retrieve the expiry date of the cookie.
    @return the date after which the cookie will no longer be handed out or <code>null</code> if it is a session cookie
    */
    public Date getExpiryDate()
    {
        return expires;
    }

    /**
    Retrieve whther this cookie is secure (if yes, it will only be handed out over secure connections).
    @return <code>true</code> if the cookie is marked as secure
    */
    public boolean isSecure()
    {
        return secure;
    }

    /**
    Retrieve the date this cookie was last sent.
    @return the date the cookie was last sent (<code>null</code> if never sent)
    */
    public Date getLastSent()
    {
        return lastSent;
    }

    /**
    Update the time this cookie was last sent: It is set to the current time.
    */
    public void sentNow()
    {
        lastSent=new Date();
    }

}
