/*
 * @(#)CookieStore.java
 *
 * Utility classes for managing cookies
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.WWW.Cookies;

import java.util.Date;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Arrays;
import java.util.StringTokenizer;
import java.util.Enumeration;
import java.net.HttpURLConnection;
import java.net.URL;
import java.io.Serializable;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;

/**
A store for cookies. Adheres to the standard (number of cookies, size, ...) as defined in the
<a href="http://home.netscape.com/newsref/std/cookie_spec.html">Cookie definition</a>
(http://home.netscape.com/newsref/std/cookie_spec.html).
<p>Sample code for usage (simplified and without exception handling):<br>
<pre>
private CookieStore cookies=new CookieStore();
...
URL url=new URL(...);
HttpURLConnection con=(HttpURLConnection)url.openConnection();
con.setRequestProperty(...);
<strong>cookies.sendCookies(con,false);</strong>
con.connect();
PrintWriter out=new PrintWriter(con.getOutputStream());
out.println(...);
out.close();
if(con.getResponseCode()!=200)
    throw new ...Exception();
<strong>cookies.parseCookies(con);</strong>
BufferedReader in=new BufferedReader(new InputStreamReader(con.getInputStream()));
String line=in.readLine();
...
</pre></p>

@author  Michael Sonntag
@version 1.0, 1.7.2000
@see FIM.Util.WWW.Cookies.CookieEnumeration
*/
public class CookieStore implements Serializable
{
    /** The maximum number of cookies stored in total. */
    public final static int TOTAL_COOKIES=300;
    /**
    The maximum number of cookies stored from one domain (exact match: "www.acme.com"
    and "acme.com" receive each 20 cookies)
    */
    public final static int COOKIES_PER_DOMAIN=20;

    /**
    Store for the cookies. Key is domain+";"+path+";"+name .
    @serial
    */
    protected Hashtable table=new Hashtable(TOTAL_COOKIES+1);

    /**
    Creates a new empty store for cookies.
    */
    public CookieStore()
    {
    }
    
    /**
    Removes all cookies from the store.
    */
    public void clearCookies()
    {
        table.clear();
    }

    /**
    Retrieves all cookies from a <code>HttpURLConnection</code> and stores them.
    @param con the connection to retrieve the cookies from. Must already be connected.
    */
    public void parseCookies(HttpURLConnection con)
    {
        String name;
        String value;
        String path;
        String expires;
        String domain;
        boolean secure;
        int pos=1;
        URL url=con.getURL();
        // Walk through all headers
        String field=con.getHeaderField(pos);
        while(field!=null)
        {
            // Only set-cookie requests are handled
            if(con.getHeaderFieldKey(pos)!=null && con.getHeaderFieldKey(pos).equalsIgnoreCase("Set-Cookie"))
            {
                // Reset values
                name="";
                value="";
                path=null;
                expires=null;
                domain=null;
                secure=false;
                // Break up in parts
                StringTokenizer tok=new StringTokenizer(field,";",false);
                if(!tok.hasMoreTokens())
                    throw new IllegalArgumentException("At least name and value must be present!");
                String nameval=tok.nextToken().trim();
                name=nameval.substring(0,nameval.indexOf("="));
                value=nameval.substring(nameval.indexOf("=")+1);
                // Handle all other attributes (if any); order is unimportant
                while(tok.hasMoreTokens())
                {
                    String token=tok.nextToken().trim(); // Remove leading space through trim()
                    if(token.toLowerCase().startsWith("expires="))
                        expires=token.substring(token.indexOf("=")+1);
                    if(token.toLowerCase().startsWith("path="))
                        path=token.substring(token.indexOf("=")+1);
                    if(token.toLowerCase().startsWith("domain="))
                        domain=token.substring(token.indexOf("=")+1);
                    if(token.equalsIgnoreCase("secure"))
                        secure=true;
                }
                if(path==null)
                    path=url.getFile(); // Default path is the same as the document
                if(domain==null)
                    domain=url.getHost();   // Default domain is the hostname of the other side
                else
                {
                    // Only hosts within the specified domain can set a cookie for a domain
                    String host=url.getHost();
                    if(!domain.toLowerCase().equals(host.toLowerCase()))
                        domain=host;    // Correct
                }
                Cookie c=new Cookie(name,value,path,domain,expires,secure);
                setCookie(c);
                // Not very efficient to call enforceLimits(domain) for each cookie (done in setCookie),
                // but easier and the total number of cookies is rather small anyway.
            }
            // Get next header
            pos++;
            field=con.getHeaderField(pos);
        }
    }

    /**
    Sends all matching cookies to a <code>HttpURLConnection</code>.
    @param con the connection to retrieve the cookies from. <strong>May not be connected!</strong>
    @param secureConn whether this is a secure connection (cookies marked as secure will be sent) or not
    */
    public void sendCookies(HttpURLConnection con,boolean secureConn)
    {
        String cookies=null;
        URL url=con.getURL();
        Vector toSend=new Vector();
        CookieEnumeration enum=new CookieEnumeration(table.elements(),url.getHost(),url.getFile());
        while(enum.hasMoreElements())
        {
            Cookie c=enum.nextCookie();
            if(c.hasExpired())
                table.remove(c.domain+";"+c.path+";"+c.name);
            else if(!c.isSecure() || secureConn)
            {
                toSend.add(c);  // Add to list of cookies to send
                c.sentNow();    // Update least used date
            }
        }
        if(toSend.size()>0)
        {
            // Sort cookies according to path specific (more specific are in the from of the list)
            Cookie[] sendArray=new Cookie[toSend.size()];
            toSend.copyInto(sendArray);
            Arrays.sort(sendArray,new CookiePathComparator());
            // Create response string
            for(int i=0;i<sendArray.length;i++)
                if(cookies==null)
                    cookies=sendArray[i].toString();
                else
                    cookies+="; "+sendArray[i].toString();
            con.setRequestProperty("Cookie",cookies);
        }
    }

    /**
    Removes all expired cookies form the store and enforces the limit of cookies per domain and the total maximum.
    If exceeded, the least recently used cookie from a domain will be deleted (only one!).
    If the total limit is exceeded, the least recently used cookie of all will be deleted (only one!).
    @param domain the domain to check the limit for
    */
    protected void enforceLimits(String domain)
    {
        Vector expired=new Vector();
        Cookie leastRecentlyUsedDomain=null;
        Cookie leastRecentlyUsedAll=null;
        int count=0;    // Number of cookies in this domain
        Enumeration enum=table.elements();
        while(enum.hasMoreElements())
        {
            Cookie c=(Cookie)enum.nextElement();
            if(c.hasExpired())
                expired.add(c); // Expired cookies are always removed
            else if(c.matchDomain(domain))
            {
                count++;    // One more from this domain
                if(leastRecentlyUsedDomain==null || c.getLastSent().before(leastRecentlyUsedDomain.getLastSent()))
                    leastRecentlyUsedDomain=c;  // Calculate least recently used from domain
            }
            else if(leastRecentlyUsedAll==null || c.getLastSent().before(leastRecentlyUsedAll.getLastSent()))
                leastRecentlyUsedAll=c;  // Calculate least recently used from all domains
        }
        // 2 steps necessary as Hashtable enumerator is not resistant against deletions
        enum=expired.elements();
        while(enum.hasMoreElements())
        {
            Cookie c=(Cookie)enum.nextElement();
            table.remove(c.domain+";"+c.path+";"+c.name);
        }
        if(count>COOKIES_PER_DOMAIN)
            // Throw out the least recently used cookie
            table.remove(leastRecentlyUsedDomain.domain+";"+leastRecentlyUsedDomain.path+";"+leastRecentlyUsedDomain.name);
        if(table.size()>TOTAL_COOKIES)
            // Throw out the least recently used cookie of all if still over limit
            if(leastRecentlyUsedDomain.getLastSent().before(leastRecentlyUsedAll.getLastSent()))
                // Will never be removed twice. If removed above, the total limit can never be
                // exceeded because we call this for each cookie we add!
                table.remove(leastRecentlyUsedDomain.domain+";"+leastRecentlyUsedDomain.path+";"+leastRecentlyUsedDomain.name);
            else
                table.remove(leastRecentlyUsedAll.domain+";"+leastRecentlyUsedAll.path+";"+leastRecentlyUsedAll.name);
    }

    /**
    Stores all non-expired and non-session-only cookies to a stream.
    @param out the stream to write to
    @throws IOException if an IO error occured during writing
    */
    private void writeObject(ObjectOutputStream out) throws IOException
    {
        // Only write cookies, which are not session-only cookies
        Vector toStore=new Vector();
        Enumeration enum=table.elements();
        while(enum.hasMoreElements())
        {
            Cookie c=(Cookie)enum.nextElement();
            if(!c.hasExpired() && !c.isSessionOnly())
                toStore.add(c); // Expired and session-only cookies are not stored
        }
        out.writeInt(toStore.size());
        for(int i=0;i<toStore.size();i++)
            out.writeObject(toStore.elementAt(i));
    }

    /**
    Restores all cookies from a stream.
    @param in the stream to read from
    @throws IOException if an IO error occured during reading
    @throws ClassNotFoundException if a class could not be restored
    */
    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException
    {
        int count=in.readInt();
        table=new Hashtable();
        for(int i=0;i<count;i++)
        {
            Cookie c=(Cookie)in.readObject();
            table.put(c.domain+";"+c.path+";"+c.name,c);
        }
    }

    /**
    Returns all cookies which match a specified domain and path. If a parameter is <code>null</code>,
    all cookies will match it.
    @param domain the domain the returned cookies are from (if <code>null</code> ignored)
    @param path the path the returned cookies are valid for (if <code>null</code> ignored)
    @return an enumeration of all matching cookies
    */
    public CookieEnumeration getCookies(String domain,String path)
    {
        return new CookieEnumeration(table.elements(),domain,path);
    }

    /**
    Retrieve a certain cookie from the store. All parameters must match exactly.
    @param name the name of the cookie
    @param domain the domain the cookie is from
    @param path the path the cookie is valid for (if "" or <code>null</code>, "/" is used)
    @return the cookie or <code>null</code> if not found
    */
    public Cookie getCookie(String name,String domain,String path)
    {
        if(name==null || name.length()<1)
            throw new IllegalArgumentException("Name must be provided");
        if(domain==null || domain.length()<1)
            throw new IllegalArgumentException("Domain must be provided");
        if(path==null || path.length()<1)
            path="/";
        return (Cookie)table.get(domain+";"+path+";"+name);
    }

    /**
    Puts a certain cookie into the store and enforces the limits. If it is already expired,
    an existing cookie with matching name, domain and path is removed from the store.
    @param c the cookie to store
    @return the previous cookie with the same name, domain and path or <code>null</code> if new
    */
    public Cookie setCookie(Cookie c)
    {
        if(c==null)
            throw new IllegalArgumentException("Cookie must be provided!");
        Cookie old=(Cookie)table.get(c.domain+";"+c.path+";"+c.name);
        if(c.hasExpired())
            table.remove(c.domain+";"+c.path+";"+c.name);
        else
        {
            table.put(c.domain+";"+c.path+";"+c.name,c);
            enforceLimits(c.domain);
        }
        return old;
    }
}

// TODO: Sort returned cookies according to specification

