/*
 * @(#)ActionElement.java
 *
 * Classes for handling HTML-Forms without UI
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.WWW.Form;

import javax.swing.text.html.HTMLDocument;
import javax.swing.text.Element;

/**
An element of a HTML-form which produced an action if invoked (e. g. resetting the form or submitting it).

@author  Michael Sonntag
@version 1.0, 1.7.2000
*/
public abstract class ActionElement extends FormElement
{
    /**
    The value of this element.
    @serial
    */
    private String value=null;

    /**
    Creates a new action element.
    @param doc the document this element belongs to
    @param parent the parent of this element
    @param startOffs start offset of the element in the document
    @param endOffs end offset of the element in the document
    @param name the name of the element
    @param value the value of the element or <code>null</code> if none exists
    */
    public ActionElement(HTMLDocument doc,Element parent,int startOffs,int endOffs,String name,String value)
    {
        super(doc,parent,startOffs,endOffs,name);
        this.value=value;
    }
    
    /**
    Resets the element to its initial state.
    Empty implementation, as action elements usually have no state.
    */
    public void reset()
    {
    }
    
    /**
    Placeholder for the action to do when the element is invoked.
    Subclasses must provide implementations.
    */
    public abstract void doAction();
    
    /**
    Returns this element in encoded (<code>application/x-www-form-urlencoded</code>) form.
    If the value is <code>null</code>, only the name is submitted.
    @return the element in encoded form
    */
    public String getEncoded()
    {
        if(value==null)
            return java.net.URLEncoder.encode(getName());
        return java.net.URLEncoder.encode(getName())+"="+java.net.URLEncoder.encode(value);
    }
    
    /**
    Retrieve the value of the element.
    @return the value of the element
    */
    public String getValue()
    {
        return value;
    }
    
    /**
    Matches this element against a number of strings and returns a measure for matching them.
    0=no match, 100=perfect match.<br>
    Extension to <code>FormElement</code>: Value match=80 exact, 70 substring
    @param matchStr an array of strings which are matched against this element
    @return measure how much this element matches the parameter
    */
    public int match(String[] matchStr)
    {
        int curMax=super.match(matchStr);
        for(int i=0;i<matchStr.length;i++)
        {
            int cur=0;
            if(value!=null && value.equalsIgnoreCase(matchStr[i]))
                cur=80;
            if(value!=null && value.toLowerCase().indexOf(matchStr[i].toLowerCase())>=0)
                cur=70;
            if(cur>curMax)
                curMax=cur;
        }
        return curMax;
    }    
    
    /**
    Returns a string representation of the action element.
    @return string representation of the form element
    */
    public String toString()
    {
        return "Action element ("+getName()+":"+value+")";
    }

}