/*
 * @(#)ElementGroup.java
 *
 * Classes for handling HTML-Forms without UI
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.WWW.Form;

import javax.swing.text.html.HTMLDocument;
import javax.swing.text.Element;
import java.util.Vector;
import java.util.Enumeration;

/**
A group of selectable elements.

@author  Michael Sonntag
@version 1.0, 1.7.2000
*/
public class ElementGroup extends FormElement
{
    /**
    The contained elements.
    @serial
    */
    protected Vector containedElems=null;
    
    /**
    Create a new and empty selection group.
    @param doc the document this element belongs to
    @param parent the parent of this element
    @param name the name of the group
    */
    public ElementGroup(HTMLDocument doc,Element parent,String name)
    {
        super(doc,parent,0,0,name);
        containedElems=new Vector();
    }
    
    /**
    Gets the number of contained elements.
    @return the number of child elements >= 0
    */
    public int getContainedElementCount()
    {
        return containedElems.size();
    }

    /**
    Fetches the contained element at the given index.
    @param index the specified index >= 0
    @return the child element
    */
    public Element getContainedElement(int index)
    {
        if(index<containedElems.size())
            return (Element)containedElems.elementAt(index);
        return null;
    }
    
    /**
    Adds an element to the group at the end.
    @param elem the element to add
    */
    public void addContainedElement(SelectableElement elem)
    {
        if(!containedElems.contains(elem))
            containedElems.add(elem);
    }

    /**
    Is this element a leaf element (<code>true</code> if no elements contained at the moment)?
    @return true if a leaf element else false
    */
    public boolean isLeaf()
    {
        return containedElems.size()==0;
    }
    
    /**
    Resets all contained elements to their initial state.
    */
    public void reset()
    {
        Enumeration enum=containedElems.elements();
        while(enum.hasMoreElements())
            ((SelectableElement)enum.nextElement()).reset();
    }
    
    /**
    Retrieve this group in encoded form. Contains all (and only) selected elements belonging to 
    this group in the form <name>=<value>. If no elements are selected, an empty string is returned
    @return the selected elements of the group in encoded form
    */
    public String getEncoded()
    {
        String res="";
        Enumeration enum=containedElems.elements();
        while(enum.hasMoreElements())
        {
            SelectableElement elem=(SelectableElement)enum.nextElement();
            if(elem.isSelected())
            {
                if(res.length()>0)
                    res+="&";
                res+=java.net.URLEncoder.encode(elem.getName())+"="+java.net.URLEncoder.encode(elem.getValue());
            }
        }
        return res;
    }
    
    /**
    Retrieve the number of elements in this group that are selected.
    @return number of selected elements
    */
    public int selectedCount()
    {
        int count=0;
        Enumeration enum=containedElems.elements();
        while(enum.hasMoreElements())
        {
            if(((SelectableElement)enum.nextElement()).isSelected())
                count++;
        }
        return count;
    }
    
    /**
    Returns a string representation of the element group.
    @return string representation of the form element
    */
    public String toString()
    {
        String res="Element group with "+containedElems.size()+" elements.";
        for(int i=0;i<containedElems.size();i++)
            res+="\nElement "+i+": "+containedElems.elementAt(i).toString();
        return res;
    }
}