/*
 * @(#)FormElement.java
 *
 * Classes for handling HTML-Forms without UI
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.WWW.Form;

import javax.swing.text.Document;
import javax.swing.text.html.HTMLDocument;
import javax.swing.text.Element;
import javax.swing.text.AttributeSet;
import javax.swing.text.MutableAttributeSet;
import javax.swing.text.SimpleAttributeSet;
import java.util.Vector;
import java.io.Serializable;

/**
An element of a HTML-form. It is connected to a HTMLDocument, has a start and end offset,
a parent element, a label and a name. It can contain a number of children and attributes.

@author  Michael Sonntag
@version 1.0, 1.7.2000
*/
public abstract class FormElement implements Element,Serializable
{
    /**
    The document this element belongs to.
    @serial
    */
    private HTMLDocument doc=null;
    /**
    Offset of the start of this element in the document.
    @serial
    */
    private int startOffs=0;
    /**
    Offset of the end of this element in the document.
    @serial
    */
    private int endOffs=-1;
    /**
    The pointer to the parent element of this element. Top-level elements point to the form thei are in.
    @serial
    */
    private Element parent=null;
    /**
    The index of the last child element searched for.
    @serial
    */
    private int lastIndex=-1;
    /**
    The label of the element. May be <code>null</code>.
    @serial
    */
    private String label=null;
    /**
    The name of the element (case-insensitive).
    @serial
    */
    private String name=null;

    /**
    The children of this element, if any.
    @serial
    */
    protected Vector children=null;
    /**
    The attributes of this element. Always exists, but may be empty.
    @serial
    */
    protected MutableAttributeSet attributes=null;
    
    /**
    Creates a new form element.
    @param doc the document this element belongs to
    @param parent the parent of this element
    @param startOffs start offset of the element in the document
    @param endOffs end offset of the element in the document
    @param name the name of the element
    */
    public FormElement(HTMLDocument doc,Element parent,int startOffs,int endOffs,String name)
    {
        if(doc==null)
            throw new IllegalArgumentException("HTMLDocument must be provided");
        this.doc=doc;
        if(parent==null)
            throw new IllegalArgumentException("Parent must be provided");
        this.parent=parent;
        if(startOffs<0)
            throw new IllegalArgumentException("Start offset must be >= 0");
        this.startOffs=startOffs;
        if(startOffs>endOffs)
            throw new IllegalArgumentException("Start offset must be <= end offset");
        this.endOffs=endOffs;
        if(name==null || name.length()<1)
            throw new IllegalArgumentException("Name must be provided");
        this.name=name;
        children=new Vector();
        attributes=new SimpleAttributeSet();        
    }
    
    /**
    Fetches the document associated with this element.
    @return the document
    */
    public Document getDocument()
    {
        return doc;
    }

    /**
    Fetches the parent element.  Will usually be the form.
    @return the parent element
    */
    public Element getParentElement()
    {
        return parent;
    }

    /**
    Fetches the name of the element. 
    @return the element name
    */
    public String getName()
    {
        return name;
    }

    /**
    Fetches the collection of attributes this element contains.
    @return the attributes for the element
    */
    public AttributeSet getAttributes()
    {
        return attributes;
    }

    /**
    Fetches the offset from the beginning of the document that this element begins at. 
    @return the starting offset >= 0
    */
    public int getStartOffset()
    {
        return startOffs;
    }

    /**
    Fetches the offset from the beginning of the document that this element ends at. 
    @return the ending offset >= 0
    */
    public int getEndOffset()
    {
        return endOffs;
    }

    /**
    Gets the child element index closest to the given offset.
    The offset is specified relative to the begining of the document.
    @param offset the specified offset >= 0
    @return the element index >= 0
    */
    public int getElementIndex(int offset)
    {
        int nchildren=children.size();
	    int index;
	    int lower=0; 
	    int upper=nchildren-1;
	    int mid=0;
	    int p0=getStartOffset();
	    int p1;
	    if (nchildren == 0)
    		return 0;
	    if (offset>=getEndOffset())
    		return nchildren-1;
	    // see if the last index can be used.
	    if((lastIndex>=lower) && (lastIndex<=upper)) 
	    {
    		Element lastHit=(Element)children.elementAt(lastIndex);
	    	p0=lastHit.getStartOffset();
		    p1=lastHit.getEndOffset();
    		if((offset>=p0) && (offset<p1)) 
	    	    return lastIndex;
    		// last index wasn't a hit, but it does give useful info about
	    	// where a hit (if any) would be.
		    if (offset<p0)
    		    upper=lastIndex;
	    	else
		        lower=lastIndex;
	    }
	    while(lower<=upper) 
	    {
		    mid=lower+((upper-lower)/2);
		    Element elem=(Element)children.elementAt(mid);
		    p0=elem.getStartOffset();
		    p1=elem.getEndOffset();
		    if((offset>=p0) && (offset<p1))
		    {
	    	    // found the location
    		    index=mid;
		        lastIndex=index;
		        return index;
		    }
		    else if(offset<p0)
		    {        
		        upper=mid-1;
    		}
    		else
    		{
		        lower=mid+1;
		    }
	    }
	    // didn't find it, but we indicate the index of where it would belong
	    if (offset<p0)
		    index=mid;
	    else 
    		index=mid+1;
	    lastIndex=index;
	    return index;   
    }

    /**
    Gets the number of child elements contained by this element.
    If this element is a leaf, a count of zero is returned.
    @return the number of child elements >= 0
    */
    public int getElementCount()
    {
        return children.size();
    }

    /**
    Fetches the child element at the given index.
    @param index the specified index >= 0
    @return the child element
    */
    public Element getElement(int index)
    {
        if(index<children.size())
            return (Element)children.elementAt(index);
        return null;
    }

    /**
    Is this element a leaf element?
    @return <code>true</code> if a leaf element else <code>false</code>
    */
    public boolean isLeaf()
    {
        return children.size()==0;
    }
    
    /**
    Fetch the label of this element.
    If none set, <code>null</code> is returned.
    @return label of the element
    */
    public String getLabel()
    {
        return label;
    }
    
    /**
    Set the label for this element.
    Removing by calling with <code>null</code> as parameter
    @param label the new label
    */
    protected void setLabel(String label)
    {
        this.label=label!=null?label.trim():null;
    }
        
    /**
    Resets this element to its initial state.
    */
    public abstract void reset();
    
    /**
    Returns this element in encoded (<code>application/x-www-form-urlencoded</code>) form.
    @return the element in encoded form or an empty string ("") if it shall not be passed to the server
    @see java.net.URLEncoder
    */
    public abstract String getEncoded();
    
    /**
    Matches this element against a number of strings and returns a measure for matching them.
    0=no match, 100=perfect match.<br>
    May be expanded to similarity-matching in the future (similar sounding words, same word-stem, etc.).
    <p>Strategy for matching:
    <ol>
    <li>Label of the element: 100 for exact match, 90 for substring (case-insensitive)</li>
    <li>Name of the lement: 95 for exact match, 85 for substring (case-insensitive)</li>
    <li>Text in front of the element: 70-<numer of characters away from element> (case-insensitive)</li>
    <li>Text after the element: 50-<numer of characters away from element> (case-insensitive)</li>
    </ol>
    The maximum value of all string provided will be returned;
    </p>
    @param matchStr an array of strings which are matched against this element
    @return measure how much this element matches the parameter
    */
    public int match(String[] matchStr)
    {
        int curMax=0;
        for(int i=0;i<matchStr.length;i++)
        {
            int cur=0;
            if(getLabel()!=null && getLabel().equalsIgnoreCase(matchStr[i]))
                cur=100+matchStr[i].length()/2;
            if(getLabel()!=null && getLabel().toLowerCase().indexOf(matchStr[i].toLowerCase())>=0)
                cur=90+matchStr[i].length()/2;
            if(getName()!=null && getName().equalsIgnoreCase(matchStr[i]))
                cur=95+matchStr[i].length()/2;
            if(getName()!=null && getName().toLowerCase().indexOf(matchStr[i].toLowerCase())>=0)
                cur=85+matchStr[i].length()/2;
            String pre=((HTMLForm)getParentElement()).getPreText(this,80).toLowerCase();
            int pos=pre.lastIndexOf(matchStr[i].toLowerCase());
            if(pos>=0)
            {
                cur=70-(pre.length()-pos-matchStr[i].length());
            }
            String post=((HTMLForm)getParentElement()).getPreText(this,50).toLowerCase();
            pos=post.indexOf(matchStr[i].toLowerCase());
            if(pos>=0)
                cur=50-pos;
            if(cur>curMax)
                curMax=cur;
        }
        return curMax;
    }    
    
    /**
    Returns a string representation of the form element.
    @return string representation of the form element
    */
    public String toString()
    {
        return "FormElement ("+getName()+")";
    }
}