/*
 * @(#)HTMLForm.java
 *
 * Classes for handling HTML-Forms without UI
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.WWW.Form;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.io.Serializable;
import java.io.IOException;
import java.net.URL;
import java.net.MalformedURLException;
import java.text.ParseException;
import javax.swing.text.StyleConstants;
import javax.swing.text.Document;
import javax.swing.text.Element;
import javax.swing.text.ElementIterator;
import javax.swing.text.AttributeSet;
import javax.swing.text.MutableAttributeSet;
import javax.swing.text.BadLocationException;
import javax.swing.text.html.HTMLDocument;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.html.HTML;
import javax.swing.text.html.Option;
import javax.swing.text.html.HTMLEditorKit;
import FIM.Util.WWW.WWWPageUtils;
import FIM.Util.WWW.Cookies.CookieStore;

/**
Models a HTML form. The internal support of <code>HTMLDocument</code> ONLY works when using a GUI, so
we cannot use it for handling it directly. Builds up on <code>HTMLDocument</code>.<br>
Sample code for parsing a certain form and submitting it:<br>
<pre>
    StringReader in=new StringReader(srcPageAsString);
    HTMLEditorKit editorKit=new HTMLEditorKit();
    HTMLDocument doc=(HTMLDocument)editorKit.createDefaultDocument();
    doc.putProperty("IgnoreCharsetDirective",new Boolean(true));
    editorKit.read(in,doc,0);
    HTMLForm form=new HTMLForm(doc,nameOfTheFormToParse);
    form.setBaseURL(baseURLOfTheForm);
    ......
    // Modify the contained elements
    ......
    String resultPage=form.doSubmit(null);
</pre>
<p>The elements of the form are stored in a flat structure and groups (e. g. radiobuttons or checkboxes
with the same name) are stored separartely.</p>
<p>>See the <a href="http://www.w3.org/TR/html4/">HTML 4.01 specification</a></p>

@author  Michael Sonntag
@version 1.0, 1.7.2000
*/
public class HTMLForm implements Element,Serializable
{
    /**
    Counter for creating unique names for elements, which have no NAME or ID property.
    @serial
    */
    private int unnamedCount=0;
    /**
    The index at which the last child we searched for was found (For optimizing the search).
    @serial
    */
    private int lastIndex=-1;
    /**
    The document this form belongs to.
    @serial
    */
    private HTMLDocument doc=null;
    /**
    The start offset of the form in the document.
    @serial
    */
    private int startOffs=0;
    /**
    The end offset of the form in the document.
    @serial
    */
    private int endOffs=-1;
    /**
    The name of the form.
    @serial
    */
    protected String formName=null;
    /**
    The elements of the form. They are stored in the order they are encountered.
    Elements are stored flat, so no hierarchy takes place. Grouping is done through the
    separate groups.
    Only those elements providing data (at least possibly) are included.
    @serial
    */
    protected Vector formElements=new Vector();
    /**
    Attributes of the form.
    @serial
    */
    protected MutableAttributeSet formAttributes=new SimpleAttributeSet();
    /**
    The groups of elements within the form. E. g. all checkboxes with thesame name or all
    radiobuttons belonging together or a group of elements to select (SELECT element).
    @serial
    */
    protected Hashtable groups=new Hashtable();

    /**
    Creates a new HTML form by extracting the information from a document. If multiple forms are in the
    document, the name of the form must be provided else the first form will be returned.
    @param doc the document in which the form is contained
    @param formName the value of the "name" attribute of the desired form. If <code>null</code>, the first form encountered is parsed
    @throws ParseException if an error occured during parsing (wrong element names, ...)
    */
    public HTMLForm(HTMLDocument doc,String formName) throws ParseException
    {
        this.doc=doc;
        this.formName=formName;
        // Get base URL if present
        if(doc.getBase()!=null)
            formAttributes.addAttribute(HTML.Tag.BASE,doc.getBase());
        HTMLDocument.Iterator iter=doc.getIterator(HTML.Tag.FORM);
        // Works only, as "FORM" is NOT a BLOCK tag (those are not implemented!)
        while(endOffs<0 && iter.isValid())
        {
            AttributeSet set=iter.getAttributes();
            formName=(String)set.getAttribute(HTML.Attribute.ID);
            if(formName==null)  // Fall back to NAME tag
                formName=(String)set.getAttribute(HTML.Attribute.NAME);
            if(formName!=null && formName.equalsIgnoreCase(this.formName))
            {
                parseForm(iter);// Matching form found
            }
            iter.next();
        }
        if(endOffs<0)
        {
            throw new IllegalArgumentException("Form with name "+formName+" not found!");
        }
    }

    /**
    Parses a certain form from a document by its position number (first form=1).
    @param doc the HTML document to extract the form from
    @param nthForm number of the form to return from the document
    @throws ParseException if an error occured during parsing (wrong element names, ...)
    */
    public HTMLForm(HTMLDocument doc,int nthForm) throws ParseException
    {
        int count=1;
        this.doc=doc;
        this.formName=formName;
        // Get base URL if present
        if(doc.getBase()!=null)
            formAttributes.addAttribute(HTML.Tag.BASE,doc.getBase());
        HTMLDocument.Iterator iter=doc.getIterator(HTML.Tag.FORM);
        // Works only, as "FORM" is NOT a BLOCK tag (those are not implemented!)
        while(count++<nthForm && iter.isValid())
            iter.next();
        // Found it or document ended?
        if(iter.isValid())
            parseForm(iter);
        if(endOffs<0)
            throw new IllegalArgumentException("Form with name "+formName+" not found!");
    }

    /**
    Parses a single form. Creates all the subelements according to the content.
    @param iter the form to parse (in <code>HTMLDocument</code> style)
    @throws ParseException if an error occured during parsing (wrong element names, ...)
    */
    protected void parseForm(HTMLDocument.Iterator iter) throws ParseException
    {        
        startOffs=iter.getStartOffset();
        endOffs=iter.getEndOffset();
        formAttributes.addAttributes(iter.getAttributes());
        // Parse formElements
        ElementIterator it=new ElementIterator(doc);
        Element elem=null;
        while((elem=it.next())!=null)
        {
            if(elem.getStartOffset()>=startOffs && elem.getEndOffset()<=endOffs)
            {
                String elementName=null;
                AttributeSet set=elem.getAttributes();
                elementName=(String)set.getAttribute(HTML.Attribute.ID);
                if(elementName==null)  // Fall back to NAME tag
                    elementName=(String)set.getAttribute(HTML.Attribute.NAME);
                if(elementName==null)   // Still not found? Synthesize a name for the element
                    elementName="FormElement"+(++unnamedCount);
                if((set.getAttribute(StyleConstants.NameAttribute) instanceof HTML.Tag)
                &&((HTML.Tag)set.getAttribute(StyleConstants.NameAttribute)).equals(HTML.Tag.INPUT))
                {
                    // This input is in the current form
                    String type=(String)set.getAttribute(HTML.Attribute.TYPE);
                    String value=(String)set.getAttribute(HTML.Attribute.VALUE);
                    boolean checked=set.getAttribute(HTML.Attribute.CHECKED)!=null;
                    String alt=(String)set.getAttribute(HTML.Attribute.ALT);
                    String src=(String)set.getAttribute(HTML.Attribute.SRC);
                    int maxLen=-1;
                    String len=(String)set.getAttribute(HTML.Attribute.MAXLENGTH);
                    if(len!=null)
                    {
                        try
                        {
                            maxLen=Integer.parseInt(len);
                        }
                        catch(NumberFormatException e)
                        {   // Ignore it then
                            System.err.print("");
                        }
                    }
                    if(type!=null && type.equalsIgnoreCase("text"))
                    {
                        InputTextElement el=new InputTextElement(doc,this,elem.getStartOffset(),elem.getEndOffset(),elementName,value!=null?value:"");
                        if(maxLen>=0)
                            el.setMaxLen(maxLen);
                        formElements.add(el);
                    }
                    else if(type!=null && type.equalsIgnoreCase("password"))
                    {
                        InputTextElement el=new InputTextElement(doc,this,elem.getStartOffset(),elem.getEndOffset(),elementName,value!=null?value:"");
                        if(maxLen>=0)
                            el.setMaxLen(maxLen);
                        formElements.add(el);
                    }
                    else if(type!=null && type.equalsIgnoreCase("hidden"))
                    {
                        TextElement el=new TextElement(doc,this,elem.getStartOffset(),elem.getEndOffset(),elementName,value!=null?value:"");
                        formElements.add(el);
                    }
                    else if(type!=null && type.equalsIgnoreCase("submit"))
                    {
                        InputSubmitElement el=new InputSubmitElement(doc,this,elem.getStartOffset(),elem.getEndOffset(),elementName,value!=null?value:"");
                        formElements.add(el);
                    }
                    else if(type!=null && type.equalsIgnoreCase("reset"))
                    {
                        InputResetElement el=new InputResetElement(doc,this,elem.getStartOffset(),elem.getEndOffset(),elementName,value!=null?value:"");
                        formElements.add(el);
                    }
                    else if(type!=null && type.equalsIgnoreCase("image"))
                    {
                        InputImageElement el=new InputImageElement(doc,this,elem.getStartOffset(),elem.getEndOffset(),elementName,alt,src);
                        formElements.add(el);
                    }
                    else if(type!=null && type.equalsIgnoreCase("checkbox"))
                    {
                        /*
                        HACK: Specification states that the value MUST be present. But pages sometimes do not
                        adhere to this. Some browsers therefore supply the standard value "on" for them. We will
                        do the same to accept as many forms as possible.
                        */
                        if(value==null)
                            value="on";
                        ElementGroup group=newGroup(doc,elementName,false);
                        InputCheckboxElement el=new InputCheckboxElement(doc,this,group,elem.getStartOffset(),elem.getEndOffset(),elementName,value,checked);
                        formElements.add(el);
                    }
                    else if(type!=null && type.equalsIgnoreCase("radio"))
                    {
                        if(value!=null)
                        {
                            ElementGroup group=newGroup(doc,elementName,true);
                            if(group instanceof SingleSelectionGroup)
                            {
                                InputRadioElement el=new InputRadioElement(doc,this,(SingleSelectionGroup)group,elem.getStartOffset(),elem.getEndOffset(),elementName,value,checked);
                                formElements.add(el);
                            }
                            else
                            {
                                // This radiobutton has the same name as a another group (SELECT or CHECKBOX)
                                throw new ParseException("The radiobutton has the same name as an existing group, which is NOT a single selection group",elem.getStartOffset());
                            }
                        }
                    }
                }
                // BUTTON element (only "submit" and "reset" handled, "button" usable only for scripts)
                if((set.getAttribute(StyleConstants.NameAttribute) instanceof HTML.UnknownTag)
                &&((HTML.UnknownTag)set.getAttribute(StyleConstants.NameAttribute)).equals(new HTML.UnknownTag("button")))
                {
                    // This button is in the current form
                    String type=(String)set.getAttribute(HTML.Attribute.TYPE);
                    String value=(String)set.getAttribute(HTML.Attribute.VALUE);
                    boolean endtag=set.getAttribute(HTML.Attribute.ENDTAG)!=null;
                    if(!endtag && (type==null || type.equalsIgnoreCase("submit")))
                    {
                        SubmitButtonElement el=new SubmitButtonElement(doc,this,elem.getStartOffset(),elem.getEndOffset(),elementName,value!=null?value:"");
                        formElements.add(el);
                    }
                    else if(!endtag && type.equalsIgnoreCase("reset"))
                    {
                        ResetButtonElement el=new ResetButtonElement(doc,this,elem.getStartOffset(),elem.getEndOffset(),elementName,value!=null?value:"");
                        formElements.add(el);
                    }
                }
                // TEXTAREA element
                if((set.getAttribute(StyleConstants.NameAttribute) instanceof HTML.Tag)
                &&((HTML.Tag)set.getAttribute(StyleConstants.NameAttribute)).equals(HTML.Tag.TEXTAREA))
                {
                    // This input is in the current form
                    String value="";
                    try
                    {
                        Document elemDoc=((Document)elem.getAttributes().getAttribute(StyleConstants.ModelAttribute));
                        value=elemDoc.getText(0,elemDoc.getLength());
                    }
                    catch(javax.swing.text.BadLocationException e)
                    {   // Should not happen; must be part of document!
    	                System.err.println(e.toString());
                	    e.printStackTrace(System.err);
                    }
                    InputTextElement el=new InputTextElement(doc,this,elem.getStartOffset(),elem.getEndOffset(),elementName,value!=null?value:"");
                    formElements.add(el);
                }
                // SELECT element
                if((set.getAttribute(StyleConstants.NameAttribute) instanceof HTML.Tag)
                &&((HTML.Tag)set.getAttribute(StyleConstants.NameAttribute)).equals(HTML.Tag.SELECT))
                {
                    boolean single=set.getAttribute(HTML.Attribute.MULTIPLE)==null;
                    ElementGroup group=getGroup(elementName);
                    if(group==null)
                    {   // May not currently exist!
                        group=newGroup(doc,elementName,single);
                        // Add all contained elemens
                        javax.swing.AbstractListModel mod=null;
                        if(single)
                        {
                            mod=(javax.swing.AbstractListModel)elem.getAttributes().getAttribute(StyleConstants.ModelAttribute);
                            if(mod instanceof javax.swing.DefaultComboBoxModel)
                            {
                                javax.swing.DefaultComboBoxModel model=(javax.swing.DefaultComboBoxModel)mod;
                                for(int i=0;i<model.getSize();i++)
                                {
                                    Option opt=(Option)model.getElementAt(i);
                                    FormElement el=new InputRadioElement(doc,this,(SingleSelectionGroup)group,elem.getStartOffset(),elem.getEndOffset(),group.getName(),opt.getValue(),opt.isSelected());
                                    el.setLabel(opt.getLabel());
                                    formElements.add(el);
                                }
                            }
                            else
                                // We expected a DefaultComboBoxModel, as this is a single-selection element
                                throw new ParseException("DefaultComboBoxModel expected as this is a single-selection group",elem.getStartOffset());
                        }
                        else
                        {
                            mod=(javax.swing.AbstractListModel)elem.getAttributes().getAttribute(StyleConstants.ModelAttribute);
                            if(mod instanceof javax.swing.DefaultListModel)
                            {
                                javax.swing.DefaultListModel model=(javax.swing.DefaultListModel)mod;
                                for(int i=0;i<model.getSize();i++)
                                {
                                    Option opt=(Option)model.getElementAt(i);
                                    FormElement el=new InputCheckboxElement(doc,this,group,elem.getStartOffset(),elem.getEndOffset(),group.getName(),opt.getValue(),opt.isSelected());
                                    el.setLabel(opt.getLabel());
                                    formElements.add(el);
                                }
                            }
                            else
                                // We expected a DefaultListModel, as this is a multiple-selection element
                                throw new ParseException("DefaultListModel expected as this is a multiple-selection group",elem.getStartOffset());
                        }
                    }
                    else
                        // A group with this name aleady exists (another SELECT, CHECKBOX or RADIOBUTTON)
                        throw new ParseException("A group with the name "+group.getName()+" already exists",elem.getStartOffset());
                }
            }
        }
        if(formAttributes.getAttribute(HTML.Attribute.METHOD)==null)
            formAttributes.addAttribute(HTML.Attribute.METHOD,"GET");
        if(formAttributes.getAttribute(HTML.Attribute.ACTION)==null)
            throw new IllegalArgumentException("Form: Action missing");
    }

    /**
    Creates a new HTML form by extracting the information from a document. If multiple forms are in the
    document, the name of the form must be provided else the first form will be returned.
    @param formPage the String of the age in which the form is contained
    @param formName the value of the "name" attribute of the desired form. If <code>null</code>, the first form encountered is parsed
    @return the parsed form from the page
    @throws ParseException if an error occured during parsing (wrong element names, ...)
    */
    public static HTMLForm getForm(String formPage,String formName) throws ParseException
    {
        java.io.StringReader in=new java.io.StringReader(formPage);
        HTMLEditorKit editorKit=new HTMLEditorKit();
        HTMLDocument doc=(HTMLDocument)editorKit.createDefaultDocument();
        // We don't want problems with changing the charset; just ignore them
        doc.putProperty("IgnoreCharsetDirective",new Boolean(true));
        try
        {
            editorKit.read(in,doc,0);
        }
        catch(javax.swing.text.BadLocationException e)
        {   // Should not happen!
            throw new RuntimeException("Internal error parsing the page ("+e.toString()+")");
        }
        catch(IOException e)
        {   // Should not happen!
            throw new RuntimeException("Internal error parsing the page ("+e.toString()+")");
        }
        return new HTMLForm(doc,formName);
    }

    /**
    Creates a new HTML form by extracting the information from a document by its position number (first form=1).
    @param formPage the String of the age in which the form is contained
    @param nthForm number of the form to return from the document
    @return the parsed form from the page
    @throws ParseException if an error occured during parsing (wrong element names, ...)
    */
    public static HTMLForm getForm(String formPage,int nthForm) throws ParseException
    {
        java.io.StringReader in=new java.io.StringReader(formPage);
        HTMLEditorKit editorKit=new HTMLEditorKit();
        HTMLDocument doc=(HTMLDocument)editorKit.createDefaultDocument();
        // We don't want problems with changing the charset; just ignore them
        doc.putProperty("IgnoreCharsetDirective",new Boolean(true));
        try
        {
            editorKit.read(in,doc,0);
        }
        catch(javax.swing.text.BadLocationException e)
        {   // Should not happen!
            throw new RuntimeException("Internal error parsing the page ("+e.toString()+")");
        }
        catch(IOException e)
        {   // Should not happen!
            throw new RuntimeException("Internal error parsing the page ("+e.toString()+")");
        }
        return new HTMLForm(doc,nthForm);
    }

    /**
    Parses all forms in the document and returns them as an array.
    @param formspage the page with the forms to parse
    @return an array of all forms in the page (length=0 if no forms found)
    @throws ParseException if an error occured during parsing (wrong element names, ...)
    */
    public static HTMLForm[] getForms(String formsPage) throws ParseException
    {
        java.io.StringReader in=new java.io.StringReader(formsPage);
        HTMLEditorKit editorKit=new HTMLEditorKit();
        HTMLDocument doc=(HTMLDocument)editorKit.createDefaultDocument();
        // We don't want problems with changing the charset; just ignore them
        doc.putProperty("IgnoreCharsetDirective",new Boolean(true));
        try
        {
            editorKit.read(in,doc,0);
        }
        catch(javax.swing.text.BadLocationException e)
        {   // Should not happen!
            throw new RuntimeException("Internal error parsing the page ("+e.toString()+")");
        }
        catch(IOException e)
        {   // Should not happen!
            throw new RuntimeException("Internal error parsing the page ("+e.toString()+")");
        }
        return HTMLForm.getForms(doc);
    }

    /**
    Parses all forms in the document and returns them as an array.
    @param doc the document to parse
    @return an array of all forms in the document (length=0 if no forms found)
    @throws ParseException if an error occured during parsing (wrong element names, ...)
    */
    public static HTMLForm[] getForms(HTMLDocument doc) throws ParseException
    {
        Vector res=new Vector();
        HTMLDocument.Iterator iter=doc.getIterator(HTML.Tag.FORM);
        // Works only, as "FORM" is NOT a BLOCK tag (those are not implemented!)
        int count=0;
        while(iter.isValid())
        {
            res.add(new HTMLForm(doc,++count));
            iter.next();
        }
        HTMLForm[] resArr=new HTMLForm[count];
        res.copyInto(resArr);
        return resArr;
    }

    /**
    Retrieve the method for submitting the form. Should either be "GET" or "POST".
    @return method for submitting the form
    */
    public String getMethod()
    {
        return (String)formAttributes.getAttribute(HTML.Attribute.METHOD);
    }

    /**
    Sets the base <code>URL</code> of the form.
    @param baseURL the new base <code>URL</code> for the form
    */
    public void setBaseURL(URL baseURL)
    {
        formAttributes.addAttribute(HTML.Tag.BASE,baseURL);
    }

    /**
    Retrieve the base <code>URL</code> of the form. Returns <code>null</code>
    if none specified in the document or explicitely set.
    @return the base <code>URL</code> of the form or <code>null</code>
    */
    public URL getBaseURL()
    {
        return (URL)formAttributes.getAttribute(HTML.Tag.BASE);
    }

    /**
    Check whether this dorm will be submitted to the same host this document is from.
    @return <code>true</code> if the form is sent to the same host it is from
    */
    public boolean isDocumentLocalSubmit()
    {
        URL action=getActionURL();
        URL base=(URL)formAttributes.getAttribute(HTML.Tag.BASE);
        return action.getHost()==null || action.getHost().equalsIgnoreCase(base.getHost());
    }

    /**
    Retrieve the <code>URL</code> where the form is submitted to. Does NOT include the parameters
    if the method for submitting is GET.
    @return <code>URl</code> for sending the form to
    */
    public URL getActionURL()
    {
        String actionString=(String)formAttributes.getAttribute(HTML.Attribute.ACTION);
        URL base=(URL)formAttributes.getAttribute(HTML.Tag.BASE);
        URL res=null;
        try
        {
            if(base!=null)
                res=new URL(base,actionString);
        }
        catch(MalformedURLException e)
        {   // Ignore
            System.err.print("");
        }
        if(res==null)
        {
            try
            {
                res=new URL(actionString);
            }
            catch(MalformedURLException e)
            {   // Ignore
                System.err.print("");
            }
        }
        return res;
    }

    /**
    Resets the form to its initial state. All elements are reset.
    */
    public void resetForm()
    {
        Enumeration enum=formElements.elements();
        while(enum.hasMoreElements())
            ((FormElement)enum.nextElement()).reset();
    }

    /**
    Submits the form. The parameter identifies the lement causing this action to happen (e. g. when
    using multiple submit buttons only the one invoked will be passed along).
    Uses a local <code>CookieStore</code> and no referer.
    @param submitter the element causing the form to submit or <code>null</code>
    @return the page the server sent in return to the post
    @throws IOException if an error occured during submitting the form or receiving the response
    */
    public String doSubmit(ActionElement submitter) throws IOException
    {
        return doSubmit(submitter,new CookieStore(),null);
    }

    /**
    Submits the form. The parameter identifies the lement causing this action to happen (e. g. when
    using multiple submit buttons only the one invoked will be passed along).
    @param submitter the element causing the form to submit or <code>null</code>
    @param store the cookies to send/receive
    @param referer the referer for this form
    @return the page the server sent in return to the post
    @throws IOException if an error occured during submitting the form or receiving the response
    */
    public String doSubmit(ActionElement submitter,CookieStore store,URL referer) throws IOException
    {
        String result="";
        Enumeration enum=formElements.elements();
        while(enum.hasMoreElements())
        {
            FormElement elem=(FormElement)enum.nextElement();
            // All selectable elements are in groups and will be submitted through them!
            // ActionElements are submitted only, if they caused the submit to happen (so reset buttons never)
            if(!(elem instanceof SelectableElement) && (!(elem instanceof ActionElement) || submitter==elem))
            {   // Add it
                if(result.length()>0)
                    result+="&"+elem.getEncoded();
                else
                    result=elem.getEncoded();
            }
        }
        // Handle the groups
        enum=groups.elements();
        while(enum.hasMoreElements())
        {
            ElementGroup elem=(ElementGroup)enum.nextElement();
            // Add it
            String group=elem.getEncoded();
            if(group.length()>0 && result.length()>0)
                result+="&"+group;
            else
                result=group;
        }
        // Retrieve the result of the form
        return WWWPageUtils.fetchForm(getActionURL(),store,result,referer,getMethod().equalsIgnoreCase("post"));
    }

    /**
    Retrieve the text visible immediately before an element (50 characters).
    @param elem the element to get the surrounding text for
    @param length the length of the text to get
    @return the text before the element
    */
    public String getPreText(FormElement elem,int length)
    {
        try
        {
            return doc.getText(Math.max(0,elem.getStartOffset()-length),length);
        }
        catch(BadLocationException e)
        {   // Should not happen anyway
            System.err.println("Error getting pre-text: "+e.toString());
            return "";
        }
    }

    /**
    Retrieve the text visible immediately after an element.
    @param elem the element to get the surrounding text for
    @param length the length of the text to get
    @return the text after the element
    */
    public String getPostText(FormElement elem,int length)
    {
        try
        {
            return doc.getText(elem.getStartOffset(),Math.min(doc.getLength()-elem.getStartOffset(),length));
        }
        catch(BadLocationException e)
        {   // Should not happen anyway
            System.err.println("Error getting post-text: "+e.toString());
            return "";
        }
    }

    /**
    Fetches the document associated with this element.
    @return the document
    */
    public Document getDocument()
    {
        return doc;
    }

    /**
    Fetches the parent element, which is always <code>null</code>, as forms have no parent.
    @return always <code>null</code>
    */
    public Element getParentElement()
    {
        return null;
    }

    /**
    Fetches the name of the form.
    @return the element name
    */
    public String getName()
    {
        return formName;
    }

    /**
    Fetches the collection of attributes this element contains.
    @return the attributes for the element
    */
    public AttributeSet getAttributes()
    {
        return formAttributes;
    }

    /**
    Fetches the offset from the beginning of the document
    that this form begins at.
    @return the starting offset >= 0
    */
    public int getStartOffset()
    {
        return startOffs;
    }

    /**
    Fetches the offset from the beginning of the document
    that this element ends at.
    @return the ending offset >= 0
    */
    public int getEndOffset()
    {
        return endOffs;
    }

    /**
    Gets the child element index closest to the given offset.
    The offset is specified relative to the begining of the
    document, not the beginning of the form.
    @param offset the specified offset >= 0
    @return the element index >= 0
    */
    public int getElementIndex(int offset)
    {
        int nchildren=formElements.size();
	    int index;
	    int lower=0;
	    int upper=nchildren-1;
	    int mid=0;
	    int p0=getStartOffset();
	    int p1;
	    if (nchildren == 0)
    		return 0;
	    if (offset>=getEndOffset())
    		return nchildren-1;
	    // see if the last index can be used.
	    if((lastIndex>=lower) && (lastIndex<=upper))
	    {
    		Element lastHit=(Element)formElements.elementAt(lastIndex);
	    	p0=lastHit.getStartOffset();
		    p1=lastHit.getEndOffset();
    		if((offset>=p0) && (offset<p1))
	    	    return lastIndex;
    		// last index wasn't a hit, but it does give useful info about
	    	// where a hit (if any) would be.
		    if (offset<p0)
    		    upper=lastIndex;
	    	else
		        lower=lastIndex;
	    }
	    while(lower<=upper)
	    {
		    mid=lower+((upper-lower)/2);
		    Element elem=(Element)formElements.elementAt(mid);
		    p0=elem.getStartOffset();
		    p1=elem.getEndOffset();
		    if((offset>=p0) && (offset<p1))
		    {
	    	    // found the location
    		    index=mid;
		        lastIndex=index;
		        return index;
		    }
		    else if(offset<p0)
		    {
		        upper=mid-1;
    		}
    		else
    		{
		        lower=mid+1;
		    }
	    }
	    // didn't find it, but we indicate the index of where it would belong
	    if (offset<p0)
		    index=mid;
	    else
    		index=mid+1;
	    lastIndex=index;
	    return index;
    }

    /**
    Gets the number of child elements contained by this element.
    @return the number of child elements >= 0
    */
    public int getElementCount()
    {
        return formElements.size();
    }

    /**
    Fetches the child element at the given index.
    @param index the specified index >= 0
    @return the child element
    */
    public Element getElement(int index)
    {
        if(index<formElements.size())
            return (FormElement)formElements.elementAt(index);
        return null;
    }

    /**
    Fetches the child elements (=form element) with the given name.
    May return multiple elements (e. g. radiobuttons returns all elements of the group, but not the
    group itself). Comparison of the name is case-insensitive.
    @param name the specified name (if <code>null</code> all elements are returned)
    @return the child elements (array of length 0 if none found)
    */
    public FormElement[] getElements(String name)
    {
        Vector res=new Vector();
        Enumeration enum=formElements.elements();
        while(enum.hasMoreElements())
        {
            FormElement elem=(FormElement)enum.nextElement();
            if(name==null || elem.getName().equalsIgnoreCase(name))
                res.add(elem);
        }
        FormElement[] resArr=new FormElement[res.size()];
        res.copyInto(resArr);
        return resArr;
    }

    /**
    Is this element a leaf element?
    @return <code>true</code> if a leaf element (no elements in the form) else <code>false</code>
    */
    public boolean isLeaf()
    {
        return formElements.size()==0;
    }

    /**
    Retrieve the group object for a certain name.
    @param groupName the name of the group
    @return the group of elements
    */
    public ElementGroup getGroup(String groupName)
    {
        return (ElementGroup)groups.get(groupName.toLowerCase());
    }

    /**
    Creates a new group element and adds it to the list of groups. If a group with this name already
    exists, this group is returned and no new one created.
    Groups can only contain subclasses of <code>SelectableElement</code>
    @param doc the document this group belongs to
    @param groupName the name of the group
    @param singleSelectionOnly if <code>true</code>, only one element of the group may be selected at any time.
    @return the created group
    @see FIM.Util.WWW.Form.SelectableElement
    */
    public ElementGroup newGroup(HTMLDocument doc,String groupName,boolean singleSelectionOnly)
    {
        ElementGroup res=null;
        res=(ElementGroup)groups.get(groupName.toLowerCase());
        if(res==null)
        {
            if(singleSelectionOnly)
                res=new SingleSelectionGroup(doc,this,groupName);
            else
                res=new ElementGroup(doc,this,groupName);
            groups.put(groupName.toLowerCase(),res);
        }
        return res;
    }

    /**
    Returns a string representation of the form
    @return string representation of the form
    */
    public String toString()
    {
        String res="";
        res+="Name: "+getName()+"\n";
        res+="Action URL: "+getActionURL()+"\n";
        res+="Starts in document at "+getStartOffset()+" and ends at "+getEndOffset()+"\n";
/*
        // Print the whole HTML-code
        java.io.StringWriter str=new java.io.StringWriter();
        try
        {
            javax.swing.text.html.HTMLWriter wr=new javax.swing.text.html.HTMLWriter(str,doc,getStartOffset(),getEndOffset()-getStartOffset());
            wr.write();
            res+="Content:\n"+str.toString()+"\n";
        }
        catch(BadLocationException e)
        {   // Should not happen anyway
            System.err.println("Error getting text of form: "+e.toString());
        }
        catch(IOException e)
        {   // Should not happen anyway
            System.err.println("Error getting text of form: "+e.toString());
        }
*/
        Enumeration enum=formElements.elements();
        while(enum.hasMoreElements())
        {
            FormElement elem=(FormElement)enum.nextElement();
            if(!(elem instanceof SelectableElement))
            {   // Add it
                res+=elem.toString()+"\n";
            }
        }
        // Handle the groups
        enum=groups.elements();
        while(enum.hasMoreElements())
        {
            ElementGroup elem=(ElementGroup)enum.nextElement();
            res+=elem.toString()+"\n";
        }
        return res;
    }

}

