/*
 * @(#)InputImageElement.java
 *
 * Classes for handling HTML-Forms without UI
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.WWW.Form;

import javax.swing.text.html.HTMLDocument;
import javax.swing.text.Element;
import java.net.URL;

/**
Class for the HTML element INPUT with subtype IMAGE.
The alternative text is stored as the value of the element.
If clicked, the form is submitted and the coordinates within the image submitted.
The resulting page and any errors are stored. Invoking it again clears this previous data.

@author  Michael Sonntag
@version 1.0, 1.7.2000
*/
public class InputImageElement extends InputSubmitElement
{
    /**
    The <code>URL</code> of the image to display (as string).
    @serial
    */
    private String src=null;
    /**
    The x coordinate clicked (0..width of image).
    @serial
    */
    private int x=-1;
    /**
    The y coordinate clicked (0..height of image).
    @serial
    */
    private int y=-1;

    /**
    Creates a new INPUT/IMAGE element.
    @param doc the document this element belongs to
    @param parent the parent of this element
    @param startOffs start offset of the element in the document
    @param endOffs end offset of the element in the document
    @param name the name of the element
    @param src the <code>URL</code> of the image
    @param alt the alternative text for the element (stored in value)
    */
    public InputImageElement(HTMLDocument doc,Element parent,int startOffs,int endOffs,String name,String alt,String src)
    {
        super(doc,parent,startOffs,endOffs,name,alt!=null?alt:"");
        if(src==null || src.length()<1)
            throw new IllegalArgumentException("URL for the image must be provided");
        this.src=src;
    }
    
    /**
    Resets the element. Both coordinates are set to -1.
    */
    public void reset()
    {
        super.reset();
        x=-1;
        y=-1;
    }
    
    /**
    Retrieve the <code>URL</code> for the image to display. Parses it in the context of the
    base-url of the form if specified.
    @return the <code>URL</code> of the image
    */
    public URL getSrcAsURL() throws java.net.MalformedURLException
    {
        URL base=(URL)((HTMLForm)getParentElement()).formAttributes.getAttribute(javax.swing.text.html.HTML.Tag.BASE);
        if(base!=null)
            return new URL(base,src);
        return new URL(src);
    }
    
    /**
    Retrieve the <code>URL</code> for the image to display.
    @return the <code>URL</code> of the image as a string
    */
    public String getSrc()
    {
        return src;
    }
    
    /**
    Set the coordinates to pass on when submitting. If negative, this coordinate will not be included.
    @param x x coordinate the user clicked
    @param y y coordinate the user clicked
    */
    public void setCoordinates(int x,int y)
    {
        this.x=x;
        this.y=y;
    }
    
    /**
    Retrieve this element in encoded form. Returns only the name if both coordinates are
    negative. Otherwise, <name>.x=<x-coord> and/or <name>.y=<y-coord> are returned.
    @return the name and possibly the coordinates the user clicked in encoded form
    */
    public String getEncoded()
    {
        if(x<0 && y<0)
            return java.net.URLEncoder.encode(getName());
        String res="";
        if(x>=0)
            res+=java.net.URLEncoder.encode(getName())+".x="+x;
        if(x>=0 && y>=0)
            res+="&";
        if(y>=0)
            res+=java.net.URLEncoder.encode(getName())+".y="+y;
        return res;
    }

    /**
    Returns a string representation of the input image element.
    @return string representation of the form element
    */
    public String toString()
    {
        return "Input image ("+getName()+":"+getValue()+"), Src="+src;
    }
}