/*
 * @(#)SelectableElement.java
 *
 * Classes for handling HTML-Forms without UI
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.WWW.Form;

import javax.swing.text.html.HTMLDocument;
import javax.swing.text.Element;

/**
Parent class for elements, which may be selected (on/off only).
All those elements must belong to a <code>ElementGroup</code> (which might contain only a
single selectable element). Forms are stored flat, so the groups are not the parent of those elements!

@author  Michael Sonntag
@version 1.0, 1.7.2000
@see FIM.Util.WWW.Form.ElementGroup
*/
public abstract class SelectableElement extends FormElement
{
    /**
    The group the element belongs to. May not be <code>null</code>.
    @serial
    */
    private ElementGroup group=null;
    /**
    Whether this element is selected or not.
    Package access for SingleSelectionGroup.
    @serial
    */
    boolean selected=false;
    /**
    Whether this element was selected initially.
    @serial
    */
    private boolean initiallySelected=false;
    /**
    The value of this element (the string submitted if it is selected).
    @serial
    */
    private String value=null;

    /**
    Creates a new INPUT/HIDDEN element, which just stores a piece of text.
    @param doc the document this element belongs to
    @param parent the parent of this element
    @param group the group this element belongs to
    @param startOffs start offset of the element in the document
    @param endOffs end offset of the element in the document
    @param name the name of the element
    @param value the value of the element (which is submitted if it is selected) (Must be provided
            according to the specification, although not always done; see {@link FIM.Util.WWW.Form.HTMLForm#parseForm(HTMLDocument.Iterator) HTMLForm.parseForm(HTMLDocument.Iterator)()} for more)
    @param initiallySelected if <code>true</code>, this element is selected at the beginning
    */
    public SelectableElement(HTMLDocument doc,Element parent,ElementGroup group,int startOffs,int endOffs,String name,String value,boolean initiallySelected)
    {
        super(doc,parent,startOffs,endOffs,name);
        if(group==null)
            throw new IllegalArgumentException("Group must be provided");
        this.group=group;
        group.addContainedElement(this);
        this.initiallySelected=initiallySelected;
        selected=initiallySelected;
        if(value==null  || value.length()<1)
            throw new IllegalArgumentException("Value must be provided");
        this.value=value;
    }

    /**
    Retrieve the group this element belongs to.
    @param the group of this element
    */
    public ElementGroup getGroup()
    {
        return group;
    }

    /**
    Resets this element to its initial state.
    */
    public void reset()
    {
        selected=initiallySelected;
    }

    /**
    Sets the state of this element.
    @param isSelected if <code>true</code>, the element is selected, otherwise not
    */
    public void setSelected(boolean isSelected)
    {
        if(isSelected && !selected)
            select();
        else if(!isSelected && selected)
            deselect();
    }

    /**
    Sets the state of the element to selected.
    */
    public void select()
    {
        selected=true;
    }

    /**
    Sets the state of the element to not selected.
    */
    public void deselect()
    {
        selected=false;
    }

    /**
    Retrieve whether this element is currently selected or not.
    @return <code>true</code> if this element is selected, <code>false</code> otherwise
    */
    public boolean isSelected()
    {
        return selected;
    }

    /**
    Retrieve the value of this element.
    @return the value of the element
    */
    public String getValue()
    {
        return value;
    }

    /**
    Selectable elements are never encoded (this is done by the group, they belong to), so
    throws always a <code>IllegalStateException</code>.
    */
    public String getEncoded()
    {
        throw new IllegalStateException("Selectable elements may not be encoded (use groups instead)");
    }

    /**
    Matches this element against a number of strings and returns a measure for matching them.
    0=no match, 100=perfect match.<br>
    Extension to <code>FormElement</code>: Value match=70 exact, 60 substring
    @param matchStr an array of strings which are matched against this element
    @return measure how much this element matches the parameter
    */
    public int match(String[] matchStr)
    {
        int curMax=super.match(matchStr);
        for(int i=0;i<matchStr.length;i++)
        {
            int cur=0;
            if(value!=null && value.equalsIgnoreCase(matchStr[i]))
                cur=70;
            if(value!=null && value.toLowerCase().indexOf(matchStr[i].toLowerCase())>=0)
                cur=60;
            if(cur>curMax)
                curMax=cur;
        }
        return curMax;
    }

    /**
    Returns a string representation of the selectable element.
    @return string representation of the form element
    */
    public String toString()
    {
        return "Selectable element ("+getName()+":"+getValue()+")="+getLabel()+", "+((selected)?"Selected":"Not selected");
    }
}