/*
 * @(#)TextElement.java
 *
 * Classes for handling HTML-Forms without UI
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.WWW.Form;

import javax.swing.text.html.HTMLDocument;
import javax.swing.text.Element;

/**
Parent class for HTML elements containing text.

@author  Michael Sonntag
@version 1.0, 1.7.2000
*/
public class TextElement extends FormElement
{
    /**
    The initial text of this control. If <code>null</code>, it is empty.
    @serial
    */
    private String initialText=null;
    /**
    The current text of the element.
    @serial
    */
    protected String currentText=null;

    /**
    Creates a new element for storing text.
    @param doc the document this element belongs to
    @param parent the parent of this element
    @param startOffs start offset of the element in the document
    @param endOffs end offset of the element in the document
    @param name the name of the element
    @param initialText the initial text content of the element
    */
    public TextElement(HTMLDocument doc,Element parent,int startOffs,int endOffs,String name,String initialText)
    {
        super(doc,parent,startOffs,endOffs,name);
        if(initialText==null)
            initialText="";
        this.initialText=initialText;
        currentText=initialText;
    }
    
    /**
    Resets the element to the original text.
    */
    public void reset()
    {
        currentText=initialText;
    }
    
    /**
    retrieve the current text of the element.
    @return the current text of the element
    */
    public String getText()
    {
        return currentText;
    }
    
    /**
    Set the current text of the element.
    @param newText the new text
    */
    public void setText(String newText)
    {
        if(newText==null)
            newText="";
        currentText=newText;
    }
    
    /**
    Retrieve the encoded form of the element: <name>=<text>.
    @return the name and the text in encoded form
    */
    public String getEncoded()
    {
        return java.net.URLEncoder.encode(getName())+"="+java.net.URLEncoder.encode(currentText);
    }

    /**
    Matches this element against a number of strings and returns a measure for matching them.
    0=no match, 100=perfect match.<br>
    Extension to <code>FormElement</code>: <strong>Current</strong> text match=50 exact, 40 substring
    @param matchStr an array of strings which are matched against this element
    @return measure how much this element matches the parameter
    */
    public int match(String[] matchStr)
    {
        int curMax=super.match(matchStr);
        for(int i=0;i<matchStr.length;i++)
        {
            int cur=0;
            if(getText().equalsIgnoreCase(matchStr[i]))
                cur=50;
            if(getText().toLowerCase().indexOf(matchStr[i].toLowerCase())>=0)
                cur=40;
            if(cur>curMax)
                curMax=cur;
        }
        return curMax;
    }    

    /**
    Returns a string representation of the text element.
    @return string representation of the form element
    */
    public String toString()
    {
        return "Text element ("+getName()+"): "+getText();
    }
}