/*
 * @(#)HTMLBreakFilterReader.java
 *
 * Utility classes for managing webpages
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.WWW;

import java.io.Reader;
import java.io.FilterReader;
import java.io.IOException;
import java.util.Hashtable;

/**
Converts a stream of HTML-code to a canonical format of linebreaks.
Removes all linebreaks (\r and \n) and changes tabs to a single space.
Afterwards, in front of &lt;p&gt;, &lt;/p&gt; and &lt;br&gt; "\n" is inserted.
The same happens in front of &lt;table&gt; and &lt;tr&gt;, although only if no
double-linebreaks are created through this.

@author  Michael Sonntag
@version 1.0, 1.7.2000
*/
public class HTMLBreakFilterReader extends FilterReader
{
    /** The last character we read from the input stream. */
    protected int lastChar=0;
    /** Marks if the last characters returned was a linebreak (following whitespaces are ignored for this). */
    protected boolean lastWhitespace=false;
    /** A buffer for reading a whole tag. */
    protected String buffer="";

    /**
    Creates a new reader.
    @param in the reader to read from 
    */
	public HTMLBreakFilterReader(Reader in)
	{
		super(in);
	}
	
	/**
    Tell whether this stream supports the mark() operation (this stream does not).
    @return always <code>false</code>
	*/
	public boolean markSupported()
	{
	    return false;
	}

	/**
    Read characters into an array. This method will block until some input is available, 
    an I/O error occurs, or the end of the stream is reached.
    @param cbuf destination buffer 
    @param off the offset in the buffer the first character will be written to
    @param len the number of characters to read
    @return the number of bytes read, or -1 if the end of the stream has been reached
    @throws IOException if an I/O error occurs 
    */
    public int read(char[] cbuf,int off,int len) throws IOException
    {
        if(off<0)
            throw new IllegalArgumentException("Offset must be at least 0");
        if(len>cbuf.length)
            throw new IllegalArgumentException("Length must be smaller or equal to length of buffer");
        if(off+len>cbuf.length)
            throw new IllegalArgumentException("Would write over end of buffer");
        int count=0;
        int cur=read();
        while(count<len && cur>=0)
        {
            cbuf[off+count]=(char)cur;
            count++;
            cur=read();
        }
        if(cur<0)
            return -1;
        return count;
    }

    /**
    Read a single character. This method will block until a character is available, 
    an I/O error occurs, or the end of the stream is reached. 
    @return the character read, as an integer in the range 0 to 16383 (0x00-0xffff), or -1 if the end of the stream has been reached
    @throws IOException if an I/O error occurs
    */
    public int read() throws IOException
    {
        if(buffer.length()==0)
        {
            int next=getNextChar();
            if(next<0)
                return -1;
            buffer+=(char)next;
            if(next=='<')
            {   // Probably a special tag
                int count=0;
                while(count<10 && next>=0 && next!='>' && !Character.isWhitespace((char)next) && !Character.isSpaceChar((char)next))
                {
                    next=getNextChar();
                    if(next>=0)
                        buffer+=(char)next;
                }
                if(count<10 && next>=0)
                {   // Check if it is one of the tags to change
                    String check=buffer.toLowerCase();
                    if((check.startsWith("<table") || check.startsWith("<tr>") || 
                            check.startsWith("<br>") || check.startsWith("<p>") || 
                            check.startsWith("</p>")) && !lastWhitespace)
                        buffer="\n"+buffer; // Insert linebreak avoiding doubles
                }
            }
        }
        int res=buffer.charAt(0);
        lastWhitespace=res=='\n' || lastWhitespace && Character.isWhitespace((char)res);
        buffer=buffer.substring(1);
        return res;
    }
    
    /**
    Retrieves the next character. Skips \r, \n and converts tabs to spaces. 
    @return the next character or -1 if end of stream reached
    @throws IOException if an I/O error occurs
    */
    protected int getNextChar() throws IOException
    {
        boolean skip=true;
        while(lastChar>=0 && skip)
        {
            lastChar=in.read();
            skip=lastChar=='\r' || lastChar=='\n' || lastChar=='\t';
        }
        return lastChar;
    }

    /**
    Skip characters. This method will block until some characters are available, 
    an I/O error occurs, or the end of the stream is reached.
    @param n the number of characters to skip 
    @return the number of characters actually skipped
    @throws IOException if an I/O error occurs 
    */
    public long skip(long n) throws IOException	
    {
        // Cannot tell in advance how much we must read, so read characters till count 
        // reached or end of stream found
        int count;
        int cur=0;
        for(count=0;count<n && cur>=0;count++)
            cur=read();
        return count;
    }
}