/*
 * @(#)WWWPageUtils.java
 *
 * Utility classes for managing webpages
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.WWW;

import java.net.URL;
import java.net.HttpURLConnection;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.io.PrintWriter;
import java.io.BufferedReader;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.Reader;
import java.io.InputStreamReader;
import java.io.IOException;
import java.net.MalformedURLException;
import FIM.Util.WWW.Cookies.CookieStore;
import FIM.Util.WWW.CanonicalFilterReader;
import FIM.Util.WWW.StripTagFilterReader;
import FIM.Util.WWW.CanonicalHTMLFilterReader;
import FIM.Util.WWW.ConvertHTMLCharFilterReader;

/**
Utility functions for managing WWW-Pages. Allows retrieving webpages and posting forms. Includes
support for using cookies.

@author  Michael Sonntag
@version 1.0, 1.7.2000
*/
public class WWWPageUtils
{
    /** Maximum number of redirects to follow (MAX_REDIRECT_DEPTH+1 will be the final page). */
    public final static int MAX_REDIRECT_DEPTH=5;


/**
Retrieve a page and return it as a single string. Lines are separated by '\n'.
@param page the page to fetch
@param cookies the <code>CookieStore</code> to use (if <code>null</code> cookies will be ignored and none given out)
@return the page as a single string
@throws IOException if an error occured during fetching the page
@throws MalformedURLException if the string could not be converted to an <code>URL</code>
*/
public static String fetchPage(String page,CookieStore cookies) throws MalformedURLException,IOException
{
    URL pageURL=new URL(page);
    return fetchPage(pageURL,cookies);
}

/**
Retrieve a page and return it as a single string. Lines are separated by '\n'.
Must be synchronized as <code>fecthPage</code> and <code>fetchForm</code> both modify
<code>HttpURLConnection.setFollowRedirects()</code>, which is a static method and the same
for all connections.
@param page the page to fetch
@param cookies the <code>CookieStore</code> to use (if <code>null</code> cookies will be ignored and none given out)
@return the page as a single string
@throws IOException if an error occured during fetching the page
*/
public synchronized static String fetchPage(URL page,CookieStore cookies) throws IOException
{
    String res="";
    // We don't do redirection by ourselfes; let framework do it
    // Must be privileged as it calls checkPermission("setFactory")
    AccessController.doPrivileged(new PrivilegedAction()
    {
        public Object run()
        {
            HttpURLConnection.setFollowRedirects(true);
            return null;
        }
    }
    );
    HttpURLConnection con=(HttpURLConnection)page.openConnection();
    if(cookies!=null)
        cookies.sendCookies(con,false);
    con.setDoInput(true);
    con.connect();
    if(con.getResponseCode()!=HttpURLConnection.HTTP_OK)
        throw new IOException("Error retrieving page: "+con.getResponseMessage()+" ("+con.getResponseCode()+")");
    if(cookies!=null)
        cookies.parseCookies(con);
    BufferedReader in=new BufferedReader(new InputStreamReader(con.getInputStream()));
    String line=in.readLine();
    while(line!=null)
    {
        res+=line+"\n";
        line=in.readLine();
    }
    in.close();
    return res;
}

/**
Send a form and retrieve the response.
@param page the <code>URL</code> the form to post to. Must include the full action (e. g. "http://www.acme.com/doIt), but not the parameters (e. g. "?do=action&amp;param=now+and+then")
@param cookies the <code>CookieStore</code> to use (if <code>null</code> cookies will be ignored and none given out)
@param request the request of the page (i. e. the content of the inputs; e. g. "?do=action&amp;param=now+and+then")
@param referer the <code>URL</code> of the page containing the form (request property "Referer") (not set if <code>null</code>)
@param post if <code>true</code>, submit method is POST, otherwise GET
@return the page as a single string
@throws IOException if an error occured during fetching the page
@throws MalformedURLException if the string could not be converted to an <code>URL</code>
*/
public static String fetchForm(String page,CookieStore cookies,String request,URL referer,boolean post) throws MalformedURLException,IOException
{
    URL pageURL=new URL(page);
    return fetchForm(pageURL,cookies,request,referer,post,0);
}

/**
Send a form and retrieve the response.
@param page the <code>URL</code> the form to post to. Must include the full action (e. g. "http://www.acme.com/doIt), but not the parameters (e. g. "?do=action&amp;param=now+and+then")
@param cookies the <code>CookieStore</code> to use (if <code>null</code> cookies will be ignored and none given out)
@param request the request of the page (i. e. the content of the inputs; e. g. "?do=action&amp;param=now+and+then")
@param referer the <code>URL</code> of the page containing the form (request property "Referer") (not set if <code>null</code>)
@param post if <code>true</code>, submit method is POST, otherwise GET
@return the page as a single string
@throws IOException if an error occured during fetching the page
@throws MalformedURLException if the string could not be converted to an <code>URL</code>
*/
public static String fetchForm(URL page,CookieStore cookies,String request,URL referer,boolean post) throws IOException
{
    return fetchForm(page,cookies,request,referer,post,0);
}

/**
Post a form (only POST supported) and retrieve the response. Internal procedure for limiting the number of redirects.
Must be synchronized as <code>fecthPage</code> and <code>fetchForm</code> both modify
<code>HttpURLConnection.setFollowRedirects()</code>, which is a static method and the same
for all connections.
@param page the <code>URL</code> the form to post to. Must include the full action (e. g. "http://www.acme.com/doIt), but not the parameters (e. g. "?do=action&amp;param=now+and+then")
@param cookies the <code>CookieStore</code> to use (if <code>null</code> cookies will be ignored and none given out)
@param request the request of the page (i. e. the content of the inputs; e. g. "?do=action&amp;param=now+and+then")
@param referer the <code>URL</code> of the page containing the form (request property "Referer") (not set if <code>null</code>)
@param post if <code>true</code>, submit method is POST, otherwise GET
@param depth the current depth (number of redirects)
@return the page as a single string
@throws IOException if an error occured during fetching the page
@throws MalformedURLException if the string could not be converted to an <code>URL</code>
*/
private synchronized static String fetchForm(URL page,CookieStore cookies,String request,URL referer,boolean post,int depth) throws IOException
{
    if(!post)
    {   // GET method used for submitting data
        try
        {
           page=new URL(page.toString()+"?"+request);
        }
        catch(MalformedURLException e)
        {
            throw new IOException("Could not create URL for GET!");
        }
    }
    // We do redirection by ourselfes
    // Must be privileged as it calls checkPermission("setFactory")
    AccessController.doPrivileged(new PrivilegedAction()
    {
        public Object run()
        {
            HttpURLConnection.setFollowRedirects(false);
            return null;
        }
    }
    );
    HttpURLConnection con=null;
    String result="";
    // Open the connection
    try
    {
        con=(HttpURLConnection)page.openConnection();
    }
    catch(IOException e)
    {
        if(con!=null)
            con.disconnect();
        throw new IOException("Couldn't connect to "+page.toString()+": "+e.toString());
    }
    if(post)
        con.setRequestMethod("POST");
    else
        con.setRequestMethod("GET");
    con.setRequestProperty("Content-Type","application/x-www-form-urlencoded");
    // Set referer
    if(referer!=null)
        con.setRequestProperty("Referer",referer.toString());
    // Send cookies
    if(cookies!=null)
        cookies.sendCookies(con,false);
    con.setDoOutput(true);
    con.setDoInput(true);
    // Connect
    con.connect();
    if(post)
    {
        // Send our request (form content)
        PrintWriter out=null;
        try
        {
            out=new PrintWriter(con.getOutputStream());
            out.print(request); // NO linefeed at the end! E. g. IIS would block then and NEVER send a response.
        }
        catch(IOException e)
        {
            if(con!=null)
                con.disconnect();
            throw new IOException("Couldn't send request to "+page.toString()+": "+e.toString());
        }
        finally
        {
            out.close();
        }
    }
    // Receive cookies
    if(cookies!=null)
        cookies.parseCookies(con);
    // 300...305 are redirection requests
    if(con.getResponseCode()!=HttpURLConnection.HTTP_OK && !(con.getResponseCode()>=300 && con.getResponseCode()<=305))   // OK or REDIRECT means we did it
        throw new IOException("Couldn't load page \""+page.toString()+"\" ("+con.getResponseCode()+": "+con.getResponseMessage()+")");
    if(con.getResponseCode()>=300 && con.getResponseCode()<=305 && depth<MAX_REDIRECT_DEPTH)   // REDIRECT -> follow it to maximum depth
    {
        URL redirectedTo=new URL(page,con.getHeaderField("Location"));
        /* The HTTP/1.1 spec says that a redirect from a POST
	     * *should not* be immediately turned into a GET, and
		 * that some HTTP/1.0 clients incorrectly did this.
    	 * Correct behavior redirects a POST to another POST.
	     * Unfortunately, since most browsers have this incorrect
		 * behavior, the web works this way now.  Typical usage
    	 * seems to be:
	     *   POST a login code or passwd to a web page.
	     *   after validation, the server redirects to another
   		 *     (welcome) page
    	 *   The second request is (erroneously) expected to be GET
	     *
    	 * We will do the incorrect thing (POST-->GET) by default.
    	 * We will provide the capability to do the "right" thing
	     * (POST-->POST) by a system property, "http.strictPostRedirect=true"
   		 COPIED FROM sun.net.www.protocol.http.HttpURLConnection.java; will do the same here
     	*/
        if(post && Boolean.getBoolean("http.strictPostRedirect"))
        {
            return fetchForm(redirectedTo,cookies,request,page,true,depth+1);
        }
        else
        {
            return fetchPage(redirectedTo.toString(),cookies);
        }
    }
    else
    {
        // Get the returned page
        BufferedReader in=new BufferedReader(new InputStreamReader(con.getInputStream()));
        try
        {
            String line=in.readLine();
            while(line!=null)
            {
                result+=line+"\n";
                line=in.readLine();
            }
            in.close();
        }
        catch(IOException e)
        {
            throw new IOException("Couldn't get result of sending: "+e.toString());
        }
        finally
        {
            con.disconnect();
        }
    }
    return result;
}

/**
Strips all tags from a string. This will remove everything between '<' and '>'.
Converts &lt;br&gt;, &lt;p&gt; and &lt;/p&gt; to "\n" and removes all other linebreaks and tabs.
Also converts the special characters (&amp;nbsp; = " ", &amp;auml; = "", ...) and types of "&amp;#160;".
@return the input string stripped from all HTML-tags
*/
public static String stripTags(String str)
{
    Reader in=new CanonicalFilterReader(new StripTagFilterReader(new CanonicalHTMLFilterReader(new ConvertHTMLCharFilterReader(new StringReader(str)))));
    StringWriter out=new StringWriter();
    try
    {
        int ch=in.read();
        while(ch>=0)
        {
            out.write((char)ch);
            ch=in.read();
        }
        out.close();
    }
    catch(IOException e)
    {   // Should never happen
    	System.err.println(e.toString());
        e.printStackTrace(System.err);
    }
    return out.toString();
}

/**
Takes a string of HTML as an input an canonicalizes it.
All linebreaks and multiple spaces within tags are removed.
E. g. "&lt;\n  /  p&gt; will be changed to "&lt;/p&gt;".
@param str the input string
@return the canonicalized string
*/
public static String canonicalize(String str)
{
    Reader in=new CanonicalHTMLFilterReader(new ConvertHTMLCharFilterReader(new StringReader(str)));
    StringWriter out=new StringWriter();
    try
    {
        int ch=in.read();
        while(ch>=0)
        {
            out.write((char)ch);
            ch=in.read();
        }
        out.close();
    }
    catch(IOException e)
    {   // Should never happen
   	    System.err.println(e.toString());
 	    e.printStackTrace(System.err);
    }
    return out.toString();
}

}