/*
 * @(#)XMLByteArray.java
 *
 * Miscellaneous XML utilities
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.XML;

import FIM.Util.XML.XMLUtils;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Text;
import org.w3c.dom.DOMException;

import java.io.Serializable;

/**
Encodes a byte array as an XML element. The name of the element is always "XMLByteArray".
@author Michael Sonntag
@version 1.0
*/
public class XMLByteArray implements XMLEncodeable, Serializable
{
    /**
    The array of bytes to be encoded/stored.
    */
    private byte[] data=null;
    
    /**
    Create an empty XMLByteArray object.
    */
    public XMLByteArray()
    {
    }
    
    /**
    Creates a new XMLByteArray object from a byte array.
    @param data the content of the XMLString
    */
    public XMLByteArray(byte[] data)
    {
        setBytes(data);
    }
    
    /**
    Creates a new XMLByteArray object from an element.
    @param data the element to parse
    */
    public XMLByteArray(Element elem)
    {
        decodeXMLElement(elem);
    }
    
    /**
    Retrieve the contained data as a byte arra.
    @return the contained data
    */
    public byte[] getBytes()
    {
        return data;
    }
    
    /**
    Set the contained data to a new value. The old value will be overwritten and is lost.
    @param data the data to set
    */
    public void setBytes(byte[] data)
    {
        this.data=new byte[data.length];
        System.arraycopy(data,0,this.data,0,data.length);
    }
    
    /**
    Retrieve the byte array as a <code>Serializable</code>.
    @return the byte array as a serializable
    */
	public Serializable getAsSerializable()
	{
	    return data;
	}
	
	/**
	Returns the string encoded as an XML element with the name "XMLString". The string itself
	can be everything, as it is contained in a CDATA section (and "]]>" is escaped, so it
	may also be contained).
	@return the string encoded in an XML element
	*/
	public Element getAsXMLElement(Document doc)
	{
	    Element el=doc.createElement("XMLByteArray");
	    try
	    {
	        el.appendChild(doc.createTextNode(XMLUtils.toHexString(data)));
	    }
	    catch(DOMException e)
	    {
	        // Should never happen, unless the document is a HTML-document and not an XML one
	        System.err.println(e.toString());
	    }
	    return el;
	}
	
	/**
	Restores the contained string from an element.
	@param elem the element containing the string
	@throws IllegalArgumentException if the element is wrong or ill-formed (tag name not "XMLString" 
	or does not contain only a CDATASection)
	*/
	public void decodeXMLElement(Element elem) throws IllegalArgumentException
	{
	    if(!elem.getTagName().equals("XMLByteArray"))
	        throw new IllegalArgumentException("Element is not a XMLByteArray!");
	    if(elem.getChildNodes().getLength()>1)
	        throw new IllegalArgumentException("Element has more than one child!");
	    if(elem.getChildNodes().getLength()<1)
	        data=new byte[0];
	    else
	    {
    	    if(!(elem.getFirstChild() instanceof Text))
	            throw new IllegalArgumentException("Element does not contain a text!");
	        data=XMLUtils.toByteArray(((Text)elem.getFirstChild()).getData());
	    }
	}	
	
	/**
	Retrieve the contained string.
	@return the contained string
	*/
    public String toString()
    {
        return XMLUtils.toHexString(data);
    }
    
    /**
    Compares two XMLByteArray objects
    @param obj the object to compare to this one
    */
    public boolean equals(Object obj)
    {
        return (obj instanceof XMLByteArray) && XMLUtils.equalByteArrays(data,((XMLByteArray)obj).getBytes());
    }
}