/*
 * @(#)XMLString.java
 *
 * Miscellaneous XML utilities
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.XML;

import FIM.Util.XML.XMLUtils;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.CDATASection;
import org.w3c.dom.DOMException;

import java.io.Serializable;

/**
Encodes a string as an XML element. The name of the element is always "XMLString".
@author Michael Sonntag
@version 1.0
*/
public class XMLString implements XMLEncodeable, Serializable
{
    /**
    The string to be encoded
    */
    private String str=null;
    
    /**
    Create an empty XMLString object.
    */
    public XMLString()
    {
    }
    
    /**
    Creates a new XMLString object from a string.
    @param str the content of the XMLString (may not be <code>null</code>)
    */
    public XMLString(String str)
    {
        setString(str);
    }
    
    /**
    Creates a new XMLString object from an element.
    @param elem the element to parse
    */
    public XMLString(Element elem)
    {
        decodeXMLElement(elem);
    }
    
    /**
    Retrieve the contained string.
    @return the contained string
    */
    public String getString()
    {
        return str;
    }
    
    /**
    Set the contained string to a new value. The old value will be overwritten and is lost.
    @param str the string to set
    */
    public void setString(String str)
    {
        if(str==null)
            throw new IllegalArgumentException("String may not be null");
        this.str=str;
    }
    
    /**
    Retrieve the string as a <code>Serializable</code>.
    @return the string as a serializable
    */
	public Serializable getAsSerializable()
	{
	    return str;
	}
	
	/**
	Returns the string encoded as an XML element with the name "XMLString". The string itself
	can be everything, as it is contained in a CDATA section (and "]]>" is escaped, so it
	may also be contained).
	@return the string encoded in an XML element
	*/
	public Element getAsXMLElement(Document doc)
	{
	    Element el=doc.createElement("XMLString");
	    try
	    {
	        el.appendChild(doc.createCDATASection(XMLUtils.CDATAQuoteText(str)));
	    }
	    catch(DOMException e)
	    {
	        // Should never happen, unless the document is a HTML-document and not an XML one
	        System.err.println(e.toString());
	    }
	    return el;
	}
	
	/**
	Restores the contained string from an element.
	@param elem the element containing the string
	@throws IllegalArgumentException if the element is wrong or ill-formed (tag name not "XMLString" 
	or does not contain only a CDATASection)
	*/
	public void decodeXMLElement(Element elem) throws IllegalArgumentException
	{
	    if(!elem.getTagName().equals("XMLString"))
	        throw new IllegalArgumentException("Element is not a XMLString!");
	    if(elem.getChildNodes().getLength()>1)
	        throw new IllegalArgumentException("Element has more than one child!");
	    if(elem.getChildNodes().getLength()<1)
	        str="";
	    else
	    {
    	    if(!(elem.getFirstChild() instanceof CDATASection))
	            throw new IllegalArgumentException("Element does not contain a CDATA section!");
	        str=XMLUtils.CDATAUnquoteText(((CDATASection)elem.getFirstChild()).getData());
	    }
	}	
	
	/**
	Retrieve the contained string.
	@return the contained string
	*/
    public String toString()
    {
        return str;
    }

    /**
    Compares two XMLString objects
    @param obj the object to compare to this one
    */
    public boolean equals(Object obj)
    {
        return (obj instanceof XMLString) && str.equals(((XMLString)obj).toString()) ||
               (obj instanceof String) && str.equals((String)obj);
    }
}