/*
 * @(#)XMLUtils.java
 *
 * Miscellaneous XML utilities
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.Util.XML;

import java.util.Random;

/**
Miscellaneous utility functions connected with XML: Quoting of text for use in a CDATA-Section, conversion to
of byte arrays to a hex-string and back, and comparing to byte arrays.
@author Michael Sonntag
@version 1.0
*/
public class XMLUtils
{
    /**
    Converts a string, so that it can be used in a CDATA section without problems. This means, it stays
    exactly the same, only "]]>" has to be changed (or the section would be ended prematurely on parsing).
    <br>Actual implementation: "]]>" is replaced by "].]>". If "].]>" does occur in the text, it is 
    doubled ("].]>].]>"). This allows recreating the original text and avoiding the end of the CDATA 
    section code.
    @param the string to quote
    @return the quoted string
    @see XMLUtils#CDATAUnquoteText(String)
    */
    public final static String CDATAQuoteText(String str)
    {
        String res="";
        while(str.indexOf("]]>")>=0 || str.indexOf("].]>")>=0)
        {
            if(str.indexOf("]]>")>=0)
            {
                res+=str.substring(0,str.indexOf("]]>"))+"].]>";
                str=str.substring(str.indexOf("]]>")+3);
            }
            else
            {
                res+=str.substring(0,str.indexOf("].]>"))+"].]>].]>";
                str=str.substring(str.indexOf("].]>")+4);
            }
        }
        res+=str;
        return res;
    }

    /**
    Converts a string from quoted back to normal.
    @param the string to unquote
    @return the original string
    @see XMLUtils#CDATAQuoteText(String)
    */
    public final static String CDATAUnquoteText(String str)
    {
        String res="";
        while(str.indexOf("].]>")>=0)
        {
            if(str.indexOf("].]>")+4<str.length() && str.substring(str.indexOf("].]>")+4).equals("].]>"))
            {   // Doubled up because we found this string during encoding
                res+=str.substring(0,str.indexOf("].]>")+4);
                str=str.substring(str.indexOf("].]>")+8);
            }
            else
            {
                res+=str.substring(0,str.indexOf("].]>"))+"]]>";
                str=str.substring(str.indexOf("].]>")+4);
            }
        }
        res+=str;
        return res;
    }
    
    /**
    Converts an array of bytes to a hex-string.
    @param data the array to convert
    @return a string with the data in hexadecimal representation
    */
    public final static String toHexString(byte[] data)
    {
        String res="";
        for(int i=0;i<data.length;i++)
        {
            char ch;
            int d=data[i];
            if(d<0)
                d+=256;
            if(d/16>9)
                ch=(char)('A'+(d/16-10));
            else
                ch=(char)('0'+(d/16));
            res+=ch;
            if(d%16>9)
                ch=(char)('A'+(d%16-10));
            else
                ch=(char)('0'+(d%16));
            res+=ch;
        }        
        return res.toUpperCase();
    }

    /**
    Converts a string of hexadecimal characters into an array of bytes.
    @param str the string to convert
    @return an array of bytes
    @throws IllegalArgumentException if the string is not a valid hexadecimal string
    */
    public final static byte[] toByteArray(String str)
    {
        str=str.toUpperCase().trim();
        if(str.length()%2!=0)
            throw new IllegalArgumentException("String must have even number of characters!");
        int len=str.length()/2;
        byte[] res=new byte[len];
        for(int i=0;i<str.length();i+=2)
        {
            char ch=str.charAt(i);
            int val=0;
            if(ch>='A')
                val=ch-'A'+10;
            else
                val=ch-'0';
            res[i/2]=(byte)(16*val);
            ch=str.charAt(i+1);
            if(ch>='A')
                val=ch-'A'+10;
            else
                val=ch-'0';
            res[i/2]+=(byte)val;
        }
        return res;
    }
    
    /**
    Compares to byte arrays. The arrays must have the same length and contain exactly the same data.
    @param ar1 the first array to compare
    @param ar2 the second array to compare
    @return <code>true</code> if both arrays have the same length and contain the same bytes
    */
    public final static boolean equalByteArrays(byte[] ar1, byte[] ar2)
    {
        if(ar1.length!=ar2.length)
            return false;
        for(int i=0;i<ar1.length;i++)
            if(ar1[i]!=ar2[i])
                return false;
        return true;
    }
    
    public static void main(String[] args)
    {
        byte[] tmp=new byte[16];
        Random rnd=new Random();
        rnd.nextBytes(tmp);
        System.out.print("Original data: ");
        for(int i=0;i<15;i++)
            System.out.print(""+tmp[i]+", ");
        System.out.println(""+tmp[15]);
        System.out.println("Encoded: "+toHexString(tmp));
        byte[] res=toByteArray(toHexString(tmp));
        System.out.print("Recreated data: ");
        for(int i=0;i<15;i++)
            System.out.print(""+res[i]+", ");
        System.out.println(""+res[15]);
        System.exit(0);
    }
    
}