/*
 * @(#)AnonymousInvoice.java
 *
 * Classes for transferring value (a.k.a. money): Paying for things
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.payment;

import java.io.*;
import FIM.Util.XML.XMLEncodeable;
import FIM.Util.XML.XMLUtils;
import java.util.Date;

import java.security.cert.Certificate;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.Signature;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidKeyException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateFactory;
import java.security.NoSuchProviderException;
import java.security.SignatureException;
import java.text.DateFormat;

import org.xml.sax.SAXException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import iaik.asn1.*;

/*
// Needed for testing; see main at end
import org.apache.xalan.serialize.Serializer;
import org.apache.xalan.serialize.SerializerFactory;
import org.apache.xalan.templates.OutputProperties;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import PkgAgentSystem.Crypto.*;
import FIM.Util.Crypto.NamedKeyAndCertificate;
*/

/**
An invoice for an anonymous recipient.
@author Michael Sonntag
@version 1.0
*/
public class AnonymousInvoice implements Invoice
{
    /** The date of the invoice. */
    private Date date;
    /** The title of the invoice. */
    private String title;
    /** The items the invoice contains of. */
    private InvoiceItem[] items;
    /** The issuer of the invoice. */
    private String issuer;
    /** Teh certificate of the issuer of the invoice. */
    private Certificate issuerCert;

    /** The actual siganture. */
    private byte[] signature=null;
    /** The algorithm used for signing. */
    private String signAlgorithm=null;
    /** The (optional) certificate associated with the signature. */
    private Certificate certificate=null;

    /**
    Create a new anonymous invoice. The currencies of the items must all be the same.
    Not both issuer and issuerCert may be null.
    @param issuer the issuer of the invoice
    @param issuerCert the certificate of the issuer of the invoice
    @param data the date for the invoice
    @param title the title of the invoice
    @param item an array of items for the invoice. At least one required
    */
    public AnonymousInvoice(String issuer,Certificate issuerCert,Date date,String title,InvoiceItem[] items)
    {
        if(issuer==null && issuerCert==null)
            throw new IllegalArgumentException("Issuer name or certificate must be provided");
        if(issuer==null)
            issuer=issuerCert.toString();
        this.issuer=issuer;
        this.issuerCert=issuerCert;
        if(date==null)
            date=new Date();
        this.date=date;
        if(title==null)
            throw new IllegalArgumentException("Title must be provided");
        this.title=title;
        if(items==null || items.length<1)
            throw new IllegalArgumentException("At least 1 item must be provided");
        // Check currencies: Must all be the same
        Currency curr=items[0].getPrice().getCurrency();
        for(int i=1;i<items.length;i++)
            if(!(items[i].getPrice().getCurrency().equals(curr)))
                throw new IllegalArgumentException("Currency of ALL items must be the same (Item "+(i+1)+"differs from first item)");
        this.items=items;
    }

    /**
    Creates a new empty invoice. Necessary so subclasses can create an empty object and then parse in the values from XML. Subclasses doing this
    must call <code>decodeXMLElement</code>.
    */
    protected AnonymousInvoice()
    {
    }

    /**
    Create a new anonymous invoice from an XML element.
    @param elem the XML element containing the encoded form
    */
    public AnonymousInvoice(Element elem)
    {
        decodeXMLElement(elem);
    }

    /**
    Retrieve the issuer of the invoice as a string.
    @return the issuer of the invoice
    */
    public String getIssuer()
    {
        return issuer;
    }

    /**
    Retrieve the issuer of the invoice as a certificate (might be <code>null</code>).
    @return the issuer of the invoice or <code>null</code>
    */
    public Certificate getIssuerCertificate()
    {
        return issuerCert;
    }

    /**
    Retrieve the date of the invoice.
    @return the date of the invoice
    */
    public Date getDate()
    {
        return date;
    }

    /**
    Retrieve the total value of all items in the invoice.
    @return the total price
    */
    public Price getTotalPrice()
    {
        double total=0.0d;
        for(int i=0;i<items.length;i++)
            total+=items[i].getPrice().getPrice();
        return new Price(items[0].getPrice().getCurrency(),total);
    }

    /**
    Retrieve the title of the invoice.
    @return the title of the invoice
    */
    public String getTitle()
    {
        return title;
    }

    /**
    Retrieve the items contained within the invoice (items or lines).
    @return the items of the invoice
    */
    public InvoiceItem[] getItems()
    {
        return (InvoiceItem[])items.clone();
    }

    /**
    Check whether this invoice is signed or not.
    @return <code>true</code> if the invoice is signed
    */
    public boolean isSigned()
    {
        return signature!=null;
    }

    /**
    Sign the invoice.
    @param signAlgorithm the algorithm to be used for signing
    @param pk the private key to be used for signing
    @param cert the certificate of the public key associated with the private key (can be <code>null</code>)
    @throws InvalidKeyException if the private key is invalid
    @throws NoSuchAlgorithmException if the algorithm is not supported
    */
    public void sign(String signAlgorithm,PrivateKey pk,Certificate cert) throws InvalidKeyException,NoSuchAlgorithmException
    {
        certificate=cert;
   		Signature sig=Signature.getInstance(signAlgorithm);
    	sig.initSign(pk);
   		try
   		{
    		sig.update(getEncoded());
	    	signature=sig.sign();
            this.signAlgorithm=signAlgorithm;
        }
		catch(SignatureException e)
        {   // Shouldn't happen: we initialized all
            System.err.println(e.toString());
		}
    }

    /**
    Retrieve the certificate passed in upon signing. Is <code>null</code> if non provided then.
    @return the certificate of the signer or <code>null</code>
    */
    public Certificate getCertificate()
    {
        return certificate;
    }

    /**
    Verify the signature of the invoice based on the certificate contained. Will fail if not signed or if no certificate is present.
    @param provider the security provider to be used for checking. if <code>null</code>, the standard provider is used
    @throws InvalidKeyException if the public key is invalid
    @throws NoSuchAlgorithmException if the algorithm used for signing is not supported
    @throws InvalidKeyException if the key in the certificate is erroneous
    @throws NoSuchProviderException if the security provider could not be found
    @throws SignatureException if the signature is invalid
    */
    public void verifySignature(String provider) throws NoSuchAlgorithmException,InvalidKeyException,NoSuchProviderException,SignatureException
    {
        verifySignature(provider,null);
    }

    /**
    Verify the signature of the invoice based on the public key passed in. Will fail if not signed. Teh certificate included is ignored.
    @param provider the security provider to be used for checking. if <code>null</code>, the standard provider is used
    @param pk the public key to use for verification. If <code>null</code>, the key from the certificate is used (if no certificate in this case, verification fails)
    @throws NoSuchAlgorithmException if the algorithm used for signing is not supported
    @throws InvalidKeyException if the public key (or the key in the certificate) is invalid
    @throws NoSuchProviderException if the security provider could not be found
    @throws SignatureException if the signature is invalid
    */
    public void verifySignature(String provider,PublicKey pk) throws NoSuchAlgorithmException,InvalidKeyException,NoSuchProviderException,SignatureException
    {
        if(pk==null && certificate==null)
            throw new SignatureException("Cannot verify signature: No certificate included and no public key specified");
        if(pk==null)
            pk=certificate.getPublicKey();
	    if(signature==null)
    		throw new SignatureException("Not signed");
   		Signature sig=null;
   		if(provider==null)
        	sig=Signature.getInstance(signAlgorithm);
   		else
        	sig=Signature.getInstance(signAlgorithm,provider);
    	sig.initVerify(pk);
   		sig.update(getEncoded());
    	if(!sig.verify(signature))
    		throw new SignatureException("Signature invalid");
    }

    /**
    Retrieve the invoice in encoded form to be used for signing. Does NOT include the signing information itself (which
    will be created based on the result!). Must be exactly the same on all calls and if the content is the same.
    @return the invoice in encoded form
    */
    public byte[] getEncoded()
    {
        SEQUENCE encoded=new SEQUENCE();
        encoded.addComponent(new PrintableString(issuer));
        if(issuerCert!=null)
        {
            try
            {
                encoded.addComponent(new OCTET_STRING(issuerCert.getEncoded()));
            }
            catch(CertificateEncodingException e)
            {
                System.err.println("Error encoding certificate for signing invoice: "+e.getMessage());
            }
        }
        encoded.addComponent(new PrintableString(title));
        encoded.addComponent(new PrintableString(""+date.getTime()));
        encoded.addComponent(new PrintableString(items[0].getPrice().getCurrency().getShortName()));
        encoded.addComponent(new INTEGER(items.length));
        for(int i=0;i<items.length;i++)
        {
            encoded.addComponent(new PrintableString(items[i].getItemName()));
            encoded.addComponent(new PrintableString(""+(int)(items[i].getPrice().getPrice()*100)));
        }
	    byte[] res=DerCoder.encode(encoded);
	    return res;
    }

    /**
    Retrieve the invoice as a string, suitable for presenting to the user in a very simple form.
    @return the invoice encoded as a string
    */
    public String toString()
    {
        String str="INVOICE from "+date.toString()+"\nIssued by: "+issuer+"\n\n"+title+"\n\n";
        for(int i=0;i<items.length;i++)
            str+=items[i].toString()+"\n";
        str+="Taxes:\n";
        str+="--------------------\nTotal: "+getTotalPrice().toString();
        if(signature!=null)
        {
            str+="\n\nInvoice is signed";
            if(certificate!=null)
                str+="\nCertificate present: "+certificate.toString();
        }
        return str;
    }

    /**
    Retrieve the invoice as a serializable (just returns <code>this</code> as this object is directly serializable).
    @return the price object as a serializable
    */
    public Serializable getAsSerializable()
	{
	    return this;
	}

    /**
    Retrieve the invoice as an XML element (name: AnonymousInvoice).
    @param doc the document within which to create the element
    @return the price as an element
    */
	public Element getAsXMLElement(Document doc)
	{
	    Element el=doc.createElement("AnonymousInvoice");
	    el.setAttribute("Title",title);
	    el.setAttribute("Date",""+date.getTime());
	    el.setAttribute("Issuer",""+issuer);
	    Element itemEl=doc.createElement("Items");
	    el.appendChild(itemEl);
	    for(int i=0;i<items.length;i++)
	        itemEl.appendChild(items[i].getAsXMLElement(doc));
	    if(issuerCert!=null)
	    {
        	Element certEl=doc.createElement("Certificate");
        	certEl.setAttribute("Type",issuerCert.getType());
	        try
	        {
	            certEl.appendChild(doc.createTextNode(XMLUtils.toHexString(issuerCert.getEncoded())));
	        }
	        catch(CertificateEncodingException e)
	        {   // Just leave it out then; shouldn't happen anyway
	            System.err.print("");
	        }
    	    el.appendChild(certEl);
	    }
	    if(signature!=null)
	    {
    	    Element sigEl=doc.createElement("Signature");
	        sigEl.setAttribute("SignatureAlgorithm",signAlgorithm);
	        sigEl.setAttribute("Signature",XMLUtils.toHexString(signature));
	        if(certificate!=null)
	        {
        	    Element certEl=doc.createElement("Certificate");
        	    certEl.setAttribute("Type",certificate.getType());
	            try
	            {
	                certEl.appendChild(doc.createTextNode(XMLUtils.toHexString(certificate.getEncoded())));
	            }
	            catch(CertificateEncodingException e)
	            {   // Just leave it out then; shouldn't happen anyway
	                System.err.print("");
	            }
    	        sigEl.appendChild(certEl);
	        }
	        el.appendChild(sigEl);
	    }
	    return el;
	}

    /**
    Parse back the invoice from an XML element.
    @param elem the element containing the price to parse back
    @throws IllegalArgumentException if the element is incorrect (e. g. wrong tag name)
    */
	public void decodeXMLElement(Element elem) throws IllegalArgumentException
	{
	    title=elem.getAttribute("Title");
	    issuer=elem.getAttribute("Issuer");
	    String dateStr=elem.getAttribute("Date");
	    date=new Date();
	    date.setTime(Long.parseLong(dateStr));
	    Element itemEl=(Element)elem.getFirstChild();
	    items=new InvoiceItem[itemEl.getChildNodes().getLength()];
	    for(int i=0;i<items.length;i++)
	        items[i]=new InvoiceItem((Element)itemEl.getChildNodes().item(i));
	    Element sigEl=(Element)itemEl.getNextSibling();
	    if(sigEl!=null && sigEl.getTagName().equals("Certificate"))
	    {
	        byte[] buf=XMLUtils.toByteArray(sigEl.getFirstChild().getNodeValue());
	        try
	        {
            	CertificateFactory cf=CertificateFactory.getInstance(sigEl.getAttribute("Type"));
    	        ByteArrayInputStream bais=new ByteArrayInputStream(buf);
	            issuerCert=cf.generateCertificate(bais);
	        }
	        catch(CertificateException e)
	        {
	            throw new RuntimeException("Could not recreate certificate");
	        }
	        sigEl=(Element)sigEl.getNextSibling();
	    }
	    if(sigEl!=null)
	    {
	        signAlgorithm=sigEl.getAttribute("SignatureAlgorithm");
	        signature=XMLUtils.toByteArray(sigEl.getAttribute("Signature"));
	        if(sigEl.getFirstChild()!=null)
	        {
	            byte[] buf=XMLUtils.toByteArray(sigEl.getFirstChild().getFirstChild().getNodeValue());
	        	try
	        	{
            		CertificateFactory cf=CertificateFactory.getInstance(((Element)sigEl.getFirstChild()).getAttribute("Type"));
    	        	ByteArrayInputStream bais=new ByteArrayInputStream(buf);
	                certificate=cf.generateCertificate(bais);
	            }
	            catch(CertificateException e)
	            {
	                throw new RuntimeException("Could not recreate certificate");
	            }
	        }
	    }
	}

    private void writeObject(java.io.ObjectOutputStream out) throws IOException
    {
        out.writeUTF(issuer);
        if(issuerCert!=null)
        {
            out.writeBoolean(true);
            out.writeUTF(issuerCert.getType());
            try
            {
                out.writeObject(issuerCert.getEncoded());
            }
            catch(CertificateEncodingException e)
            {
	            throw new NotSerializableException(e.toString());
	        }
        }
        else
            out.writeBoolean(false);
        out.writeUTF(title);
        out.writeInt(items.length);
        for(int i=0;i<items.length;i++)
            out.writeObject(items[i]);
        out.writeLong(date.getTime());
        if(signature!=null)
        {
            out.writeBoolean(true);
            out.writeUTF(signAlgorithm);
            out.writeObject(signature);
            if(certificate!=null)
            {
                out.writeBoolean(true);
                out.writeUTF(certificate.getType());
                try
                {
                    out.writeObject(certificate.getEncoded());
                }
                catch(CertificateEncodingException e)
                {
	                throw new NotSerializableException(e.toString());
	            }
            }
            else
                out.writeBoolean(false);
        }
        else
            out.writeBoolean(false);
    }

    private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException
    {
        issuer=in.readUTF();
        if(in.readBoolean())
        {
            String type=in.readUTF();
            byte[] bytes=(byte[])in.readObject();
            try
            {
                CertificateFactory cf=CertificateFactory.getInstance(type);
    	        ByteArrayInputStream bais=new ByteArrayInputStream(bytes);
	            issuerCert=cf.generateCertificate(bais);
	        }
	        catch(CertificateException e)
	        {
	            throw new IOException(e.toString());
	        }
        }
        title=in.readUTF();
        int count=in.readInt();
        items=new InvoiceItem[count];
        for(int i=0;i<count;i++)
            items[i]=(InvoiceItem)in.readObject();
        date=new Date();
        date.setTime(in.readLong());
        if(in.readBoolean())
        {
            signAlgorithm=in.readUTF();
            signature=(byte[])in.readObject();
            if(in.readBoolean())
            {
                String type=in.readUTF();
                byte[] bytes=(byte[])in.readObject();
                try
                {
                    CertificateFactory cf=CertificateFactory.getInstance(type);
    	            ByteArrayInputStream bais=new ByteArrayInputStream(bytes);
	                certificate=cf.generateCertificate(bais);
	            }
	            catch(CertificateException e)
	            {
	                throw new IOException(e.toString());
	            }
	        }
	    }
    }

/*
	public static void main(String[] args) throws Exception
	{
	    InvoiceItem[] items=new InvoiceItem[2];
	    items[0]=new InvoiceItem("Players Handbook",new Price(Currency.findCurrency("USD"),19.9));
	    items[1]=new InvoiceItem("Monstrous Manual",new Price(Currency.findCurrency("USD"),24.9));
	    InvoiceItem[] taxes=new InvoiceItem[1];
	    taxes[0]=new InvoiceItem("VAT",new Price(Currency.findCurrency("USD"),8.96));
	    AnonymousInvoice inv=new AnonymousInvoice(null,"Test-Invoice",items,taxes);
	    // Next 2 lines needed for loading the security store with agent certificates in it
        java.net.URL.setURLStreamHandlerFactory(new PkgAgentSystem.AMPStreamHandlerFactory());
	    java.security.Security.addProvider(new PkgAgentSystem.Crypto.AgentCryptographyProvider());
        PersonalSecurityStore store=PersonalSecurityStoreFactory.openPersonalSecurityStore("d:\\data\\java\\agents\\agenthome\\keystore.store","password".toCharArray());
	    NamedKeyAndCertificate nkc=store.getOwnerPrivateKeyAndCertificate("Server Certificate");
        inv.sign("MD5/RSA",nkc.getPrivateKey(),nkc.getCertificate());
        inv.verifySignature(null);
	    System.out.println("Original:\n"+inv.toString()+"\n\n");
        String str="<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<Invoice/>";
        DocumentBuilderFactory dFactory=DocumentBuilderFactory.newInstance();
        DocumentBuilder dBuilder=dFactory.newDocumentBuilder();
        Document document=dBuilder.parse(new java.io.ByteArrayInputStream(str.getBytes()));
        document.getDocumentElement().appendChild(inv.getAsXMLElement(document));
        ByteArrayOutputStream bos=new ByteArrayOutputStream();
        OutputProperties prop=new OutputProperties(OutputProperties.getDefaultMethodProperties("xml"));
        Serializer serializer=SerializerFactory.getSerializer(prop.getProperties());
        serializer.setOutputStream(bos);
        serializer.asDOMSerializer().serialize(document);
        System.out.println(bos.toString());
        inv=new AnonymousInvoice((Element)document.getDocumentElement().getFirstChild());
        System.out.println("\n\nParsed back:\n"+inv);
	}
*/
}