/*
 * @(#)CreditCardPayment.java
 *
 * Classes for transferring value (a.k.a. money): Paying for things
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.payment;

import java.util.Date;
import java.util.Calendar;
import java.io.*;
import FIM.Util.XML.XMLEncodeable;
import FIM.Util.XML.XMLUtils;

import java.security.cert.Certificate;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.Signature;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidKeyException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateFactory;
import java.security.NoSuchProviderException;
import java.security.SignatureException;
import java.text.DateFormat;

import org.xml.sax.SAXException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
A class representing the payment by creditcard.
@author Michael Sonntag
@version 1.0
@see FIM.payment.Invoice
*/
public class CreditCardPayment extends PaymentBase
{
    /** The name of the company which issued the card */
    private String cardCompany=null;
    /** The name on the card (=holder) */
    private String nameOnCard=null;
    /** The number on the card */
    private byte[] number=null;
    /** The expiry date of the card */
    private Date expiry=null;

    /**
    Create a new empty object. To be used for parsing in XML data later. Must be public so it can be
    created by factory methods.
    */
    public CreditCardPayment()
    {
    }

    /**
    Create a new creditcard payment.
    @param invoice the invoice this payment is for
    @param cardCompany the company which issued the card
    @param nameOnCard the name on the card (holder)
    @param number the number of the creditcard as an array of bytes
    @param expiry the expiry date
    */
    public CreditCardPayment(Invoice inv,String cardCompany,String nameOnCard,byte[] number,Date expiry)
    {
        super(inv,"Creditcard");
        if(cardCompany==null || cardCompany.length()<1)
            throw new IllegalArgumentException("Name of the credit card company must be provided");
        this.cardCompany=cardCompany;
        if(nameOnCard==null || nameOnCard.length()<1)
            throw new IllegalArgumentException("Name on the credit card must be provided");
        this.nameOnCard=nameOnCard;
        if(number==null || number.length<1)
            throw new IllegalArgumentException("Number of the credit card must be provided");
        this.number=number;
        if(expiry==null)
            throw new IllegalArgumentException("Expiry date must be provided");
        this.expiry=expiry;
    }

    /**
    Create a new data payment from an XML element.
    @param elem the XML element containing the encoded form
    */
    public CreditCardPayment(Element elem)
    {
        decodeXMLElement(elem);
    }

    /**
    Retrieve the payment as a string, suitable for presenting to the user in a very simple form.
    @return the payment encoded as a string
    */
    public String toString()
    {
        String str="Payment for Invoice from "+getInvoice().getDate().toString()+" ("+getInvoice().getTitle()+")\n\n";
        str+="Payment by Creditcard from "+cardCompany+", Name on card: "+nameOnCard+" (";
        for(int i=0;i<number.length;i++)
        {
            str+=""+number[i];
            if((i+1)%4==0)
                str+=" ";
        }
        Calendar cal=Calendar.getInstance();
        cal.setTime(expiry);
        str+=", "+(cal.get(Calendar.MONTH)+1)+"/"+cal.get(Calendar.YEAR)+")\n";
        return str;
    }

    /**
    Retrieve the payment in encoded form to be used for signing. Does NOT include the signing information itself (which
    will be created based on the result!). Must be exactly the same on all calls and if the content is the same.
    @return the payment in encoded form
    */
    public byte[] getEncoded()
    {
        ByteArrayOutputStream bos=new ByteArrayOutputStream();
        try
        {
            ObjectOutputStream out=new ObjectOutputStream(bos);
            out.writeObject(super.getEncoded());
            out.writeUTF(cardCompany);
            out.writeUTF(nameOnCard);
            out.writeObject(number);
            Calendar cal=Calendar.getInstance();
            cal.setTime(expiry);
            out.writeUTF(""+(cal.get(Calendar.MONTH)+1)+"/"+cal.get(Calendar.YEAR));
            out.close();
        }
        catch(IOException e)
        {   // Shouldn't happen
            System.err.println(e.toString());
        }
        return bos.toByteArray();
    }


    /**
    Retrieve the payment as an XML element.
    @param doc the document within which to create the element
    @return the currency as an element
    */
	public Element getAsXMLElement(Document doc)
	{
	    Element el=doc.createElement("CreditCardPayment");
	    el.appendChild(super.getAsXMLElement(doc));
	    Element ccEl=doc.createElement("CreditCard");
	    Element companyEl=doc.createElement("CardCompany");
	    companyEl.appendChild(doc.createTextNode(cardCompany));
	    Element nameEl=doc.createElement("NameOnCard");
	    nameEl.appendChild(doc.createTextNode(nameOnCard));
	    Element numberEl=doc.createElement("Number");
	    String nb="";
        for(int i=0;i<number.length;i++)
            nb+=""+number[i];
	    numberEl.appendChild(doc.createTextNode(nb));
	    Element expiryEl=doc.createElement("ExpiryDate");
	    Calendar cal=Calendar.getInstance();
        cal.setTime(expiry);
	    expiryEl.appendChild(doc.createTextNode(""+(cal.get(Calendar.MONTH)+1)+"/"+cal.get(Calendar.YEAR)));
	    ccEl.appendChild(companyEl);
	    ccEl.appendChild(nameEl);
	    ccEl.appendChild(numberEl);
	    ccEl.appendChild(expiryEl);
	    el.appendChild(ccEl);
	    return el;
	}

    /**
    Parse back the payment from an XML element.
    @param elem the element containing the currency to parse back
    @throws IllegalArgumentException if the element is incorrect (e. g. wrong tag name)
    */
	public void decodeXMLElement(Element elem) throws IllegalArgumentException
	{
	    super.decodeXMLElement((Element)elem.getFirstChild());
	    Element ccEl=(Element)elem.getFirstChild().getNextSibling();
	    cardCompany=ccEl.getFirstChild().getFirstChild().getNodeValue();
	    nameOnCard=ccEl.getFirstChild().getNextSibling().getFirstChild().getNodeValue();
	    String nb=ccEl.getFirstChild().getNextSibling().getNextSibling().getFirstChild().getNodeValue().trim();
	    number=new byte[nb.length()];
	    for(int i=0;i<number.length;i++)
	        number[i]=Byte.parseByte(""+nb.charAt(i));
	    String exp=ccEl.getLastChild().getFirstChild().getNodeValue();
	    Calendar cal=Calendar.getInstance();
	    cal.set(Integer.parseInt(exp.substring(exp.indexOf("/")+1)),Integer.parseInt(exp.substring(0,exp.indexOf("/")))-1,1);
	    expiry=cal.getTime();
	}

}