/*
 * @(#)Currency.java
 *
 * Classes for transferring value (a.k.a. money): Paying for things
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.payment;

import java.io.Serializable;
import java.io.IOException;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Locale;
import java.text.NumberFormat;
import java.text.ParseException;

import FIM.Util.XML.XMLEncodeable;

import org.xml.sax.SAXException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
A class representing a currency. Also includes a (fixed) conversion factor.
<br>Includes a static list of the currencies available.
<br>Currently includes Euro, ATS, DEM and US$
@author Michael Sonntag
@version 1.0
*/
public class Currency implements XMLEncodeable,Serializable
{
    /**
    The short name of the currency.
    */
    private String shortName=null;

    /**
    The long name of the currency.
    */
    private String longName=null;

    /**
    The list of abbreviations of the currency.
    */
    protected String[] abbrev=null;

    /**
    If true, the currency is in front of values, otherwise behind.
    */
    private boolean pre=true;

    /**
    The locale for parsing values in this currency.
    */
    private Locale loc=null;

    /**
    The exchange rate to one Euro.
    */
    private double exchangeRateToStandard=0.0d;

    /**
    The static list of currencies available.
    */
    protected static Vector currencies=new Vector();

    /** Abbreviations for USD */
    private final static String[] CURRENCY_USD_DEF={"USD","US$","$","Dollar"};
    /** Abbreviations for DEM */
    private final static String[] CURRENCY_DEM_DEF={"DM","DEM","Mark"};
    /** Abbreviations for ATS */
    private final static String[] CURRENCY_ATS_DEF={"ATS","S","S","Schilling"};
    /** Abbreviations for Euro */
    private final static String[] CURRENCY_EUR_DEF={"EUR","ECU","Euro"};

    /**
    Creates the four standard objects and adds them to the static list.
    */
    static
    {
        currencies.add(new Currency("ATS","Schilling",true,CURRENCY_ATS_DEF,Locale.GERMANY,13.7603));
        currencies.add(new Currency("EUR","Euro",false,CURRENCY_EUR_DEF,Locale.GERMANY,1.0));
        currencies.add(new Currency("DEM","Deutsche Mark",true,CURRENCY_DEM_DEF,Locale.GERMANY,1.95583));
        currencies.add(new Currency("USD","Dollar",true,CURRENCY_USD_DEF,Locale.US,0.89219942));
    }

    /**
    Creates a new currency object.
    @param shortName the short name for the currency
    @param longName the long name for the currency
    @param pre if <code>true</code>, the currency is in front of the value, otherwise behind
    @param abbrev the list of abbreviations for this currency (may be empty or <code>null</code>)
    @param loc the locale to use for parsing values in this currency
    @param exchangeRate the exchange rate to 1 Euro (xxxx units of this currency are 1 Euro); provide 0.0 if not applicable or unknown
    */
    public Currency(String shortName,String longName,boolean pre,String[] abbrev,Locale loc,double exchangeRate)
    {
        if(shortName==null || shortName.length()!=3)
            throw new IllegalArgumentException("Short name must consist of 3 characters");
        this.shortName=shortName;
        if(longName==null || longName.length()<1)
            throw new IllegalArgumentException("Long name must be provided");
        this.longName=longName;
        this.pre=pre;
        if(abbrev==null || abbrev.length<1)
        {
            this.abbrev=new String[1];
            this.abbrev[0]=shortName;
        }
        else
            this.abbrev=abbrev;
        if(loc==null)
            throw new IllegalArgumentException("Locale for parsing must be provided");
        this.loc=loc;
        if(exchangeRate<0.0d)
            throw new IllegalArgumentException("The exchange rate must be >=0.0!");
        exchangeRateToStandard=exchangeRate;
    }

    /**
    Creates a new currency object from values stored in an XML Element.
    @param elem the XML element to parse the values from
    @throws IllegalArgumentException if the currency could not be parsed back completely
    */
    public Currency(Element elem)
    {
        decodeXMLElement(elem);
        if(shortName==null || shortName.length()!=3)
            throw new IllegalArgumentException("Short name must consist of 3 characters");
        if(longName==null || longName.length()<1)
            throw new IllegalArgumentException("Long name must be provided");
        if(loc==null)
            throw new IllegalArgumentException("Locale for parsing must be provided");
        if(exchangeRateToStandard<0.0d)
            throw new IllegalArgumentException("The exchange rate must be >=0.0!");
    }

    /**
    Set the exchange rate for this currency to the Euro.
    xxxx units of this currency are 1 Euro.
    If no applicable or unknown, provide 0.0d.
    @param exchangeRate the factor for conversion (msut be >=0.0)
    */
    public void setExchangeRateToStandard(double exchangeRate)
    {
        if(exchangeRate<0.0d)
            throw new IllegalArgumentException("The exchange rate must be >=0.0!");
        exchangeRateToStandard=exchangeRate;
    }

    /**
    Retrieve the exchange rate for this currency to the Euro.
    xxxx units of this currency are 1 Euro.
    @return the factor for conversion or 0.0 if not known/applicable
    */
    public double getExchangeRateToStandard()
    {
        return exchangeRateToStandard;
    }

    /**
    Retrieve the abbreviations of this currency.
    @return the array of abbreviations of this corrency
    */
    public String[] getAbbreviations()
    {
        return (String[])abbrev.clone();
    }

    /**
    Retrieve the short name of this currency.
    @return the short name of this currency
    */
    public String getShortName()
    {
        return shortName;
    }

    /**
    Retrieve the long name of this currency.
    @return the long name of this currency
    */
    public String getLongName()
    {
        return longName;
    }

    /**
    Retrieve the locale to be used for parsing values of this currency.
    @return the locale for this currency
    */
    public Locale getLocale()
    {
        return loc;
    }

    /**
    Retrieve whether this currency is a prefix to the value (e. g. US$ 30.99) or a postfix (e. g. 19,99 S).
    @return <code>true</code> if the currency text is in front of the value
    */
    public boolean isPreValue()
    {
        return pre;
    }

    /**
    Returns a string representation of the object; in this case the short name of the currency.
    @return the short name of the currency
    */
    public String toString()
    {
        return shortName;
    }


    /**
    Matches a string to the abbreviations for this currency (case-sensitive).
    @param str the string to match
    @return <code>true</code> if the string matches one of the abbrevations
    */
    protected boolean matchString(String str)
    {
        boolean res=false;
        for(int i=0;i<abbrev.length && !res;i++)
            res=res || abbrev[i].equals(str);
        return res;
    }

    /**
    Adds a currency to the list. If the currency already exists (exact match, excluding only the
    conversion factor!), the request is ignored.
    @param cur the currency to add.
    */
    public static void addCurrency(Currency cur)
    {
        synchronized(currencies)
        {
            boolean found=false;
            Enumeration enum=currencies.elements();
            while(!found && enum.hasMoreElements())
                found=cur.equals((Currency)enum.nextElement());
            if(!found)
                currencies.add(cur);
        }
    }

    /**
    Removes a currency from the list. Ignored if not in the list.
    @param cur the currency to remove
    */
    public static void removeCurrency(Currency cur)
    {
        synchronized(currencies)
        {
            currencies.remove(cur);
        }
    }

    /**
    Clears the list of all currencies.
    */
    public static void clearCurrencies()
    {
        synchronized(currencies)
        {
            currencies.clear();
        }
    }

    /**
    Retrieve the number of currencies in the list
    @return the number of currencies
    */
    public static int currenciesCount()
    {
        synchronized(currencies)
        {
            return currencies.size();
        }
    }

    /**
    Retrieve the index of a certain currency
    @param cur the currency to find
    @return the position of the currency in the list or -1 if not found
    */
    public static int currencyPos(Currency cur)
    {
        synchronized(currencies)
        {
            return currencies.indexOf(cur);
        }
    }

    /**
    Retrieve an array of all currencies available.
    @param an array of all currency-objects available
    */
    public static Currency[] getCurrencies()
    {
        synchronized(currencies)
        {
            Currency[] res=new Currency[currencies.size()];
            currencies.copyInto(res);
            return res;
        }
    }

    /**
    Finds the first currency equivalent to the string. Returns a currency object,
    where one of the abbreviations equals the string.
    If the occurs in the abbrveiations of more than one currency, the first one is returned
    (which is more or less random).
    @param str the string containing the value of the currency denominator
    @return <code>null</code> or the currency for the string
    */
    public static Currency findCurrency(String str)
    {
        Currency res=null;
        synchronized(currencies)
        {
            Enumeration enum=currencies.elements();
            while(enum.hasMoreElements() && res==null)
            {
                Currency cur=(Currency)enum.nextElement();
                if(cur.matchString(str))
                    res=cur;
            }
        }
        return res;
    }

    /**
    Parses the value from a string according to the currency ('.' or ',' as thousands-separator, ...).
    @param str the string containing the value to parse
    @return Double.NaN in case of error, otherwise the parsed value according to the currency
    */
    public double parseCurrencyValue(String str)
    {
        str=str.trim();
        Locale locOverride=null;
        // Lion.cc has e. g. ATS 259.00 although this should be 259,00
        // Try to compensate this through override
        if(str.length()>2 && str.charAt(str.length()-3)=='.')
            locOverride=Locale.UK;
        if(str.length()>2 && str.charAt(str.length()-3)==',')
            locOverride=Locale.GERMANY;
        NumberFormat nf=null;
        if(locOverride!=null)
            nf=NumberFormat.getInstance(locOverride);
        else
        {
            if(loc!=null)
                nf=NumberFormat.getInstance(loc);
            else
                nf=NumberFormat.getInstance();
        }
        Number n=null;
        try
        {
            n=nf.parse(str);
        }
        catch(ParseException e)
        {
            return Double.NaN;
        }
        return n.doubleValue();
    }

    /**
    Compares two currencies for equality. Includes all the data WITHOUT the exchange factor.
    Abbreviations need not be in the same order, but all must be included and no other may
    be included.
    @param obj the object this one to compare to
    @return <code>true</code> if both objects are equal
    */
    public boolean equals(Object obj)
    {
        if(!(obj instanceof Currency))
            return false;
        Currency cur=(Currency)obj;
        boolean res=shortName.equals(cur.getShortName()) && longName.equals(cur.getLongName()) &&
            pre==cur.isPreValue() && loc.equals(cur.getLocale()) && abbrev.length==cur.getAbbreviations().length;
        String[] curAbbrev=cur.getAbbreviations();
        for(int i=0;res && i<abbrev.length;i++)
        {
            res=false;
            for(int j=0;!res && j<curAbbrev.length;j++)
                res=abbrev[i].equals(curAbbrev[j]);
        }
        return res;
    }

    /**
    Retrieve the currency as a serializable (just returns <code>this</code> as this object is directly serializable).
    @return the currency object as a serializable
    */
    public Serializable getAsSerializable()
	{
	    return this;
	}

    /**
    Retrieve the currency as an XML element.
    @param doc the document within which to create the element
    @return the currency as an element
    */
	public Element getAsXMLElement(Document doc)
	{
	    Element el=doc.createElement("Currency");
	    Element abbrevEl=doc.createElement("Abbreviations");
	    el.setAttribute("ShortName",shortName);
	    el.setAttribute("LongName",longName);
	    el.setAttribute("Pre",pre?"TRUE":"FALSE");
	    el.setAttribute("ExchangeRateToStandard",""+exchangeRateToStandard);
	    Element locEl=doc.createElement("Locale");
        locEl.setAttribute("Country",loc.getCountry());
        locEl.setAttribute("Language",loc.getLanguage());
        locEl.setAttribute("Variant",loc.getVariant());
        for(int i=0;i<abbrev.length;i++)
        {
            Element abEl=doc.createElement("Abbreviation");
            abEl.appendChild(doc.createTextNode(abbrev[i]));
	        abbrevEl.appendChild(abEl);
	    }
	    el.appendChild(abbrevEl);
	    el.appendChild(locEl);
	    return el;
	}

    /**
    Parse back the currency from an XML element.
    @param elem the element containing the currency to parse back
    @throws IllegalArgumentException if the element is incorrect (e. g. wrong tag name)
    */
	public void decodeXMLElement(Element elem) throws IllegalArgumentException
	{
	    shortName=elem.getAttribute("ShortName");
	    longName=elem.getAttribute("LongName");
	    pre=elem.getAttribute("Pre").equalsIgnoreCase("TRUE");
	    try
	    {
	        exchangeRateToStandard=Double.parseDouble(elem.getAttribute("ExchangeRateToStandard"));
	    }
	    catch(NumberFormatException e)
	    {
	        throw new IllegalArgumentException("Could not parse exchange rate to standard ("+e.getMessage()+")");
	    }
	    Element abEl=(Element)elem.getFirstChild();
	    abbrev=new String[abEl.getChildNodes().getLength()];
	    for(int i=0;i<abbrev.length;i++)
	        abbrev[i]=abEl.getChildNodes().item(i).getFirstChild().getNodeValue();
	    Element locEl=(Element)elem.getLastChild();
	    String country=locEl.getAttribute("Country");
	    String language=locEl.getAttribute("Language");
	    String variant=locEl.getAttribute("Variant");
	    loc=new Locale(language,country,variant);
	}
}