/*
 * @(#)DataPayment.java
 *
 * Classes for transferring value (a.k.a. money): Paying for things
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.payment;

import java.util.Date;
import java.util.Calendar;
import java.io.*;
import FIM.Util.XML.*;

import java.security.cert.Certificate;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.Signature;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidKeyException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateFactory;
import java.security.NoSuchProviderException;
import java.security.SignatureException;
import java.text.DateFormat;

import org.xml.sax.SAXException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
A class representing the payment by providing some valuable piece of data. This could be either some kind 
of E-Cash or replacing the purchase with an exchange (data-for-data). The data can be one of the following:
<ul>
<li>A string
<li>An array of bytes
<li>A serializable object
</ul>
@author Michael Sonntag
@version 1.0
@see FIM.payment.Invoice
*/
public class DataPayment extends PaymentBase
{
    /** The type of the content: NONE */
    public final static int TYPE_EMPTY=0;
    /** The type of the content: String */
    public final static int TYPE_STRING=1;
    /** The type of the content: Array of bytes */
    public final static int TYPE_BYTES=2;
    /** The type of the content: Serializable object */
    public final static int TYPE_SERIALIZABLE=3;

    /** The content: String */
    private String stringData=null;
    /** The content: Array of bytes */
    private byte[] byteData=null;
    /** The content: Serializable object */
    private Serializable serialData=null;

    /**
    Create a new empty object. To be used for parsing in XML data later. Must be public so it can be
    created by factory methods.
    */
    public DataPayment()
    {
    }

    /**
    Create a new data payment with the type string.
    @param invoice the invoice this payment is for
    @param characteristic the description of the content
    @param stringData the string data
    */
    public DataPayment(Invoice inv,String characteristic,String stringData)
    {
        super(inv,characteristic);
        if(stringData==null || stringData.length()<1)
            throw new IllegalArgumentException("String data must be provided");
        this.stringData=stringData;
    }

    /**
    Create a new data payment with the type array of bytes.
    @param invoice the invoice this payment is for
    @param characteristic the description of the content
    @param byteData the byte data
    */
    public DataPayment(Invoice inv,String characteristic,byte[] byteData)
    {
        super(inv,characteristic);
        if(byteData==null || byteData.length<1)
            throw new IllegalArgumentException("Byte data must be provided");
        this.byteData=(byte[])byteData.clone();
    }

    /**
    Create a new data payment with the type serializable object.
    @param invoice the invoice this payment is for
    @param characteristic the description of the content
    @param serialData the serializable object
    */
    public DataPayment(Invoice inv,String characteristic,Serializable serialData)
    {
        super(inv,characteristic);
        if(serialData==null)
            throw new IllegalArgumentException("Serializable data must be provided");
        this.serialData=serialData;
    }

    /**
    Create a new data payment from an XML element.
    @param elem the XML element containing the encoded form
    */
    public DataPayment(Element elem)
    {
        decodeXMLElement(elem);
    }

    /**
    Retrieve the data in case of the string variant.
    @return the string data
    @throws IllegalStateException if this object ist not of the data subtype or not initialized
    */
    public String getStringData()
    {
        if(stringData==null)
            throw new IllegalStateException("Not initialized or not a string varaiant of data payment");
        return stringData;
    }

    /**
    Retrieve the data in case of the bytes variant.
    @return the array of bytes
    @throws IllegalStateException if this object ist not of the byte-array subtype or not initialized
    */
    public byte[] getByteData()
    {
        if(byteData==null)
            throw new IllegalStateException("Not initialized or not a byte array varaiant of data payment");
        return byteData;
    }

    /**
    Retrieve the data in case of the serializable object variant.
    @return the the serializable object
    @throws IllegalStateException if this object ist not of the serializable object subtype or not initialized
    */
    public Serializable getSerializableData()
    {
        if(serialData==null)
            throw new IllegalStateException("Not initialized or not a serializable object varaiant of data payment");
        return serialData;
    }

    /**
    Retrieve the type of this payment.
    @return the payment type
    @see #TYPE_EMPTY
    @see #TYPE_STRING
    @see #TYPE_BYTES
    @see #TYPE_SERIALIZABLE
    */
    public int getType()
    {
        if(stringData!=null)
            return TYPE_STRING;
        if(stringData!=null)
            return TYPE_BYTES;
        if(stringData!=null)
            return TYPE_SERIALIZABLE;
        return TYPE_EMPTY;
    }

    /**
    Retrieve the payment as a string, suitable for presenting to the user in a very simple form.
    @return the payment encoded as a string
    */
    public String toString()
    {
        String str="Payment for Invoice from "+getInvoice().getDate().toString()+" ("+getInvoice().getTitle()+")\n\n";
        str+="Payment by Data (Characteristic: "+characteristics()+"; Content: ";
        if(stringData!=null)
            str+="String: "+stringData;
        if(byteData!=null)
            str+="Array of bytes; Length: "+byteData.length;
        if(serialData!=null)
            str+="Serializable object; Class: "+serialData.getClass().getName();
        str=str+=")\n";
        return str;
    }

    /**
    Retrieve the payment in encoded form to be used for signing. Does NOT include the signing information itself (which
    will be created based on the result!). Must be exactly the same on all calls and if the content is the same.
    @return the payment in encoded form
    */
    public byte[] getEncoded()
    {
        ByteArrayOutputStream bos=new ByteArrayOutputStream();
        try
        {
            ObjectOutputStream out=new ObjectOutputStream(bos);
            out.writeObject(super.getEncoded());
            out.writeInt(getType());
            switch(getType())
            {
                case 1:
                    out.writeUTF(stringData);
                break;
                case 2:
                    out.writeObject(byteData);
                break;
                case 3:
                    out.writeObject(serialData);
                break;
            }
            out.close();
        }
        catch(IOException e)
        {   // Shouldn't happen
            System.err.println(e.toString());
        }
        return bos.toByteArray();
    }

    /**
    Retrieve the payment as an XML element.
    @param doc the document within which to create the element
    @return the currency as an element
    */
	public Element getAsXMLElement(Document doc)
	{
	    Element el=doc.createElement("DataPayment");
	    el.appendChild(super.getAsXMLElement(doc));
	    Element data=null;
        if(stringData!=null)
        {
            data=doc.createElement("StringData");
            XMLString xstr=new XMLString(stringData);
            data.appendChild(xstr.getAsXMLElement(doc));
        }
        else if(byteData!=null)
        {
            data=doc.createElement("ByteData");
            XMLByteArray xby=new XMLByteArray(byteData);
            data.appendChild(xby.getAsXMLElement(doc));
        }
        else
        {   // Serializable
            data=doc.createElement("SerialData");
            try
            {
                ByteArrayOutputStream bos=new ByteArrayOutputStream();
                ObjectOutputStream os=new ObjectOutputStream(bos);
                os.writeObject(serialData);
                bos.close();
                data.appendChild(doc.createTextNode(XMLUtils.toHexString(bos.toByteArray())));
            }
            catch(IOException e)
            {
System.err.println(e.toString());
            }
        }
        el.appendChild(data);
	    return el;
	}

    /**
    Parse back the payment from an XML element.
    @param elem the element containing the currency to parse back
    @throws IllegalArgumentException if the element is incorrect (e. g. wrong tag name)
    */
	public void decodeXMLElement(Element elem) throws IllegalArgumentException
	{
	    super.decodeXMLElement((Element)elem.getFirstChild());
	    Element dataEl=(Element)elem.getFirstChild().getNextSibling();
	    if(dataEl.getTagName().equals("StringData"))
	    {
	        XMLString xstr=new XMLString();
	        xstr.decodeXMLElement((Element)dataEl.getFirstChild());
	        stringData=xstr.getString();
	    }
	    else if(dataEl.getTagName().equals("ByteData"))
	    {
	        XMLByteArray xby=new XMLByteArray();
	        xby.decodeXMLElement((Element)dataEl.getFirstChild());
	        byteData=xby.getBytes();
	    }
	    else
	    {
	        ByteArrayInputStream bin=new ByteArrayInputStream(XMLUtils.toByteArray(dataEl.getFirstChild().getNodeValue()));
	        try
	        {
    	        ObjectInputStream in=new ObjectInputStream(bin);
	            serialData=(Serializable)in.readObject();
	        }
	        catch(IOException e)
	        {
System.err.println(e.toString());
	        }
	        catch(ClassNotFoundException e)
	        {
System.err.println(e.toString());
	        }
	    }
	}

}