/*
 * @(#)Invoice.java
 *
 * Classes for transferring value (a.k.a. money): Paying for things
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.payment;

import java.io.*;
import FIM.Util.XML.XMLEncodeable;
import java.util.Date;

import java.security.cert.Certificate;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidKeyException;
import java.security.cert.CertificateException;
import java.security.NoSuchProviderException;
import java.security.SignatureException;

/**
This class models an invoice to another entity. It might be signed. Can be encoded in XML.
Contains the date, a title, and the items.
@author Michael Sonntag
@version 1.0
@see FIM.payment.InvoiceItem
*/
public interface Invoice extends XMLEncodeable,Serializable
{
    /**
    Retrieve the date of the invoice.
    @return the date of the invoice
    */
    public Date getDate();

    /**
    Retrieve the title of the invoice.
    @return the title of the invoice
    */
    public String getTitle();

    /**
    Retrieve the items contained within the invoice (items or lines).
    @return the items of the invoice
    */
    public InvoiceItem[] getItems();

    /**
    Retrieve the total value of all items in the invoice.
    @return the total price
    */
    public Price getTotalPrice();

    /**
    Retrieve the issuer of the invoice as a string.
    @return the issuer of the invoice
    */
    public String getIssuer();

    /**
    Retrieve the issuer of the invoice as a certificate (might be <code>null</code>).
    @return the issuer of the invoice or <code>null</code>
    */
    public Certificate getIssuerCertificate();

    /**
    Check whether this invoice is signed or not.
    @return <code>true</code> if the invoice is signed
    */
    public boolean isSigned();

    /**
    Sign the invoice.
    @param signAlgorithm the algorithm to be used for signing
    @param pk the private key to be used for signing
    @param cert the certificate of the public key associated with the private key (can be <code>null</code>)
    @throws InvalidKeyException if the private key is invalid
    @throws NoSuchAlgorithmException if the algorithm is not supported
    */
    public void sign(String signAlgorithm,PrivateKey pk,Certificate cert) throws InvalidKeyException,NoSuchAlgorithmException;

    /**
    Retrieve the certificate passed in upon signing. Is <code>null</code> if non provided then.
    @return the certificate of the signer or <code>null</code>
    */
    public Certificate getCertificate();

    /**
    Verify the signature of the invoice based on the certificate contained. Will fail if not signed or if no certificate is present.
    @param provider the security provider to be used for checking. if <code>null</code>, the standard provider is used
    @throws InvalidKeyException if the public key is invalid
    @throws NoSuchAlgorithmException if the algorithm used for signing is not supported
    @throws InvalidKeyException if the key in the certificate is erroneous
    @throws NoSuchProviderException if the security provider could not be found
    @throws SignatureException if the signature is invalid
    */
    public void verifySignature(String provider) throws NoSuchAlgorithmException,InvalidKeyException,NoSuchProviderException,SignatureException;

    /**
    Verify the signature of the invoice based on the public key passed in. Will fail if not signed. Teh certificate included is ignored.
    @param provider the security provider to be used for checking. if <code>null</code>, the standard provider is used
    @param pk the public key to use for verification. If <code>null</code>, the key from the certificate is used (if no certificate in this case, verification fails)
    @throws NoSuchAlgorithmException if the algorithm used for signing is not supported
    @throws InvalidKeyException if the public key (or the key in the certificate) is invalid
    @throws NoSuchProviderException if the security provider could not be found
    @throws SignatureException if the signature is invalid
    */
    public void verifySignature(String provider,PublicKey pk) throws NoSuchAlgorithmException,InvalidKeyException,NoSuchProviderException,SignatureException;

    /**
    Retrieve the invoice in encoded form to be used for signing. Does NOT include the signing information itself (which
    will be created based on the result!). Must be exactly the same on all calls and if the content is the same.
    @return the invoice in encoded form
    */
    public byte[] getEncoded();

    /**
    Retrieve the invoice as a string, suitable for presenting to the user in a very simple form.
    @return the invoice encoded as a string
    */
    public String toString();

}