/*
 * @(#)InvoiceItem.java
 *
 * Classes for transferring value (a.k.a. money): Paying for things
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.payment;

import java.io.*;
import FIM.Util.XML.XMLEncodeable;

import org.xml.sax.SAXException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
Helper class: Models a line in an invoice. Just contains a text and the price for this item.
@author Michael Sonntag
@version 1.0
*/
public class InvoiceItem implements XMLEncodeable,Serializable
{
    /** The name for this line */
    private String item=null;
    /** The price associated with the line */
    private Price price=null;
    
    /**
    Creates a new line for an invoice.
    @param item the name for the item
    @param price the price for this item
    */
    public InvoiceItem(String item,Price price)
    {
        if(item==null || item.length()<1)
            throw new IllegalArgumentException("Item name must be provided");
        this.item=item;
        if(price==null)
            throw new IllegalArgumentException("Price must be provided");
        this.price=price;
    }
    
    /**
    Creates a new invoice item object from values stored in an XML Element.
    @param elem the XML element to parse the values from
    @throws IllegalArgumentException if the price could not be parsed back completely
    */
    public InvoiceItem(Element elem)
    {
        decodeXMLElement(elem);
        if(item==null || item.length()<1)
            throw new IllegalArgumentException("Item name must be provided");
        if(price==null)
            throw new IllegalArgumentException("Price must be provided");
    }

    /**
    Retrieve the name of this item.
    @return the name
    */
    public String getItemName()
    {
        return item;
    }
    
    /**
    Retrieve the price of this item.
    @return the price
    */
    public Price getPrice()
    {
        return price;
    }
    
    /** 
    Retrieve the item as a serializable (just returns <code>this</code> as this object is directly serializable).
    @return the price object as a serializable
    */
    public Serializable getAsSerializable()
	{
	    return this;
	}
	
    /** 
    Retrieve the item as an XML element.
    @param doc the document within which to create the element
    @return the price as an element
    */
	public Element getAsXMLElement(Document doc)
	{
	    Element el=doc.createElement("InvoiceItem");
	    Element nameEl=doc.createElement("Itemname");
	    nameEl.appendChild(doc.createTextNode(""+item));
	    el.appendChild(nameEl);
	    Element priceEl=price.getAsXMLElement(doc);
	    el.appendChild(priceEl);
	    return el;
	}
	
    /** 
    Parse back the item from an XML element.
    @param elem the element containing the price to parse back
    @throws IllegalArgumentException if the element is incorrect (e. g. wrong tag name)
    */
	public void decodeXMLElement(Element elem) throws IllegalArgumentException
	{
	    Element nameEl=(Element)elem.getFirstChild();
	    item=nameEl.getFirstChild().getNodeValue();
	    Element priceEl=(Element)elem.getLastChild();
	    price=new Price(priceEl);
	}

    /**
    Returns a strign representation for an invoice item. Contains the name and the price.
    @return string representation
    */
    public String toString()
    {
        return item+": "+price.toString();
    }
    
}