/*
 * @(#)NamedInvoice.java
 *
 * Classes for transferring value (a.k.a. money): Paying for things
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.payment;

import java.io.*;
import FIM.Util.XML.XMLEncodeable;
import FIM.Util.XML.XMLUtils;
import java.util.Date;

import java.security.cert.Certificate;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.Signature;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidKeyException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateFactory;
import java.security.NoSuchProviderException;
import java.security.SignatureException;
import java.text.DateFormat;

import org.xml.sax.SAXException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import iaik.asn1.*;

/*
//Needed for testing; see main at end
import org.apache.xalan.serialize.Serializer;
import org.apache.xalan.serialize.SerializerFactory;
import org.apache.xalan.templates.OutputProperties;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import PkgAgentSystem.Crypto.*;
import FIM.Util.Crypto.NamedKeyAndCertificate;
*/

/**
An invoice for a certain entity (addressee, not issuer!), which is described as a string or a certificate (or both).
@author Michael Sonntag
@version 1.0
*/
public class NamedInvoice extends AnonymousInvoice
{
    /** The addressee of the invoice. */
    private String addressee=null;
    /** The certificate of the addressee. */
    private Certificate addresseeCert=null;

    /**
    Create a new invoice. The currencies of the items must all be the same.
    Not both addressee and addresseeCert may be null.
    @param issuer the issuer of the invoice
    @param issuerCert the certificate of the issuer of the invoice
    @param data the date for the invoice
    @param title the title of the invoice
    @param item an array of items for the invoice. At least one required
    @param addressee the addressee of the invoice
    @param addresseeCert the certificate of the addressee of the invoice
    */
    public NamedInvoice(String issuer,Certificate issuerCert,Date date,String title,InvoiceItem[] items,String addressee,Certificate addresseeCert)
    {
        super(issuer,issuerCert,date,title,items);
        if(addressee==null && addresseeCert==null)
            throw new IllegalArgumentException("Addressee must be provided (either name or certificate)");
        this.addressee=addressee;
        this.addresseeCert=addresseeCert;
    }

    /**
    Create a new invoice from an XML element.
    @param elem the XML element containing the encoded form
    */
    public NamedInvoice(Element elem)
    {
        decodeXMLElement(elem);
    }

    /**
    Creates a new empty invoice. Necessary so subclasses can create an empty object and then parse in the values from XML. Subclasses doing this
    must call <code>decodeXMLElement</code>.
    */
    protected NamedInvoice()
    {
    }

    /**
    Retrieve the addressee of the invoice as a string.
    @return the addressee of the invoice
    */
    public String getAddressee()
    {
        return addressee;
    }

    /**
    Retrieve the addressee of the invoice as a certificate (might be <code>null</code>).
    @return the addressee of the invoice or <code>null</code>
    */
    public Certificate getAddresseeCertificate()
    {
        return addresseeCert;
    }

    /**
    Retrieve the invoice in encoded form to be used for signing. Does NOT include the signing information itself (which
    will be created based on the result!). Must be exactly the same on all calls and if the content is the same.
    @return the invoice in encoded form
    */
    public byte[] getEncoded()
    {
        SEQUENCE encoded=new SEQUENCE();
        byte[] res=super.getEncoded();
        encoded.addComponent(new OCTET_STRING(res));
        encoded.addComponent(new PrintableString(addressee));
        if(addresseeCert!=null)
        {
            try
            {
                encoded.addComponent(new OCTET_STRING(addresseeCert.getEncoded()));
            }
            catch(CertificateEncodingException e)
            {   // Should never happen
                System.err.println(e.toString());
            }
        }
	    return DerCoder.encode(encoded);
    }

    /**
    Retrieve the invoice as a string, suitable for presenting to the user in a very simple form.
    @return the invoice encoded as a string
    */
    public String toString()
    {
        String str="INVOICE from "+getDate().toString()+"Issued by: "+getIssuer()+"\n\n"+getTitle()+"\n\n";
        if(addressee!=null || addresseeCert!=null)
            str+="\n";
        if(addressee!=null)
            str+="Addressee: "+addressee+"\n";
        if(addresseeCert!=null)
            str+="Addressee certificate present\n";
        str+="\n";
        InvoiceItem[] items=getItems();
        for(int i=0;i<items.length;i++)
            str+=items[i].toString()+"\n";
        str+="--------------------\nTotal: "+getTotalPrice().toString();
        if(isSigned())
        {
            str+="\n\nInvoice is signed";
            if(getCertificate()!=null)
                str+="\nCertificate present: "+getCertificate().toString();
        }
        return str;
    }

    /**
    Retrieve the invoice as an XML element.
    @param doc the document within which to create the element
    @return the price as an element
    */
	public Element getAsXMLElement(Document doc)
	{
	    Element el=doc.createElement("NamedInvoice");
	    if(addressee!=null)
	        el.setAttribute("AddresseeName",addressee);
	    el.appendChild(super.getAsXMLElement(doc));
        if(addresseeCert!=null)
        {
            Element certEl=doc.createElement("AddresseeCertificate");
        	certEl.setAttribute("Type",addresseeCert.getType());
	        try
	        {
	            certEl.appendChild(doc.createTextNode(XMLUtils.toHexString(addresseeCert.getEncoded())));
	        }
	        catch(CertificateEncodingException e)
	        {   // Just leave it out then; shouldn't happen anyway
	            System.err.print("");
	        }
    	    el.appendChild(certEl);
	    }
	    return el;
	}

    /**
    Parse back the invoice from an XML element.
    @param elem the element containing the price to parse back
    @throws IllegalArgumentException if the element is incorrect (e. g. wrong tag name)
    */
	public void decodeXMLElement(Element elem) throws IllegalArgumentException
	{
	    super.decodeXMLElement((Element)elem.getFirstChild());
	    addressee=elem.getAttribute("AddresseeName");
	    if(elem.getChildNodes().getLength()>1)
	    {
	        // Addressee certificate present
            byte[] buf=XMLUtils.toByteArray(elem.getLastChild().getFirstChild().getNodeValue());
	        try
	        {
                CertificateFactory cf=CertificateFactory.getInstance(((Element)elem.getLastChild()).getAttribute("Type"));
    	        ByteArrayInputStream bais=new ByteArrayInputStream(buf);
	            addresseeCert=cf.generateCertificate(bais);
	        }
	        catch(CertificateException e)
	        {
	            throw new RuntimeException("Could not recreate addressee certificate");
	        }
	    }
	}

    private void writeObject(java.io.ObjectOutputStream out) throws IOException
    {
        out.writeUTF(addressee);
        out.writeUTF(addresseeCert.getType());
        try
        {
            out.writeObject(addresseeCert.getEncoded());
        }
        catch(CertificateEncodingException e)
        {
	        throw new NotSerializableException(e.toString());
        }
    }

    private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException
    {
        addressee=in.readUTF();
        String type=in.readUTF();
        byte[] bytes=(byte[])in.readObject();
        try
        {
            CertificateFactory cf=CertificateFactory.getInstance(type);
    	    ByteArrayInputStream bais=new ByteArrayInputStream(bytes);
	        addresseeCert=cf.generateCertificate(bais);
	    }
	    catch(CertificateException e)
	    {
	        throw new IOException(e.toString());
	    }
    }

/*
	public static void main(String[] args) throws Exception
	{
	    InvoiceItem[] items=new InvoiceItem[2];
	    items[0]=new InvoiceItem("Players Handbook",new Price(Currency.findCurrency("USD"),19.9));
	    items[1]=new InvoiceItem("Monstrous Manual",new Price(Currency.findCurrency("USD"),24.9));
	    InvoiceItem[] taxes=new InvoiceItem[1];
	    taxes[0]=new InvoiceItem("VAT",new Price(Currency.findCurrency("USD"),8.96));
	    // Next 2 lines needed for loading the security store with agent certificates in it
        java.net.URL.setURLStreamHandlerFactory(new PkgAgentSystem.AMPStreamHandlerFactory());
	    java.security.Security.addProvider(new PkgAgentSystem.Crypto.AgentCryptographyProvider());
        PersonalSecurityStore store=PersonalSecurityStoreFactory.openPersonalSecurityStore("d:\\data\\java\\agents\\keystore.store","password".toCharArray());
	    NamedKeyAndCertificate nkc=store.getOwnerPrivateKeyAndCertificate("Server Certificate");
	    NamedInvoice inv=new NamedInvoice(null,"Test-Invoice",items,taxes,"Michael Sonntag",nkc.getCertificate());
        inv.sign("MD5/RSA",nkc.getPrivateKey(),nkc.getCertificate());
        inv.verifySignature(null);
	    System.out.println("Original:\n"+inv.toString()+"\n\n");
        String str="<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<Invoice/>";
        DocumentBuilderFactory dFactory=DocumentBuilderFactory.newInstance();
        DocumentBuilder dBuilder=dFactory.newDocumentBuilder();
        Document document=dBuilder.parse(new java.io.ByteArrayInputStream(str.getBytes()));
        document.getDocumentElement().appendChild(inv.getAsXMLElement(document));
        ByteArrayOutputStream bos=new ByteArrayOutputStream();
        OutputProperties prop=new OutputProperties(OutputProperties.getDefaultMethodProperties("xml"));
        Serializer serializer=SerializerFactory.getSerializer(prop.getProperties());
        serializer.setOutputStream(bos);
        serializer.asDOMSerializer().serialize(document);
        System.out.println(bos.toString());
        inv=new NamedInvoice((Element)document.getDocumentElement().getFirstChild());
        System.out.println("\n\nParsed back:\n"+inv);
	}
*/
}