/*
 * @(#)Payment.java
 *
 * Classes for transferring value (a.k.a. money): Paying for things
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.payment;

import java.io.*;
import FIM.Util.XML.XMLEncodeable;

import java.security.cert.Certificate;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidKeyException;
import java.security.cert.CertificateException;
import java.security.NoSuchProviderException;
import java.security.SignatureException;

/**
A class representing the payment for a certain invoice. The actual payment information is implementation specific.
The payment can be signed.
@author Michael Sonntag
@version 1.0
@see FIM.payment.Invoice
*/
public interface Payment extends XMLEncodeable,Serializable
{
    /**
    Retrieve the invoice this payment is for.
    @return the invoice
    */
    public Invoice getInvoice();

    /**
    Check whether this payment is signed or not.
    @return <code>true</code> if the payment is signed
    */
    public boolean isSigned();

    /**
    Sign the payment.
    @param signAlgorithm the algorithm to be used for signing
    @param pk the private key to be used for signing
    @param cert the certificate of the public key associated with the private key (can be <code>null</code>)
    @throws InvalidKeyException if the private key is invalid
    @throws NoSuchAlgorithmException if the algorithm is not supported
    */
    public void sign(String signAlgorithm,PrivateKey pk,Certificate cert) throws InvalidKeyException,NoSuchAlgorithmException;

    /**
    Retrieve the certificate passed in upon signing. Is <code>null</code> if non provided then.
    @return the certificate of the signer or <code>null</code>
    */
    public Certificate getCertificate();

        /**
    Verify the signature of the invoice based on the certificate contained. Will fail if not signed or if no certificate is present.
    @param provider the security provider to be used for checking. if <code>null</code>, the standard provider is used
    @throws InvalidKeyException if the public key is invalid
    @throws NoSuchAlgorithmException if the algorithm used for signing is not supported
    @throws InvalidKeyException if the key in the certificate is erroneous
    @throws NoSuchProviderException if the security provider could not be found
    @throws SignatureException if the signature is invalid
    */
    public void verifySignature(String provider) throws NoSuchAlgorithmException,InvalidKeyException,NoSuchProviderException,SignatureException;

    /**
    Verify the signature of the invoice based on the public key passed in. Will fail if not signed. Teh certificate included is ignored.
    @param provider the security provider to be used for checking. if <code>null</code>, the standard provider is used
    @param pk the public key to use for verification. If <code>null</code>, the key from the certificate is used (if no certificate in this case, verification fails)
    @throws NoSuchAlgorithmException if the algorithm used for signing is not supported
    @throws InvalidKeyException if the public key (or the key in the certificate) is invalid
    @throws NoSuchProviderException if the security provider could not be found
    @throws SignatureException if the signature is invalid
    */
    public void verifySignature(String provider,PublicKey pk) throws NoSuchAlgorithmException,InvalidKeyException,NoSuchProviderException,SignatureException;

    /**
    Retrieve a characteristic string for this type of payment (e. g. "Creditcard").
    @return the characteristic of this payment
    */
    public String characteristics();

    /**
    Retrieve the payment in encoded form to be used for signing. Does NOT include the signing information itself (which
    will be created based on the result!). Must be exactly the same on all calls and if the content is the same.
    @return the payment in encoded form
    */
    public byte[] getEncoded();

    /**
    Retrieve the payment as a string, suitable for presenting to the user in a very simple form.
    @return the payment encoded as a string
    */
    public String toString();

}