/*
 * @(#)PaymentBase.java
 *
 * Classes for transferring value (a.k.a. money): Paying for things
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.payment;

import java.util.Date;
import java.util.Calendar;
import java.io.*;
import FIM.Util.XML.*;

import java.security.cert.Certificate;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.Signature;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidKeyException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateFactory;
import java.security.NoSuchProviderException;
import java.security.SignatureException;
import java.text.DateFormat;

import org.xml.sax.SAXException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
A helper class with common methods of interface <code>Payment</code> implemented.
@author Michael Sonntag
@version 1.0
@see FIM.payment.Invoice
*/
public abstract class PaymentBase implements Payment
{
    /** Invoice this payment is for */
    private Invoice invoice=null;
    /** Characteristic for this payment */
    private String characteristic=null;

    /** The (optional) certificate associated with the signature. */
    private Certificate certificate=null;
    /** The algorithm used for signing. */
    private String signAlgorithm=null;
    /** The actual siganture. */
    private byte[] signature=null;

    /**
    Create a new payment object with a characteristic.
    @param invoice the invoice this payment is for
    @param characteristic the description of the content
    @param stringData the string data
    */
    public PaymentBase(Invoice inv,String characteristic)
    {
        if(inv==null)
            throw new IllegalArgumentException("Invoice must be provided");
        invoice=inv;
        if(characteristic==null || characteristic.length()<1)
            throw new IllegalArgumentException("Characteristic must be provided");
        this.characteristic=characteristic;
    }

    /**
    Create a new empty object. To be used for parsing in XML data later.
    */
    protected PaymentBase()
    {
    }

    /**
    Retrieve the invoice this payment is for.
    @return the invoice
    */
    public Invoice getInvoice()
    {
        return invoice;
    }

    /**
    Retrieve the certificate passed in upon signing. Is <code>null</code> if non provided then.
    @return the certificate of the signer or <code>null</code>
    */
    public Certificate getCertificate()
    {
        return certificate;
    }

    /**
    Check whether this payment is signed or not.
    @return <code>true</code> if the payment is signed
    */
    public boolean isSigned()
    {
        return signature!=null;
    }

    /**
    Sign the payment.
    @param signAlgorithm the algorithm to be used for signing
    @param pk the private key to be used for signing
    @param cert the certificate of the public key associated with the private key (can be <code>null</code>)
    @throws InvalidKeyException if the private key is invalid
    @throws NoSuchAlgorithmException if the algorithm is not supported
    */
    public void sign(String signAlgorithm,PrivateKey pk,Certificate cert) throws InvalidKeyException,NoSuchAlgorithmException
    {
        certificate=cert;
   		Signature sig=Signature.getInstance(signAlgorithm);
    	sig.initSign(pk);
   		try
   		{
    		sig.update(getEncoded());
	    	signature=sig.sign();
            this.signAlgorithm=signAlgorithm;
        }
		catch(SignatureException e)
        {   // Shouldn't happen: we initialized all
            System.err.println(e.toString());
		}
    }

    /**
    Verify the signature of the payment based on the certificate contained. Will fail if not signed or if no certificate is present.
    @param provider the security provider to be used for checking. if <code>null</code>, the standard provider is used
    @throws InvalidKeyException if the public key is invalid
    @throws NoSuchAlgorithmException if the algorithm used for signing is not supported
    @throws InvalidKeyException if the key in the certificate is erroneous
    @throws NoSuchProviderException if the security provider could not be found
    @throws SignatureException if the signature is invalid
    */
    public void verifySignature(String provider) throws NoSuchAlgorithmException,InvalidKeyException,NoSuchProviderException,SignatureException
    {
        verifySignature(provider,null);
    }

    /**
    Verify the signature of the payment based on the public key passed in. Will fail if not signed. Teh certificate included is ignored.
    @param provider the security provider to be used for checking. if <code>null</code>, the standard provider is used
    @param pk the public key to use for verification. If <code>null</code>, the key from the certificate is used (if no certificate in this case, verification fails)
    @throws NoSuchAlgorithmException if the algorithm used for signing is not supported
    @throws InvalidKeyException if the public key (or the key in the certificate) is invalid
    @throws NoSuchProviderException if the security provider could not be found
    @throws SignatureException if the signature is invalid
    */
    public void verifySignature(String provider,PublicKey pk) throws NoSuchAlgorithmException,InvalidKeyException,NoSuchProviderException,SignatureException
    {
        if(pk==null && certificate==null)
            throw new SignatureException("Cannot verify signature: No certificate included and no public key specified");
        if(pk==null)
            pk=certificate.getPublicKey();
	    if(signature==null)
    		throw new SignatureException("Not signed");
   		Signature sig=null;
   		if(provider==null)
        	sig=Signature.getInstance(signAlgorithm);
   		else
        	sig=Signature.getInstance(signAlgorithm,provider);
    	sig.initVerify(pk);
   		sig.update(getEncoded());
    	if(!sig.verify(signature))
    		throw new SignatureException("Signature invalid");
    }

    /**
    Retrieve the payment in encoded form to be used for signing. Does NOT include the signing information itself (which
    will be created based on the result!). Must be exactly the same on all calls and if the content is the same.
    @return the payment in encoded form
    */
    public byte[] getEncoded()
    {
        ByteArrayOutputStream bos=new ByteArrayOutputStream();
        try
        {
            ObjectOutputStream out=new ObjectOutputStream(bos);
            out.writeUTF(characteristic);
            out.writeObject(invoice);
            out.close();
        }
        catch(IOException e)
        {   // Shouldn't happen
            System.err.println(e.toString());
        }
        return bos.toByteArray();
    }

    /**
    Retrieve the characteristic string for this type of payment (description of the content)
    @return the characteristic of this payment
    */
    public String characteristics()
    {
        return characteristic;
    }


    /**
    Retrieve the payment as a serializable (just returns <code>this</code> as this object is directly serializable).
    @return the currency object as a serializable
    */
    public Serializable getAsSerializable()
	{
	    return this;
	}

    /**
    Retrieve the payment as an XML element.
    @param doc the document within which to create the element
    @return the currency as an element
    */
	public Element getAsXMLElement(Document doc)
	{
	    Element el=doc.createElement("PaymentBase");
	    el.setAttribute("Characteristic",characteristic);
	    el.appendChild(invoice.getAsXMLElement(doc));
	    if(signature!=null)
	    {
    	    Element sigEl=doc.createElement("Signature");
	        sigEl.setAttribute("SignatureAlgorithm",signAlgorithm);
	        sigEl.setAttribute("Signature",XMLUtils.toHexString(signature));
	        if(certificate!=null)
	        {
        	    Element certEl=doc.createElement("Certificate");
        	    certEl.setAttribute("Type",certificate.getType());
	            try
	            {
	                certEl.appendChild(doc.createTextNode(XMLUtils.toHexString(certificate.getEncoded())));
	            }
	            catch(CertificateEncodingException e)
	            {   // Just leave it out then; shouldn't happen anyway
	                System.err.print("");
	            }
    	        sigEl.appendChild(certEl);
	        }
	        el.appendChild(sigEl);
	    }
	    return el;
	}

    /**
    Parse back the payment from an XML element.
    @param elem the element containing the currency to parse back
    @throws IllegalArgumentException if the element is incorrect (e. g. wrong tag name)
    */
	public void decodeXMLElement(Element elem) throws IllegalArgumentException
	{
	    characteristic=elem.getAttribute("Characteristic");
	    Element invEl=(Element)elem.getFirstChild();
	    if(invEl.getTagName().equals("AnonymousInvoice"))
	        invoice=new AnonymousInvoice(invEl);
	    else if(invEl.getTagName().equals("NamedInvoice"))
	        invoice=new NamedInvoice(invEl);
	    else
	        throw new IllegalArgumentException("Unknown type of invoice: "+invEl.getTagName());
	    Element sigEl=(Element)invEl.getNextSibling();
	    if(sigEl!=null)
	    {
	        signAlgorithm=sigEl.getAttribute("SignatureAlgorithm");
	        signature=XMLUtils.toByteArray(sigEl.getAttribute("Signature"));
	        if(sigEl.getFirstChild()!=null)
	        {
	            byte[] buf=XMLUtils.toByteArray(sigEl.getFirstChild().getFirstChild().getNodeValue());
	        	try
	        	{
            		CertificateFactory cf=CertificateFactory.getInstance(((Element)sigEl.getFirstChild()).getAttribute("Type"));
    	        	ByteArrayInputStream bais=new ByteArrayInputStream(buf);
	                certificate=cf.generateCertificate(bais);
	            }
	            catch(CertificateException e)
	            {
    	            throw new RuntimeException("Could not recreate certificate");
	            }
	        }
	    }
	}

}