/*
 * @(#)Price.java
 *
 * Classes for transferring value (a.k.a. money): Paying for things
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.payment;

import java.io.*;
import java.text.NumberFormat;
import java.text.ParseException;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Locale;

import FIM.Util.XML.XMLEncodeable;

import org.xml.sax.SAXException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
A class for representing a price by a currency object and a value. Also supports parsing
both from a string.
@author Michael Sonntag
@version 1.0
*/
public class Price implements XMLEncodeable,Serializable, Comparable
{
    /**
    The currency of this price.
    */
    private Currency curr=null;

    /**
    The value of this price.
    */
    private double price=-1.0d;

    /**
    Creates a new price object.
    @param curr the currency
    @param price the value
    @throws IllegalArgumentException if the currency is <code>null</code> or the price is &lt; 0.0
    */
    public Price(Currency curr,double price)
    {
        if(curr==null)
            throw new IllegalArgumentException("Currency must be provided");
        this.curr=curr;
        if(price<0.0d)
            throw new IllegalArgumentException("Value must be at least 0.0");
        this.price=price;
    }

    /**
    Creates a new price object from values stored in an XML Element.
    @param elem the XML element to parse the values from
    @throws IllegalArgumentException if the price could not be parsed back completely
    */
    public Price(Element elem)
    {
        decodeXMLElement(elem);
        if(curr==null)
            throw new IllegalArgumentException("Currency must be provided");
        if(price<0.0d)
            throw new IllegalArgumentException("Value must be at least 0.0");
    }

    /**
    Retrieve the currency of this price.
    @return the currency
    */
    public Currency getCurrency()
    {
        return curr;
    }

    /**
    Retrieve the value of this price.
    @return the value
    */
    public double getPrice()
    {
        return price;
    }

    /**
    Retrieve the value of this price in Euro.
    Uses the exchange factor stored in the currency.
    @return the value in Euro
    @param IllegalStateException if no exchange value is present
    */
    public double getStandardPrice()
    {
        if(curr.getExchangeRateToStandard()>0.0d)
            return price/curr.getExchangeRateToStandard();
        throw new IllegalStateException("No exchange rate present in currency "+curr.getShortName());
    }

    /**
    Retrieve a string representation of this price. Includes the price and the value.
    Uses the locale for formatting.
    @return the price as a string
    */
    public String toString()
    {
        NumberFormat nf=NumberFormat.getNumberInstance(curr.getLocale());
        nf.setMinimumFractionDigits(2);
        nf.setMaximumFractionDigits(2);
        String priceStr;
        priceStr=nf.format(price);
        if(price==Double.POSITIVE_INFINITY)
            priceStr="+INF";
        if(price==Double.NEGATIVE_INFINITY)
            priceStr="-INF";
        if(price==Double.NaN)
            priceStr="NaN";
        if(curr.isPreValue())
            return curr.getShortName()+" "+priceStr;
        else
            return priceStr+" "+curr.getShortName();
    }

    /**
    Compares two prices for equality.
    @param obj the object to compare to this one
    @return <code>true</code> if the currency and the value is the same
    */
    public boolean equals(Object obj)
    {
        if(!(obj instanceof Price))
            return false;
        Price other=(Price)obj;
        return other.curr.equals(curr) && other.price==price;
    }

    /**
    Compares this price to another one. If the currency is the same, the value is compared.
    Otherwise the value in Euro is compared. If no conversion is abailable for one of the
    currencies, a <code>ClassCastException</code> is thrown.
    @param o the object to compare to this one
    @return returns a negative integer, zero, or a positive integer as this object is less than, equal to, or greater than the specified object
    @throws ClassCastException if the other object is not a price
    @throws ClassCastException if the other object has a different currency and one of the
        currencies cannot be converted to Euros
    */
    public int compareTo(Object o)
    {
        if(!(o instanceof Price))
            throw new ClassCastException("Cannot compare Price with "+o.getClass().getName());
        Price comp=(Price)o;
        if(curr.equals(comp.getCurrency()))
        {
            if(price<comp.getPrice())
                return -1;
            if(price>comp.getPrice())
                return +1;
            return 0;
        }
        try
        {
            getStandardPrice();
        }
        catch(IllegalStateException e)
        {
            throw new ClassCastException("Conversion to Euro not possible for this object");
        }
        try
        {
            comp.getStandardPrice();
        }
        catch(IllegalStateException e)
        {
            throw new ClassCastException("Conversion to Euro not possible for other object");
        }
        if(getStandardPrice()<comp.getStandardPrice())
            return -1;
        if(getStandardPrice()>comp.getStandardPrice())
            return +1;
        return 0;
    }

    /**
    Retrieve the price as a serializable (just returns <code>this</code> as this object is directly serializable).
    @return the price object as a serializable
    */
    public Serializable getAsSerializable()
	{
	    return this;
	}

    /**
    Retrieve the price as an XML element.
    @param doc the document within which to create the element
    @return the price as an element
    */
	public Element getAsXMLElement(Document doc)
	{
	    Element el=doc.createElement("Price");
	    Element currencyEl=curr.getAsXMLElement(doc);
	    Element priceEl=doc.createElement("Value");
	    priceEl.appendChild(doc.createTextNode(""+price));
	    el.appendChild(currencyEl);
	    el.appendChild(priceEl);
	    return el;
	}

    /**
    Parse back the price from an XML element.
    @param elem the element containing the price to parse back
    @throws IllegalArgumentException if the element is incorrect (e. g. wrong tag name)
    */
	public void decodeXMLElement(Element elem) throws IllegalArgumentException
	{
	    Element currencyEl=(Element)elem.getFirstChild();
	    curr=new Currency(currencyEl);
	    Element priceEl=(Element)elem.getLastChild();
	    try
	    {
	        price=Double.parseDouble(priceEl.getFirstChild().getNodeValue());
	    }
	    catch(NumberFormatException e)
	    {
	        throw new IllegalArgumentException("Could not parse price value ("+e.getMessage()+")");
	    }
	}

}