/*
 * @(#)VoucherPayment.java
 *
 * Classes for transferring value (a.k.a. money): Paying for things
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.payment;

import java.util.Date;
import java.util.Calendar;
import java.io.*;
import FIM.Util.XML.*;

import java.security.cert.Certificate;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.Signature;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidKeyException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateFactory;
import java.security.NoSuchProviderException;
import java.security.SignatureException;
import java.text.DateFormat;

import org.xml.sax.SAXException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/*
// Needed for testing; see main at end
import org.apache.xalan.serialize.Serializer;
import org.apache.xalan.serialize.SerializerFactory;
import org.apache.xalan.templates.OutputProperties;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import PkgAgentSystem.Crypto.*;
import FIM.Util.Crypto.NamedKeyAndCertificate;
*/

/**
A class representing the payment by providing a voucher (which must be a certificate).
@author Michael Sonntag
@version 1.0
@see FIM.payment.Invoice
*/
public class VoucherPayment extends PaymentBase
{
    /** The voucher itself */
    private Certificate voucher=null;

    /**
    Create a new empty object. To be used for parsing in XML data later. Must be public so it can be
    created by factory methods.
    */
    public VoucherPayment()
    {
    }

    /**
    Create a new voucher payment.
    @param invoice the invoice this payment is for
    @param characteristic the description of the content
    @param voucher the voucher
    */
    public VoucherPayment(Invoice inv,String characteristic,Certificate voucher)
    {
        super(inv,characteristic);
        if(voucher==null)
            throw new IllegalArgumentException("Voucher must be provided");
        this.voucher=voucher;
    }

    /**
    Create a new voucher payment from an XML element.
    @param elem the XML element containing the encoded form
    */
    public VoucherPayment(Element elem)
    {
        decodeXMLElement(elem);
    }

    /**
    Retrieve the voucher.
    @return the voucher
    */
    public Certificate getVoucher()
    {
        return voucher;
    }

    /**
    Retrieve the payment as a string, suitable for presenting to the user in a very simple form.
    @return the payment encoded as a string
    */
    public String toString()
    {
        String str="Payment for Invoice from "+getInvoice().getDate().toString()+" ("+getInvoice().getTitle()+")\n\n";
        str+="Payment by Voucher: "+voucher.toString()+"\n";
        return str;
    }

    /**
    Retrieve the payment in encoded form to be used for signing. Does NOT include the signing information itself (which
    will be created based on the result!). Must be exactly the same on all calls and if the content is the same.
    @return the payment in encoded form
    */
    public byte[] getEncoded()
    {
        ByteArrayOutputStream bos=new ByteArrayOutputStream();
        try
        {
            ObjectOutputStream out=new ObjectOutputStream(bos);
            out.writeObject(super.getEncoded());
            out.writeUTF(voucher.getType());
            out.writeObject(voucher.getEncoded());
            out.close();
        }
        catch(IOException e)
        {   // Shouldn't happen
            System.err.println(e.toString());
        }
        catch(CertificateEncodingException e)
        {   // Shouldn't happen
            System.err.println(e.toString());
        }
        return bos.toByteArray();
    }

    /**
    Retrieve the payment as an XML element.
    @param doc the document within which to create the element
    @return the currency as an element
    */
	public Element getAsXMLElement(Document doc)
	{
	    Element el=doc.createElement("VoucherPayment");
	    el.appendChild(super.getAsXMLElement(doc));
	    Element data=doc.createElement("Voucher");
	    data.setAttribute("Type",voucher.getType());
	    try
	    {
	        data.appendChild(doc.createTextNode(XMLUtils.toHexString(voucher.getEncoded())));
	    }
	    catch(CertificateEncodingException e)
	    {
	        System.err.println(e.toString());
	    }
        el.appendChild(data);
	    return el;
	}

    /**
    Parse back the payment from an XML element.
    @param elem the element containing the currency to parse back
    @throws IllegalArgumentException if the element is incorrect (e. g. wrong tag name)
    */
	public void decodeXMLElement(Element elem) throws IllegalArgumentException
	{
	    super.decodeXMLElement((Element)elem.getFirstChild());
	    Element dataEl=(Element)elem.getFirstChild().getNextSibling();
        byte[] vbuf=XMLUtils.toByteArray(dataEl.getFirstChild().getNodeValue());
	    try
	    {
                CertificateFactory cf=CertificateFactory.getInstance(dataEl.getAttribute("Type"));
    	        ByteArrayInputStream bais=new ByteArrayInputStream(vbuf);
	            voucher=cf.generateCertificate(bais);
	    }
	    catch(CertificateException e)
	    {
	        throw new RuntimeException("Could not recreate voucher");
	    }
	}

/*
	public static void main(String[] args) throws Exception
	{
	    InvoiceItem[] items=new InvoiceItem[3];
	    items[0]=new InvoiceItem("Players Handbook",new Price(Currency.findCurrency("USD"),19.9));
	    items[1]=new InvoiceItem("Monstrous Manual",new Price(Currency.findCurrency("USD"),24.9));
	    items[2]=new InvoiceItem("VAT",new Price(Currency.findCurrency("USD"),8.96));
	    AnonymousInvoice inv=new AnonymousInvoice("FIM-Test",null,null,"Test-Invoice",items);
	    // Next 2 lines needed for loading the security store with agent certificates in it
        java.net.URL.setURLStreamHandlerFactory(new PkgAgentSystem.AMPStreamHandlerFactory());
	    java.security.Security.addProvider(new PkgAgentSystem.Crypto.AgentCryptographyProvider());
        PersonalSecurityStore store=PersonalSecurityStoreFactory.openPersonalSecurityStore("d:\\data\\java\\agents\\keystore.store","password".toCharArray());
	    NamedKeyAndCertificate nkc=store.getOwnerPrivateKeyAndCertificate("Server Certificate");
        inv.sign("MD5/RSA",nkc.getPrivateKey(),nkc.getCertificate());
	    Payment py=new VoucherPayment(inv,"Test-Voucher",nkc.getCertificate());
        py.sign("MD5/RSA",nkc.getPrivateKey(),nkc.getCertificate());
        py.verifySignature(null);
	    System.out.println("Original:\n"+py.toString()+"\n\n");
        String str="<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<Payment/>";
        DocumentBuilderFactory dFactory=DocumentBuilderFactory.newInstance();
        DocumentBuilder dBuilder=dFactory.newDocumentBuilder();
        Document document=dBuilder.parse(new java.io.ByteArrayInputStream(str.getBytes()));
        document.getDocumentElement().appendChild(py.getAsXMLElement(document));
        ByteArrayOutputStream bos=new ByteArrayOutputStream();
        OutputProperties prop=new OutputProperties(OutputProperties.getDefaultMethodProperties("xml"));
        Serializer serializer=SerializerFactory.getSerializer(prop.getProperties());
        serializer.setOutputStream(bos);
        serializer.asDOMSerializer().serialize(document);
        System.out.println(bos.toString());
        py=new VoucherPayment((Element)document.getDocumentElement().getFirstChild());
        System.out.println("\n\nParsed back:\n"+py);
	}
*/
}