/*
 * @(#)Mutex.java
 *
 * Miscellaneous utilities for synchronization
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */
 
package FIM.synchronisation;

/**
A standard mutual exclusion as defined in books on operating systems. Used if only
one thread at any time may execaute a certain piece of code. At the beginning of the critical
section call <code>lock</code> and at the end <code>unlock</code>.

@author  Michael Sonntag
@version 1.0, 1.7.2000
@see     java.lang.Thread
*/
public final class Mutex
{
    /** The current owner of the mutex*/    
	private Thread owner=null;
	/** The number of threads waiting to acquire the mutex */
	private int wait_count=0;
	
	/**
	Try to acquire the mutex for up to a maximum of the provided time in milliseconds.
	@param millis number of milliseconds to wait. Providing 0 will wait indefinitely
	@return true if the mutex was acquired, false if the time ran out without success
	@throws InterruptedException if the wait for the mutex was {@link java.lang.Thread#interrupt() interrupt}
    @see java.lang.Thread    
	*/
	public synchronized boolean lock(int millis) throws InterruptedException
	{
	    // If already the owner return
		if(owner==Thread.currentThread())
			return true;
	    // Loop necessary for a certain combination of circumstances
	    /*
        In einem besonderen Fall kann es passieren, da zwei Threads gleichzeitig das wait verlassen:
        1. Der MutEx ist belegt (Thread-1) und wird gerade vom Besitzer freigegeben. 
           Nach Betreten der synchronized-Methode unlock luft bei einem anderen Thread (Thread-2)
           die Wartezeit ab und er wird in den Zustand "Runnable" versetzt (aber nicht gestartet,
           da der Monitor noch von unlock belegt wird!).
        2. In unlock wird mit notify ein weiterer Thread (Thread-3) aufgeweckt und in den Zustand 
           "Runnable" versetzt. Dieser Thread sollte jedoch unbestimmt warten (es wurde wait(0)
           aufgerufen). Auch dieser kann (noch) nicht weiterlaufen.
        3. Sobald unlock verlassen wird, kann Thread-2 (auf Wartezeit gestartet) das wait 
           verlassen und erhlt den MutEx zugesprochen.
        4. Sobald Thread-2 lock verlt, kann Thread-3 aus dem wait herauskommen. Dieser Thread 
           sollte jedoch unbestimmt warten und der MutEx ist zur Zeit nicht frei! Daher mu 
           dieser Thread zurck zum wait geschickt werden. Dies kann beliebig oft hintereinander 
           auftreten. 
        Daher ist eine WHILE-Schleife ntig!
	    */
		while(owner!=null)
		{
			try
			{
			    // Increase number of waiting threads
				wait_count++;
				// Wait for the time provided
				wait(millis);
			}
			finally
			{   // Always reduce the number of waiting threads, also when interrupted
				wait_count--;
			}
			// If we should wait for a certain time and the mutex is not free, the time
			// has elapsed (if interrupted, this code will not be reached)
			if(millis!=0 && owner!=null)
				return false;
		}
		// Current thread is now the owner
		owner=Thread.currentThread();
		return true;
	}
	
	/**
	Try to acquire the mutex. Will wait indefinitely till the mutex is acquired or it is
	{@link java.lang.Thread#interrupt() interrupt}.
	@return true if the mutex was acquired, false if the time ran out without success
	@throws InterruptedException if the wait for the mutex was interrupted
    @see java.lang.Thread    
	*/
	public synchronized boolean lock() throws InterruptedException
	{
		return lock(0);
	}
	
	/**
	Releases the mutex. May only be called by the thread possessing the mutex (else a 
	<code>RuntimeException</code> will be thrown). If threads are waiting, one of them
	is released and can acquire the mutex. Which of them is NOT defined!
	*/
	public synchronized void unlock()
	{
	    // Check for ownership
		if(owner!=Thread.currentThread())
			throw new RuntimeException("Cannot unlock: Thread is not Mutex owner");
	    // Release it
		owner=null;
		// If at least one thread is waiting wake up one of them
		if(wait_count>0)
			notify();
	}
}
