/*
 * @(#)Semaphor.java
 *
 * Miscellaneous utilities for synchronization
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.synchronisation;

/**
A standard semaphor as defined in books on operating systems.
Negative values are the number of threads waiting, positive values the
number of possible waits without blocking.<br>
Because of a collision with java.lang.Object, the methods are all
prefixed with "Sem": SemWait, SemSignal, SemValue.

@author  Michael Sonntag
@version 1.0, 1.7.2000
@see     java.lang.Thread
*/
public final class Semaphor
{
    /** Current count of the semaphor */
	private long count=1;
	
	/**
	Create a new semaphor with a count of 1.
	*/
	public Semaphor()
	{
		SemInit(1);
	}
	
	/**
	Create a new semaphor with the provided count.
	@param init initial value of the semaphor
	*/
	public Semaphor(long init)
	{
		SemInit(init);
	}
	
	/**
	Re-Initializes the semaphor with the value provided. Take care, strange results may happen if
	this change is not expected by all other threads (and they are progammed for this)!
	If threads are waiting and the parameter is greater than their number, the according number of 
	them is set free immediately. If the parameter is less than the number of waiting threads, the
	waiting threads are set free with the first <code>SemSignal</code>s and the excess value is
	only required for new <code>SemWait</code>s to not block.
	@param init the value to initialize the semaphor with
	*/
	public synchronized void SemInit(long init)
	{
		if(count<0 && init>count)
		{
			for(long i=count;i<init;i++)
			    notify();
		}
		count=init;
	}
	
	/**
	Returns the current value of the semaphor.
	Three ranges are possible:
	<ul>
	<li>Positive value: That may <code>SemWait</code> will succeed without blocking</li>
	<li>Zero value: No thread is waiting, the next call to <code>SemWait</code> will block till <code>SemSignal</code> is called</li>
	<li>Negative value: That many threads are currently waiting</li>
	</ul>
	@return current value
	*/
	public synchronized long SemValue()
	{
		return count;
	}
	
    /**
    Waits till the Semaphor is available.
    @throws InterruptedException if {@link java.lang.Thread#interrupt() interrupt} is called
    @see java.lang.Thread    
    */
	public synchronized void SemWait() throws InterruptedException
	{
		count--;
		if(count<0)
			wait(0);
	}

    /**
    Waits till the Semaphor is available.<br>
    ATTENTION: Will ignore any interrupts (no {@link java.lang.Thread#interrupt() interrupt} possible; swallows the <code>InterruptedException</code>)
    and wait again till finally a SemSignal is done.
    No termination in between possible!
    @see java.lang.Thread
    */
	public void SemWaitNoInterrupt()
	{
	    boolean end=false;
	    while(!end)
	    {
	        try
	        {
    	        SemWait();
	            end=true;
	        }
	        catch(InterruptedException e)
	        {   // Do nothing here, just loop back and wait again
	            System.err.print("");
	        }
	    }
	}
	
	/**
	Send a signal to the Semaphor. If threads are waiting, one of them is started.
	Which one is NOT defined.
	*/
	public synchronized void SemSignal()
	{
		count++;
		if(count<=0)
			notify();
	}

}
