/*
 * @(#)FIMFactory.java
 *
 * Part of the AgentSystem "POND"
 * Developed at the Institute for Information Processing and Microprocessor Technology (FIM)
 * Johannes Kepler University Linz
 * Altenbergerstr. 69, A-4040 Linz, Austria
 */

package FIM.test;

import FIM.Util.ClassLoaderObjectInputStream;
import FIM.payment.*;
import java.security.NoSuchAlgorithmException;
import java.security.KeyFactory;
import java.security.spec.X509EncodedKeySpec;
import java.security.spec.InvalidKeySpecException;
import java.security.cert.*;
import java.io.*;
import java.util.Collection;
import java.util.Locale;
import java.math.BigInteger;
import java.text.DateFormat;
import java.text.ParseException;

import iaik.asn1.*;

/**
Factory for creating <code>FIMVoucher</code> certificates form input streams. Only restoring a single certificate is supported.
Multiple certificates and certificate revocation lists are not implemented and will throw an exception if called.<br>
The certificate will be loaded through the same classloader as the <code>InputStream</code> to ensure correct access and security.

@author Michael Sonntag
@version 1.0, 1.7.2000
@see FIM.test.FIMVoucher
@see FIM.test.FIMCryptographyProvider
@see FIM.Util.ClassLoaderObjectInputStream
*/
public class FIMFactory extends CertificateFactorySpi
{
    
    /**
    Creates a new factory
    */
    public FIMFactory()
    {
    }
        
    /**
    Generates a certificate object and initializes it with
    the data read from the input stream <code>inStream</code>.<br>
    The given input stream <code>inStream</code> must contain a single
    PublicKeyCertificate certificate.<br>
    The stream must immediately start with the certificate (and only this data is consumed).
    @param inStream an input stream with the certificate data.
    @return an AgentIdentity with the data from the input stream.
    @exception CertificateException on parsing errors.
    */
    public Certificate engineGenerateCertificate(InputStream inStream) throws CertificateException
    {
        FIMVoucher res=null;
        try
	    {	    
            ClassLoaderObjectInputStream in=new ClassLoaderObjectInputStream(inStream);
            in.setClassLoader(inStream.getClass().getClassLoader());
            res=new FIMVoucher();
    	    SEQUENCE decoded=null;
	        try
	        {
	            decoded=(SEQUENCE)DerCoder.decode(in);
	        }
	        catch(CodingException e)
	        {
	            throw new CertificateException("Could not decode identity: "+e.getMessage());
	        }
		    // Read subject certificate
	        String type=(String)((PrintableString)decoded.getComponentAt(0)).getValue();
   			byte[] buf=((OCTET_STRING)decoded.getComponentAt(1)).getWholeValue();
    		CertificateFactory cf=CertificateFactory.getInstance(type);
	    	ByteArrayInputStream bais=new ByteArrayInputStream(buf);
	        res.subject=cf.generateCertificate(bais);
	        // Read issuer certificate
	        type=(String)((PrintableString)decoded.getComponentAt(2)).getValue();
   			buf=((OCTET_STRING)decoded.getComponentAt(3)).getWholeValue();
    		cf=CertificateFactory.getInstance(type);
	    	bais=new ByteArrayInputStream(buf);
	        res.issuer=cf.generateCertificate(bais);
    	    // Read currency
	        String shortName=(String)((PrintableString)decoded.getComponentAt(4)).getValue();
	        Currency cur=Currency.findCurrency(shortName);
	        if(cur==null)
	            throw new CertificateException("Could not find currency "+cur+"!");
	        double value=((double)((BigInteger)((INTEGER)decoded.getComponentAt(5)).getValue()).longValue())/100.0;
	        res.value=new Price(cur,value);
	        String validity=(String)((PrintableString)decoded.getComponentAt(6)).getValue();
            DateFormat df=DateFormat.getDateInstance(DateFormat.SHORT,Locale.GERMANY);
            res.validTill=df.parse(validity);
            res.signAlgorithm=(String)((PrintableString)decoded.getComponentAt(7)).getValue();
		    res.signature=((OCTET_STRING)decoded.getComponentAt(8)).getWholeValue();          
   		}
	    catch(ParseException e)
    	{
	        throw new CertificateException("Error parsing back validity date: "+e.toString());
    	}
	    catch(IOException e)
    	{
	        throw new CertificateException(e.toString());
    	}
    	return res;
    }

    /**
    Returns a (possibly empty) collection view of the certificates read
    from the given input stream <code>inStream</code>.
    <br><strong>NOT IMPLEMENTED</strong>
    @param inStream the input stream with the certificates.
    @return a (possibly empty) collection view of java.security.cert.Certificate objects
    initialized with the data from the input stream.
    @exception CertificateException on parsing errors.
    */
    public Collection engineGenerateCertificates(InputStream inStream)throws CertificateException
    {
        throw new RuntimeException("Not implemented (Only single certificates supported).!");
    }

    /**
    Generates a certificate revocation list (CRL) object and initializes it
    with the data read from the input stream <code>inStream</code>.
    <br><strong>NOT IMPLEMENTED</strong>
    @param inStream an input stream with the CRL data.
    @return a CRL object initialized with the data from the input stream.
    @exception CRLException on parsing errors.
    */
    public CRL engineGenerateCRL(InputStream inStream)throws CRLException
    {
        throw new RuntimeException("Not implemented (CRLs not supported).!");
    }

    /**
    Returns a (possibly empty) collection view of the CRLs read
    from the given input stream <code>inStream</code>.
    <br><strong>NOT IMPLEMENTED</strong>
    @param inStream the input stream with the CRLs.
    @return a (possibly empty) collection view of java.security.cert.CRL objects initialized with the data from the input stream.
    @exception CRLException on parsing errors.
    */
    public Collection engineGenerateCRLs(InputStream inStream)throws CRLException
    {
        throw new RuntimeException("Not implemented (CRLs not supported).!");
    }
    
}