package FIM.test;

import java.security.cert.*;
import java.security.*;
import java.util.*;
import java.math.BigInteger;
import java.text.*;

import FIM.payment.*;

import iaik.asn1.*;

public class FIMVoucher extends java.security.cert.Certificate 
{
    // Package access, so we can fill in the values in the FIMFactory!
    java.security.cert.Certificate subject=null;
    java.security.cert.Certificate issuer=null;
    Price value=null;
    Date validTill=null;
    
	/** The algorithm used for signing */
	protected String signAlgorithm=null;
	/** The signature */
	protected byte[] signature=null;

    public FIMVoucher()
    {
        super("FIM-Voucher");
    }
    
    public FIMVoucher(java.security.cert.Certificate subject,java.security.cert.Certificate issuer,Price value,Date validTill,PrivateKey pk)
    {
        super("FIM-Voucher");
        if(subject==null)
            throw new IllegalArgumentException("Subject certificate must be provided");
        this.subject=subject;
        if(issuer==null)
            throw new IllegalArgumentException("Issuer certificate must be provided");
        this.issuer=issuer;
        if(value==null)
            throw new IllegalArgumentException("Value must be provided");
        this.value=value;
        if(validTill==null)
            throw new IllegalArgumentException("Validity date be provided");
        this.validTill=validTill;
        checkValidity();
        try
        {
            String algo=pk.getAlgorithm();
            if(algo.equalsIgnoreCase("RSA"))
                algo="MD5/RSA";
            Signature sign=Signature.getInstance(algo);
            sign.initSign(pk);
            sign.update(getEncodedForSigning());
            signature=sign.sign();
        }
        catch(Exception e)
        {
            throw new IllegalArgumentException("Could not sign voucher: "+e.toString());
        }
    }
    
    public Price getValue()
    {
        return value;
    }
    
    public java.security.cert.Certificate getSubject()
    {
        return subject;
    }
    
    public java.security.cert.Certificate getIssuer()
    {
        return issuer;
    }
    
    public Date getValidTill()
    {
        return validTill;
    }
    
    public void checkValidity() throws IllegalStateException
    {
        Calendar cal=Calendar.getInstance();
        cal.setTime(new Date());
        if(cal.after(validTill))
            throw new IllegalStateException("Voucher already expired!");
    }
    
    /**
    Returns the encoded form of this certificate. It is assumed that each certificate type would have 
    only a single form of encoding; for example, X.509 certificates would be encoded as ASN.1 DER.
    @return the encoded form of this certificate
    @throws CertificateEncodingException if an encoding error occurs
    */
    public byte[] getEncoded() throws CertificateEncodingException
    {
        SEQUENCE encoded=new SEQUENCE();
        // Write subject certificate
        encoded.addComponent(new PrintableString(subject.getType()));
        encoded.addComponent(new OCTET_STRING(subject.getEncoded()));
        // Write issuer certificate
        encoded.addComponent(new PrintableString(issuer.getType()));
        encoded.addComponent(new OCTET_STRING(issuer.getEncoded()));
        // Write currency and value
        encoded.addComponent(new PrintableString(value.getCurrency().getShortName()));
        encoded.addComponent(new INTEGER(BigInteger.valueOf((long)(value.getPrice()*100))));
        // Write validity date
        DateFormat df=DateFormat.getDateInstance(DateFormat.SHORT,Locale.GERMANY);
        encoded.addComponent(new PrintableString(df.format(validTill)));
        // Write signature algorithm
        encoded.addComponent(new PrintableString(signAlgorithm));
        // Write signature
        encoded.addComponent(new OCTET_STRING(signature));
	    return DerCoder.encode(encoded);
    }

    protected byte[] getEncodedForSigning() throws CertificateEncodingException
    {
        SEQUENCE encoded=new SEQUENCE();
        // Write subject certificate
        encoded.addComponent(new PrintableString(subject.getType()));
        encoded.addComponent(new OCTET_STRING(subject.getEncoded()));
        // Write issuer certificate
        encoded.addComponent(new PrintableString(issuer.getType()));
        encoded.addComponent(new OCTET_STRING(issuer.getEncoded()));
        // Write currency and value
        encoded.addComponent(new PrintableString(value.getCurrency().getShortName()));
        encoded.addComponent(new INTEGER(BigInteger.valueOf((long)(value.getPrice()*100))));
        // Write validity date
        DateFormat df=DateFormat.getDateInstance(DateFormat.SHORT,Locale.GERMANY);
        encoded.addComponent(new PrintableString(df.format(validTill)));
	    return DerCoder.encode(encoded);
    }

    /**
    Verifies that this certificate was signed using the private key that corresponds to the specified 
    public key.
    @param key the PublicKey used to carry out the verification
    @throws NoSuchAlgorithmException on unsupported signature algorithms
    @throws InvalidKeyException on incorrect key
    @throws NoSuchProviderException if there's no default provider
    @throws SignatureException on signature errors
    @throws CertificateException on encoding errors
    */
    public void verify(PublicKey key) throws CertificateException, NoSuchAlgorithmException,InvalidKeyException, NoSuchProviderException,SignatureException
    {
        verify(key,null);
    }

    /**
    Verifies that this certificate was signed using the private key that corresponds to the 
    specified public key. This method uses the signature verification engine supplied by 
    the specified provider.
    @param key the PublicKey used to carry out the verification
    @param sigProvider the name of the signature provider
    @throws NoSuchAlgorithmException on unsupported signature algorithms
    @throws InvalidKeyException on incorrect key
    @throws NoSuchProviderException on incorrect provider
    @throws SignatureException on signature errors
    @throws CertificateException on encoding errors
    */
    public void verify(PublicKey key, String sigProvider) throws CertificateException, NoSuchAlgorithmException,InvalidKeyException, NoSuchProviderException,SignatureException
    {
   		Signature sign=null;
   		if(sigProvider==null)
        	sign=Signature.getInstance(signAlgorithm);
   		else
        	sign=Signature.getInstance(signAlgorithm,sigProvider);
    	sign.initVerify(issuer.getPublicKey());
   		sign.update(getEncodedForSigning());
    	if(!sign.verify(signature))
    		throw new SignatureException("Signature invalid");
    }

    /**
    Returns a string representation of this certificate.
    @return a string representation of this certificate
    */
    public String toString()
    {
        String str="FIMVoucher for "+subject.toString()+"\n Value: "+value.toString()+"\n";
        DateFormat df=DateFormat.getDateInstance(DateFormat.SHORT);
        str+="Valid till: "+df.format(validTill);
        return str;
    }

    /**
    Gets the public key from this certificate.
    @return the public key
    */
    public PublicKey getPublicKey()
    {
        return subject.getPublicKey();
    }
}
